#ifndef BODY_H
#define BODY_H


#include <vector>
#include <list>
#include <string>
#include <vector3D.h>



#include <PQP.h>

class Color
{
  public:
   float r, g, b, a;

   Color()
   {
     r= g= b= a= 1.0; 
   }
   void set(float red, float green ,float blue, float alpha= 1.0)
   {
     r= red;
     g= green;
     b= blue;
     a= alpha;
   }
};

/**
 * \enum PrimitiveType
 *
 * The different types of primitive.
 */
typedef enum PrimitiveType
{
  BOX,
  CYLINDER,
  SPHERE
} PrimitiveType;

/**
 * \enum Material
 *
 * Materials a body can be made of (used when a body model is exported to POVRay).
 */
typedef enum Material
{
  PLASTIC,
  SHINY_PLASTIC,
  METAL,
  GLASS,
  CONCRETE
} Material;

std::string materialToString(Material material);

/**
 * \enum BodyType
 *
 * Role of the body (tells wether if it used for display, collision tests or both).
 */
typedef enum BodyType
{
  DISPLAY_ONLY,
  COLLISION_ONLY,
  DISPLAY_AND_COLLISION,
  NO_DISPLAY_NO_COLLISION
} BodyType;

/*! \brief The Triangle class is used by the TriangleMesh class. */
class Triangle
{
  public:
   //! indices of the triangle vertices (in the vertex array of a TriangleMesh variable)
   unsigned int v[3]; 

   //! indices of the texture vertices of the triangle (in the corresponding array of a TriangleMesh variable).
   //! It is set to -1 by default to indicate that there is no texture coordinates defined for this triangle.
   int tv[3]; 

   //!indices of the vertex normals of the triangle (in the corresponding array of a TriangleMesh variable)
   //! It is set to -1 by default to indicate that there is no vertex normal defined for this triangle.
   int vn[3]; 

   //! normal of the triangle surface
   Vector3D normal; 

   //! RBGA color of the triangle
   Color color;

   Triangle()
   {
     v[0]= v[1]= v[2]= 0;
     tv[0]= tv[1]= tv[2]= -1;
     vn[0]= vn[1]= vn[2]= -1;
     color.set(0.0, 0.0, 0.0, 0.0);
   }
   void setVertexIndices(unsigned int i1, unsigned int i2, unsigned int i3)
   {
     v[0]= i1; v[1]= i2; v[2]= i3;
   }
   void setNormalIndices(unsigned int i1, unsigned int i2, unsigned int i3)
   {
     vn[0]= i1; vn[1]= i2; vn[2]= i3;
   }
   void setTextureVertexIndices(unsigned int i1, unsigned int i2, unsigned int i3)
   {
     tv[0]= i1; tv[1]= i2; tv[2]= i3;
   }
   void noTextureVertex()
   {
     tv[0]= tv[1]= tv[2]= -1;
   }
};


/*! \brief The TriangleMesh class is used to describe the geometry of a body. */
class TriangleMesh
{
 public:
  //! geometric vertices of the mesh
  std::vector<Vector3D> vertices;

  //! array of normal vectors for the mesh vertices
  std::vector<Vector3D> normals;  /*!< NB: there is no one-to-one mapping between the normal array and the vertex array */

  //! texture coordinates of the mesh vertices
  std::vector<Vector2D> tvertices; /*!< NB: there is no one-to-one mapping between the texture vertex array and the vertex array */

  //! triangles of the mesh
  std::vector<Triangle> triangles; 

  TriangleMesh();
  TriangleMesh(const TriangleMesh &mesh);
  ~TriangleMesh();
  unsigned int nbVertices() const  { return vertices.size(); }
  unsigned int nbNormals() const   { return normals.size(); }
  unsigned int nbTriangles() const { return triangles.size(); }
  int getVertex(unsigned int index, Vector3D &vertex);
  int getTriangleVertices(unsigned int index, Vector3D &p1, Vector3D &p2, Vector3D &p3);
  TriangleMesh & operator=(const TriangleMesh &mesh);
  int addVertex(float x, float y, float z);
  int addNormal(float x, float y, float z);
  int addTextureVertex(float u, float v);
  int addTriangle(Triangle t);
  int addTriangle(Vector3D a, Vector3D b, Vector3D c);

  int makeBox(float dimX, float dimY, float dimZ);
  int makeCylinder(float radius, float height, unsigned int nbSlices= 20);
  int makeSphere(float radius, unsigned int nbSlices= 20, unsigned int nbStacks= 10);
  int readOBJfile(std::string filename, float scale= 1.0);
  int saveAsOBJfile(std::string filename, float scale= 1.0);
  int readP3Dfile(std::string filename, float scale= 1.0);
  int reset();
  void draw(bool smooth= true);
  int computeTriangleNormals();
  int computeVertexNormals();
  int vertexTransform(const HTMatrix &T);
  int invertTriangleNormals();
  Vector3D closestPoint(Vector3D p); /*!< to be finished */
  bool pointInclusionTest(Vector3D p); /*!< to be finished */
  int fitInSphere(float radius);
  class AABB* kdTreeDecomposition(float resolution);
};

class AABBTree
{
  std::list<class AABB*> tree;
  class AABB *root;
};

typedef enum AABB_state
{
  INSIDE,
  OUTSIDE,
  NEITHER
} AABB_state;


//! experimental Axis Aligned Bounding Box class (WIP)
class AABB
{
  private:
    AABB_state _state;
    float _xmin, _xmax, _ymin, _ymax, _zmin, _zmax;
    Vector3D _center;
    float _dimX, _dimY, _dimZ;
    bool _root, _leaf, _terminate;
    float _minimum_size;
    unsigned int _level;
    class AABB *_children[2];
    class AABB *_brother;   
    class TriangleMesh *_mesh;
    std::list<unsigned int> _inner_triangles;
  public:
    AABB(Vector3D center, float dimX, float dimY, float dimZ, float minSize, TriangleMesh *mesh, std::list<unsigned int> inner_triangles);
    AABB(float xmin, float xmax, float ymin, float ymax, float zmin, float zmax, float minSize, TriangleMesh *mesh, std::list<unsigned int> inner_triangles);

    AABB(AABB *previous, Vector3D center, float dimX, float dimY, float dimZ);
    AABB(AABB *previous, float xmin, float xmax, float ymin, float ymax, float zmin, float zmax);

    bool isPointInside(Vector3D p);
    int divide();
    int divide2();
    int adjust(int axis);
    int adjust();
    int testTriangles();
    int testTrianglesSmall(int axis);
    int testTrianglesLarge();
    int draw();
    int draw(unsigned int level, bool fill= false);   
};

/*! \brief The Body class is used for robot links and environment obstacles. */
class Body
{
  protected:
    std::string _name; /*!< name of the body */
    HTMatrix _pose; /*!< homogeneous matrix of the body (absolute) pose */
    TriangleMesh _mesh;  /*!< geometric model of the body (for display) */

    PQP_Model *_pqpModel; /*!< collision model of the body (used by PQP)  */


    BodyType _type; 
    Color _color; /*!< RGBA color of the body */ 
    std::string _texture; /*!< name of the texture to apply to the body */

    //for link bodies:
    class Joint *_joint;  /*!< pointer to the joint the body is bound to (NULL if the body is not linked to a joint)*/  
    HTMatrix _Tjoint;  /*!< pose of the body relatively to the joint frame */ 

  public:
    Material material; /*!< body's material (only used for POVRay export for now) */
    bool colliding;  /*!< indicates if the body is colliding or not */ 
    bool enable_collision; /*!< indicates if collision detection will be performed with this body */ 
    bool enable_display;  /*!< indicates if the body will be displayed or not */ 
    bool enable_shadow;  /*!< indicates if the shadow of the body on the floor will be displayed or not */
    bool enable_texture;  /*!< indicates if the body's texture will be displayed or not */

    Body();
    Body(BodyType type);
    Body(const Body &body);
    ~Body();

    Body(std::string name, BodyType type, std::string filename, float scale= 1.0);
    Body(const std::string &name, const BodyType type, const TriangleMesh &mesh);
    Body & operator = (const Body &body);
    std::string name();
    HTMatrix pose();
    TriangleMesh& mesh();
    Joint * joint();
    HTMatrix Tjoint();
    PQP_Model* pqpModel();
    BodyType type();

    int loadMeshFile(std::string filename, float scale); 
    void setPosition(float t[3]);
    void setPosition(float x, float y, float z);
    void setOrientation(float yaw, float pitch, float roll);
    void setPose(float x, float y, float z, float yaw, float pitch, float roll);
    void setPose(const HTMatrix &M);
    int setJoint(Joint *joint);
    void setTjoint(const HTMatrix &M);
    void rename(const std::string name);
    void setTexture(const std::string texture);
    void printConfiguration();
    void setColor(float r, float g, float b, float alpha);
    void setMaterial(Material _material);

    int draw(bool smooth);
    int drawInBlack();
    int exportMeshToPOVRay(std::string foldername);
    int addToPOVfile(FILE *file);
    int createPQPmodel();
    int drawPQPmodel(float red, float green, float blue);
    int drawPQPtriangle(unsigned int index);
    int drawOBBs(unsigned int level);
};

/*! \brief A box-shaped body class inheriting from Body. */
class Box: public Body
{
  private:
   float _dimX; /*!< box's dimension along X axis */
   float _dimY; /*!< box's dimension along Y axis */
   float _dimZ; /*!< box's dimension along Z axis */
  public:
   Box(std::string name, BodyType type, float dimX, float dimY, float dimZ);
   float dimX() { return _dimX; }
   float dimY() { return _dimY; }
   float dimZ() { return _dimZ; }
};

/*! \brief A cylinder-shaped body class inheriting from Body. */
class Cylinder: public Body
{
  private:
   float _radius; /*!< cylinder's radius */ 
   float _height; /*!< cylinder's height (length along Z axis) */
  public:
   Cylinder(std::string name, BodyType type, float _radius, float _height, unsigned int nbSlices= 20);
   float radius() { return _radius; }
   float height() { return _height; }
};

/*! \brief A sphere-shaped body class inheriting from Body. */
class Sphere: public Body
{
  private:
   float _radius; /*!< sphere's radius */ 
  public:
   Sphere(std::string name, BodyType type, float _radius, unsigned int nbSlices= 20, unsigned int nbStacks= 10);
   float radius() { return _radius; }
};


int drawOBBs_recursive(PQP_Model *pqpModel, float T[4][4], int bn, unsigned int currentLevel, unsigned int level);
int collisionTest(Body *body1, Body *body2);
float distanceComputation(Body *body1, Body *body2, float p1[3], float p2[3]);


#endif


