#ifndef FILE_READER_H
#define FILE_READER_H

#include <string>
#include <vector>
#include <list>
#include <iostream>
#include <sstream>
#include <robot.h>


#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/tree.h>


/**
 * \struct ElementParserData
 *
 * A structure to store the information about simple elements (elements without children).
 */
typedef struct ElementParserData
{
  std::string name, file, models_directory, robot_directory, robot_file, type;
  std::string subrobot, master, slave, texture, connection, material_type;
  JointType joint_type;
  ConstraintType constraint_type;
  bool enable_collision, enable_display, enable_shadow;
  float scale, dimX, dimY, dimZ, radius, height;
  float bounding_box[6];
  Vector3D translation;
  float YPR_angles[3], angle_axis[4], dh_parameters[4];
  HTMatrix ht_matrix;
  float color[4];
  std::vector<float> q, qmin, qmax, robot_configuration;
  std::vector<float> parameters;
  unsigned int master_dof, slave_dof;
} ElementParserData;

/**
 * \struct TransformationParserData
 *
 * A structure to store the information about a transformation element.
 */
typedef struct TransformationParserData
{
  HTMatrix matrix;
} TransformationParserData;

/**
 * \struct SubrobotParserData
 *
 * A structure to store the information about a subrobot element.
 */
typedef struct SubrobotParserData
{
  std::string connection;
  HTMatrix T;
  std::string robot_file;
  bool enable_collision, enable_display;
} SubrobotParserData;


/**
 * \struct BodyParserData
 *
 * A structure to store the information about a body element.
 */
typedef struct BodyParserData
{
  std::string name, file, texture;
  float scale;
  bool is_primitive, enable_collision, enable_display, enable_shadow;
  HTMatrix T;
  float color[4], dimX, dimY, dimZ, radius, height;
  PrimitiveType primitive_type;
  Material material;
} BodyParserData;


/**
 * \struct ConstraintParserData
 *
 * A structure to store the information about a constraint element.
 */
typedef struct ConstraintParserData
{
  std::string master, slave;
  ConstraintType type;
  unsigned int master_dof, slave_dof;
  std::vector<float> parameters;
} ConstraintParserData;


/**
 * \struct JointParserData
 *
 * A structure to store the information about a joint element.
 */
typedef struct JointParserData
{
  std::string name;
  bool relative;
  JointType type;
  Vector3D axis, center, translation;
  float dh_parameters[4];
  HTMatrix transformation;
  std::vector<float> q, qmin, qmax;
} JointParserData;

int logfile(const char *format,...);
bool isElementDefined(std::string element);



std::string getNodeString(xmlDocPtr doc, xmlNodePtr node);
void warningMessage(int line_number, const xmlChar *URL, const xmlChar *element_name, std::string &message);
void formatErrorMessage(int line_number, const xmlChar *URL, const xmlChar *element_name, std::string &message);
void elementMissingMessage(int line_number, const xmlChar *URL, const xmlChar *element_name, std::string &message);
bool parseElement(xmlDocPtr doc, xmlNodePtr entry_node, std::string element, ElementParserData &data);
bool parseTransformation(xmlDocPtr doc, xmlNodePtr entry_node, TransformationParserData &data);
bool parseSubrobot(xmlDocPtr doc, xmlNodePtr entry_node, SubrobotParserData &data);
bool parseBody(xmlDocPtr doc, xmlNodePtr entry_node, BodyParserData &data);
bool parseJoint(xmlDocPtr doc, xmlNodePtr entry_node, Robot *robot);
bool parseCollisionPair(xmlDocPtr doc, xmlNodePtr entry_node, std::string &bodyName1, std::string &bodyName2);

int parseRobotFile(std::string path, std::string filename, Robot *robot);
int parseWorldFile(std::string path, std::string filename, World *world);

//! Converts a variable to a string.
template<class T> std::string convertToString(const T& t)
{
  std::ostringstream stream;
  stream << t;
  return stream.str();
}

#endif

