#ifndef JOINT_H
#define JOINT_H

#include <vector>
#include <string>
#include <list>
#include <body.h>
#include <math.h>

#define ANGLE_STEP  ( M_PI/180.0f )
#define TRANSLATION_STEP  0.01f

/**
 * \enum DOFstate
 *
 * DOFstate enumerates the different possible states of a degree of freedom (DOF).
 */
typedef enum DOFstate
{
  INDEPENDENT, /*!< Ordinary DOF (active). */ 
  CONSTRAINED, /*!< Passive DOF: its value is computed from another DOF value. */ 
  BLOCKED /*!< The value of the DOF is constant (equal to the minimal and maximal bounds (that are the same)). */ 
} DOFstate;

/**
 * \enum DOFunit
 *
 * DOFUnit enumerates the different possible units of a degree of freedom (DOF).
 */
typedef enum DOFunit
{
  METER,
  RADIAN,
  DEGREE,
} DOFUnit;

/** WIP
 * \enum DOFtopology
 *
 * DOFtopology enumerates the different possible topology of a degree of freedom (DOF).
 */
typedef enum DOFtopology
{
  R_TOPOLOGY,  /*!< unit interval topology */ 
  S_TOPOLOGY,  /*!< circle topology */ 
  SO3_TOPOLOGY, /*!< SO3 (space rotations) topology */ 
} DOFtopology;

/**
 * \enum ConstraintType
 *
 * DOFUnit enumerates the different types of constraint (for now, there is only one).
 */
typedef enum ConstraintType
{
  LINEAR
} ConstraintType;


/*! \brief The Constraint class is used to describe how active DOFs control passive DOFs. */
class Constraint
{
 public:
  //! type of the constraint
  ConstraintType type;

  //! vector of the constraint parameters (its size depends on the constraint type)
  std::vector<float> parameters;

  //! pointer to the master (active) joint
  class Joint *master;

  //! pointer to the slave (passive) joint
  class Joint *slave;

  //! index of the master's DOF that will control the slave's DOF 
  unsigned int master_dof;

  //! index of the slave's DOF that will be controlled by the master's DOF 
  unsigned int slave_dof;

  Constraint()
  {
    type= LINEAR;
    master= NULL;
    slave= NULL;
  }
};


/*! \brief  The class DOF contains all the information about a degree of freedom. */
class DOF
{
  private:
    float _min; /*!< minimal bound on the DOF value */
    float _max; /*!< maximal bound on the DOF value */
    float _value; /*!< current value in the associated unit */
    float _velocity; /*!< current velocity in the associated unit */
    float _acceleration; /*!< current acceleration in the associated unit */
    DOFunit _unit; /*!< unit of the DOF */
  public:
    DOFstate state;
    DOFtopology topology;

    DOF();
    float min();
    float max();
    float value();
    float velocity();
    float acceleration();
    DOFunit unit();
    int setValue(float value_);
    int setVelocity(float velocity_);
    int setAcceleration(float acceleration_);
    int setUnit(DOFunit unit);
    int setBounds(float min, float max);
    int getBounds(float &min, float &max);
};

/**
 * \enum JointType
 *
 * JointType enumerates the different joint types: FREEFLYER (6 dofs), REVOLUTE (1 dof) and PRISMATIC (1 dof).
 */
typedef enum JointType {
  BASE,  /*!< R R S */
  FREEFLYER,  /*!< R R R SO3 SO3 SO3 */
  REVOLUTE,  /*!< R  */
  S_REVOLUTE, /*!< S */
  PRISMATIC /*!< R  */
} JointType;


/*! \brief The class Joint used by the class Robot. */
class Joint
{
  private:
    std::string _name; /*!< name of the joint */

  public:
    JointType type;    /*!< type of the joint (see JointType) */
    Joint *previous;  /*!< pointer to the previous joint in the kinematic tree */
    std::vector<Joint*> next;  /*!< vector of pointer(s) to the next joint(s) in the kinematic tree */
    std::vector<DOF> q; /*!< DOF vector of the joint */
    HTMatrix frame; /*!< current joint frame in the global referential */
    HTMatrix Tconst; /*!<  constant transform between the previous joint frame and the current one */

    //! transform due to the joint with its current parameter value
    /*! NB: we have the relation: frame= (previous_joint_frame) * Tconst * T   */
    HTMatrix T; 

    //! indicates if the values of 'frame' and 'T' are up-to-date
    bool up_to_date; 

    bool selected;  /*!< for GUI: indicates if the joint is selected or not */

    Joint(JointType _type);
    ~Joint();
    std::string name()    { return _name; }
    void rename(const std::string name);
    int draw(float length);
    int printInfo();
    int computeTransform();
    int setDOFvalue(unsigned int index, float x);
    int getDOFvalue(unsigned int index, float &x);
    int setDOFvelocity(unsigned int index, float x);
    int getDOFvelocity(unsigned int index, float &x);
    int setDOFacceleration(unsigned int index, float x);
    int getDOFacceleration(unsigned int index, float &x);
    int incrementDOFvalue(unsigned int index, float ratio);
    int decrementDOFvalue(unsigned int index, float ratio);
    unsigned int nbDOFs();
};

std::string jointTypeToString(JointType type);
std::string DOFTopologyToString(DOFtopology topology);
std::string DOFunitToString(DOFunit unit);
std::string DOFstateToString(DOFstate state);

#endif
