#include <stdio.h>
#include <math.h>
#include <iostream>
#include <fstream>
#include <sstream>
#if defined(ENABLE_DISPLAY)
#include <GL/gl.h>
#endif
#include <coldman.h>
#include <body.h>
#include <robot.h>
#include <file_reader.h>

#if defined(ENABLE_DISPLAY)
static double PALETTE[16][3]= { {1,1,1}, {1,0,0}, {0,1,0}, {0,0,1}, {1,1,0}, {1,0,1}, {0,1,1} , {1,0.5,0.5}, {0.5,1,0.5}, {0.5,0.5,1}, {1,0.25,0.5}, {1,0.5,0.25}, {0.25,1.0,0.5}, {0.5,1,0.25}, {0.25,0.5,1}, {0.5,0.25,1}  };
static int PALETTE_SIZE= 16;
static int CURRENT_COLOR= 0;
#endif

//! Converts a Material to a std::string.
std::string materialToString(Material material)
{
  switch(material)
  {
    case PLASTIC:
      return "PLASTIC";
    break;
    case SHINY_PLASTIC:
      return "SHINY_PLASTIC";
    break;
    case METAL:
      return "METAL";
    break;
    case GLASS:
      return "GLASS";
    break;
    case CONCRETE:
      return "CONCRETE";
    break;
  }

  return "UNDEFINED";
}

TriangleMesh::TriangleMesh()
{

}

TriangleMesh::TriangleMesh(const TriangleMesh &mesh)
{
  unsigned int i, j;
  Vector2D p2;
  Vector3D p3;
  Triangle triangle;

  vertices.clear();
  normals.clear();
  tvertices.clear();
  triangles.clear();

  vertices.reserve(mesh.vertices.size());  
  normals.reserve(mesh.normals.size());  
  tvertices.reserve(mesh.tvertices.size());  
  triangles.reserve(mesh.triangles.size());  

  for(i=0; i<mesh.vertices.size(); ++i)
  {
     p3= mesh.vertices[i];
     vertices.push_back(p3);
  }

  for(i=0; i<mesh.normals.size(); ++i)
  {
     p3= mesh.normals[i];
     normals.push_back(p3);
  }

  for(i=0; i<mesh.tvertices.size(); ++i)
  {
     p2= mesh.tvertices[i];
     tvertices.push_back(p2);
  }

  for(i=0; i<mesh.triangles.size(); ++i)
  {
     for(j=0; j<3; j++)
     {
       triangle.v[j] = mesh.triangles[i].v[j];
       triangle.tv[j]= mesh.triangles[i].tv[j];
       triangle.vn[j]= mesh.triangles[i].vn[j];
     }     
     triangle.normal= mesh.triangles[i].normal;

     triangles.push_back(triangle);
  }
}

TriangleMesh::~TriangleMesh()
{
  vertices.clear();
  normals.clear();
  tvertices.clear();
  triangles.clear();
}


TriangleMesh & TriangleMesh::operator = (const TriangleMesh &mesh)
{
  if(this==&mesh)
  {  return *this;  }

  unsigned int i, j;
  Vector2D p2;
  Vector3D p3;
  Triangle triangle;

  vertices.clear();  
  normals.clear();  
  tvertices.clear();  
  triangles.clear();  

  vertices.reserve(mesh.vertices.size());  
  normals.reserve(mesh.normals.size());  
  tvertices.reserve(mesh.tvertices.size());  
  triangles.reserve(mesh.triangles.size());  

  for(i=0; i<mesh.vertices.size(); ++i)
  {
     p3= mesh.vertices[i];
     vertices.push_back(p3);
  }

  for(i=0; i<mesh.normals.size(); ++i)
  {
     p3= mesh.normals[i];
     normals.push_back(p3);
  }

  for(i=0; i<mesh.tvertices.size(); ++i)
  {
     p2= mesh.tvertices[i];
     tvertices.push_back(p2);
  }

  for(i=0; i<mesh.triangles.size(); ++i)
  {
     for(j=0; j<3; j++)
     {
       triangle.v[j] = mesh.triangles[i].v[j];
       triangle.tv[j]= mesh.triangles[i].tv[j];
       triangle.vn[j]= mesh.triangles[i].vn[j];
     }     
     triangle.normal= mesh.triangles[i].normal;

     triangles.push_back(triangle);
  }

  return *this;
}

//! Returns the coordinates of the mesh vertex with given index.
//! \param index index of the vertex in the mesh vertex array (starts from zero)
//! \param vertex a 3D vector that will be filled with the coordinates of the vertex
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::getVertex(unsigned int index, Vector3D &vertex)
{
  if(vertices.empty())
  {
    printf("%s: %d: TriangleMesh::getVertex(): there is no vertex in the mesh.\n", __FILE__, __LINE__);
    return 0; 
  }

  if( index > vertices.size()-1 )
  {
    printf("%s: %d: TriangleMesh::getVertex(): vertex index (%d) exceeds the vertex number of the mesh (%d).\n", __FILE__, __LINE__, index, vertices.size());
    return 0;
  }

  vertex= vertices[index];

  return 1;
}

//! Gets the three vertices of the triangle with given index.
//! \param index index of the triangle in the mesh triangle array (starts from zero)
//! \param p1 a 3D vector that will be filled with the coordinates of the triangle first vertex
//! \param p2 a 3D vector that will be filled with the coordinates of the triangle second vertex
//! \param p3 a 3D vector that will be filled with the coordinates of the triangle third vertex
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::getTriangleVertices(unsigned int index, Vector3D &p1, Vector3D &p2, Vector3D &p3)
{
  if(triangles.empty())
  {
    printf("%s: %d: TriangleMesh::getTriangleVertices(): there is no triangle in the mesh.\n", __FILE__, __LINE__);
    return 0; 
  }
  if( index > triangles.size()-1 )
  {
    printf("%s: %d: TriangleMesh::getTriangleVertices(): triangle index (%d) exceeds the triangle number of the mesh (%d).\n", __FILE__, __LINE__, index, triangles.size());
    return 0; 
  }

  if( !getVertex(triangles[index].v[0], p1) )
  {  return 0;  }

  if( !getVertex(triangles[index].v[1], p2) )
  {  return 0;  }

  if( !getVertex(triangles[index].v[2], p3) )
  {  return 0;  }


  return 1;
}

//! Adds a vertex to the mesh.
//! \param x x coordinate of the vertex to add
//! \param y y coordinate of the vertex to add
//! \param z z coordinate of the vertex to add
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::addVertex(float x, float y, float z)
{
  Vector3D v(x, y, z);

  vertices.push_back(v);
  return 1;
}

//! Adds a normal to the mesh.
//! \param x x coordinate of the normal to add
//! \param y y coordinate of the normal to add
//! \param z z coordinate of the normal to add
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::addNormal(float x, float y, float z)
{
  Vector3D v(x, y, z);
 
  if(!v.normalize())
  {
    v.set(0, 0, 1);
  }
  normals.push_back(v);
  return 1;
}

//! Adds a texture vertex to the mesh.
//! \param u u coordinate of the texture vertex to add
//! \param v v coordinate of the texture vertex to add
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::addTextureVertex(float u, float v)
{
  Vector2D vt(u, v);
 
  tvertices.push_back(vt);
  return 1;
}

//! Adds a triangle to the mesh. 
//! \param t  the triangle to add
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::addTriangle(Triangle t)
{
  Vector3D edge1, edge2;
  Triangle triangle;
  
  if(vertices.empty())
  {
    printf("%s: %d: TriangleMesh::addTriangle(): the mesh does not have any vertex yet.\n", __FILE__, __LINE__);
    return 0;
  }    

  if( (t.v[0] > vertices.size()-1) || (t.v[1] > vertices.size()-1) || (t.v[2] > vertices.size()-1) )
  {
    printf("%s: %d: TriangleMesh::addTriangle(): one of the triangle's vertex indices (%d %d %d) exceeds the mesh's number of vertices (%d).\n", __FILE__, __LINE__,t.v[0], t.v[1], t.v[2], vertices.size());
    return 0; 
  }

  if( (t.tv[0] > (int) tvertices.size()-1) || (t.tv[1] > (int) tvertices.size()-1) || (t.tv[2] > (int) tvertices.size()-1) )
  {
    printf("%s: %d: TriangleMesh::addTriangle(): one of the triangle's texture vertex indices (%d %d %d) exceeds the mesh's number of the texture vertices (%d).\n", __FILE__, __LINE__,t.tv[0], t.tv[1], t.tv[2], tvertices.size());
    return 0; 
  }

  if( (t.vn[0] > (int) normals.size()-1) || (t.vn[1] > (int) normals.size()-1) || (t.vn[2] > (int) normals.size()-1) )
  {
    printf("%s: %d: TriangleMesh::addTriangle(): one of the triangle's normal indices (%d %d %d) exceeds the mesh's number of normals (%d).\n", __FILE__, __LINE__,t.vn[0], t.vn[1], t.vn[2], normals.size());
    return 0; 
  }

  edge1= vertices[t.v[1]] - vertices[t.v[0]];
  edge2= vertices[t.v[2]] - vertices[t.v[0]];
 
  t.normal= crossProduct(edge1, edge2);
  t.normal.normalize();

  triangles.push_back(t);
  
  return 1;
}

//! Adds a triangle to the mesh from three new vertices.
//! This function must be used only to create a "triangle soup" mesh. 
//! \param a  the triangle's first vertex
//! \param b  the triangle's second vertex
//! \param c  the triangle's third vertex
//! \return OK in case of success, ERROR otherwise
int TriangleMesh::addTriangle(Vector3D a, Vector3D b, Vector3D c)
{
  if(this==NULL)
  {
    printf("%s: %d: TriangleMesh::addTriangle(): the calling instance is NULL.\n",__FILE__,__LINE__);
    return ERROR;
  }

  Vector3D edge1, edge2;
  Triangle triangle;

  addVertex(a.x, a.y, a.z);
  addVertex(b.x, b.y, b.z);
  addVertex(c.x, c.y, c.z);

  triangle.v[0]= nbVertices()-3;
  triangle.v[1]= nbVertices()-2;
  triangle.v[2]= nbVertices()-1;

  edge1= b - a;
  edge2= c - a;
 
  triangle.normal= crossProduct(edge1, edge2);
  triangle.normal.normalize();

  triangles.push_back(triangle);
  
  return OK;
}


//! Fills the mesh to be a box, centered on (0,0,0).
//! The texture coordinate unit square is projected on each face.
//! \param dimX box dimension along X axis
//! \param dimY box dimension along Y axis
//! \param dimZ box dimension along Z axis
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::makeBox(float dimX, float dimY, float dimZ)
{
  if( (dimX < 0) || (dimY < 0) || (dimZ < 0) )
  { 
    printf("%s: %d:  TriangleMesh::makeBox(): an input dimension is negative.\n", __FILE__, __LINE__);
    dimX= fabs(dimX);
    dimY= fabs(dimY);
    dimZ= fabs(dimZ);
  }

  Triangle triangle;

  reset();

  vertices.reserve(8);  
  normals.reserve(6);  
  tvertices.reserve(4);  
  triangles.reserve(12);  

  addVertex(-0.5*dimX, -0.5*dimY,  0.5*dimZ);
  addVertex( 0.5*dimX, -0.5*dimY,  0.5*dimZ);
  addVertex( 0.5*dimX,  0.5*dimY,  0.5*dimZ);
  addVertex(-0.5*dimX,  0.5*dimY,  0.5*dimZ);

  addVertex(-0.5*dimX, -0.5*dimY, -0.5*dimZ);
  addVertex( 0.5*dimX, -0.5*dimY, -0.5*dimZ);
  addVertex( 0.5*dimX,  0.5*dimY, -0.5*dimZ);
  addVertex(-0.5*dimX,  0.5*dimY, -0.5*dimZ);

  addNormal( 1,  0,  0);
  addNormal(-1,  0,  0);
  addNormal( 0,  1,  0);
  addNormal( 0, -1,  0);
  addNormal( 0,  0,  1);
  addNormal( 0,  0, -1);

  addTextureVertex(0, 0);
  addTextureVertex(1, 0);
  addTextureVertex(1, 1);
  addTextureVertex(0, 1);

  // +Z face:
  triangle.setVertexIndices(0, 1, 2);
  triangle.setNormalIndices(4, 4, 4);
  triangle.setTextureVertexIndices(0, 1, 2);
  addTriangle(triangle);
  triangle.setVertexIndices(0, 2, 3);
  triangle.setNormalIndices(4, 4, 4);
  triangle.setTextureVertexIndices(0, 2, 3);
  addTriangle(triangle);
 
  // -Z face:
  triangle.setVertexIndices(4, 7, 6);
  triangle.setNormalIndices(5, 5, 5);
  triangle.setTextureVertexIndices(3, 0, 1);
  addTriangle(triangle);
  triangle.setVertexIndices(4, 6, 5);
  triangle.setNormalIndices(5, 5, 5);
  triangle.setTextureVertexIndices(3, 1, 2);
  addTriangle(triangle);

  // +X face:
  triangle.setVertexIndices(1, 5, 6);
  triangle.setNormalIndices(0, 0, 0);
  triangle.setTextureVertexIndices(3, 0, 1);
  addTriangle(triangle);
  triangle.setVertexIndices(1, 6, 2);
  triangle.setNormalIndices(0, 0, 0);
  triangle.setTextureVertexIndices(3, 1, 2);
  addTriangle(triangle);

  // -X face:
  triangle.setVertexIndices(0, 3, 7);
  triangle.setNormalIndices(1, 1, 1);
  triangle.setTextureVertexIndices(2, 3, 0);
  addTriangle(triangle);
  triangle.setVertexIndices(0, 7, 4);
  triangle.setNormalIndices(1, 1, 1);
  triangle.setTextureVertexIndices(2, 0, 1);
  addTriangle(triangle);

  // +Y face:
  triangle.setVertexIndices(2, 6, 7);
  triangle.setNormalIndices(2, 2, 2);
  triangle.setTextureVertexIndices( 3, 0, 1);
  addTriangle(triangle);
  triangle.setVertexIndices(2, 7, 3);
  triangle.setNormalIndices(2, 2, 2);
  triangle.setTextureVertexIndices(3, 1, 2);
  addTriangle(triangle);

  // -Y face:
  triangle.setVertexIndices(0, 4, 5);
  triangle.setNormalIndices(3, 3, 3);
  triangle.setTextureVertexIndices( 3, 0, 1);
  addTriangle(triangle);
  triangle.setVertexIndices(0, 5, 1);
  triangle.setNormalIndices(3, 3, 3);
  triangle.setTextureVertexIndices(3, 1, 2);
  addTriangle(triangle);

  return 1;
}


//! Fills the mesh to be a cylinder, centered on (0,0,0) and aligned on Z axis.
//! The texture coordinate unit square is projected (rolled) on the side of the cylinder.
//! \param radius cylinder radius
//! \param height cylinder height
//! \param nbSlices number of segment around Z axis
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::makeCylinder(float radius, float height, unsigned int nbSlices)
{
  if( (radius < 0) || (height < 0) )
  { 
    printf("%s: %d:  TriangleMesh::makeCylinder(): an input dimension is negative.\n", __FILE__, __LINE__);
    radius= fabs(radius);
    height= fabs(height);
  }

  if(nbSlices<3)
  { nbSlices= 10; }

  unsigned int i;
  float *cosT= NULL, *sinT= NULL, alpha, dalpha;
  Triangle triangle;

  reset();

  vertices.reserve(2*nbSlices + 2);  
  normals.reserve(nbSlices + 2);  
  tvertices.reserve(2*nbSlices + 2);  
  triangles.reserve(4*nbSlices);  

  cosT= new float[nbSlices];
  sinT= new float[nbSlices];
  dalpha= 2*M_PI/((float) nbSlices);

  alpha= 0;
  for(i=0; i<nbSlices; ++i)
  {
    cosT[i]= cos(alpha);
    sinT[i]= sin(alpha);
    alpha+= dalpha;
  }

  // first, fill the vertex array:
  addVertex(0, 0, 0.5*height);
  addNormal(0, 0, 1.0);
  for(i=0; i<nbSlices; ++i)
  {
    addVertex(radius*cosT[i], radius*sinT[i], 0.5*height);
    addNormal(cosT[i], sinT[i], 0.0);
    addTextureVertex(((float) i)/((float) nbSlices) , 1.0);
  }
  addVertex(0, 0, -0.5*height);
  addNormal(0, 0, -1.0);
  for(i=0; i<nbSlices; ++i)
  {
    addVertex(radius*cosT[i], radius*sinT[i], -0.5*height);
    addTextureVertex(((float) i)/((float) nbSlices) , 0.0);
  }
  addTextureVertex(1.0 , 1.0);
  addTextureVertex(1.0 , 0.0);

  // fill the triangle array:
  // top and bottom faces:
  for(i=0; i<nbSlices-1; ++i) // top
  {
    triangle.setVertexIndices(0, 1+i, 2+i);
    triangle.setNormalIndices(0, 0, 0);
    triangle.noTextureVertex();
    addTriangle(triangle);
  }
  triangle.setVertexIndices(0, nbSlices, 1);
  triangle.setNormalIndices(0, 0, 0);
  triangle.noTextureVertex();
  addTriangle(triangle);

  for(i=0; i<nbSlices-1; ++i) // bottom
  {
    triangle.setVertexIndices(nbSlices+1, (nbSlices+1)+2+i, (nbSlices+1)+1+i);
    triangle.setNormalIndices(nbSlices+1, nbSlices+1, nbSlices+1);
    triangle.noTextureVertex();
    addTriangle(triangle);
  }
  triangle.setVertexIndices(nbSlices+1, (nbSlices+1)+1, nbSlices+1+nbSlices);
  triangle.setNormalIndices(nbSlices+1, nbSlices+1, nbSlices+1);
  triangle.noTextureVertex();
  addTriangle(triangle);

  // side faces:
  for(i=0; i<nbSlices-1; ++i)
  {
    triangle.setVertexIndices(1+i, nbSlices+2+i, 2+i);
    triangle.setNormalIndices(1+i, 1+i, 2+i);
    triangle.setTextureVertexIndices(i, nbSlices+i, 1+i);
    addTriangle(triangle);
    triangle.setVertexIndices(nbSlices+2+i, nbSlices+2+i+1, 2+i);
    triangle.setNormalIndices(1+i, 2+i, 2+i);
    triangle.setTextureVertexIndices(nbSlices+i, nbSlices+1+i, 1+i);
    addTriangle(triangle);
  }
  triangle.setVertexIndices(nbSlices, 2*nbSlices+1, 1);
  triangle.setNormalIndices(nbSlices, nbSlices, 1);
  triangle.setTextureVertexIndices(nbSlices-1, 2*nbSlices-1, tvertices.size()-2);
  addTriangle(triangle);
  triangle.setVertexIndices(2*nbSlices+1, nbSlices+2, 1);
  triangle.setNormalIndices(nbSlices, 1, 1);
  triangle.setTextureVertexIndices(2*nbSlices-1, tvertices.size()-1, tvertices.size()-2);
  addTriangle(triangle);

  delete [] cosT;
  delete [] sinT;

  return 1;
}

//! Fills the mesh to be a sphere.
//! The texture coordinate unit square is rolled around Z-axis.
//! \param radius sphere radius
//! \param nbSlices number of segment around Z axis
//! \param nbStacks number of divisions along Z axis
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::makeSphere(float radius, unsigned int nbSlices, unsigned int nbStacks)
{
  if(radius < 0)
  { 
    printf("%s: %d:  TriangleMesh::makeSphere(): an input dimension is negative.\n", __FILE__, __LINE__);
    radius= fabs(radius);
  }

  if(nbSlices<3)
  { nbSlices= 10; }

  if(nbStacks<3)
  { nbStacks= 10; }


  unsigned int i, j, shift1, shift2;
  unsigned int i1, i2, i3, it1, it2, it3;
  float u, v;
  float *cosT= NULL, *sinT= NULL, alpha, dalpha;
  float *cosTz= NULL, *sinTz= NULL, beta, dbeta;
  Triangle triangle;

  reset();

  vertices.reserve(nbSlices*(nbStacks-1) + 2);  
  normals.reserve(nbSlices*(nbStacks-1) + 2);  
  tvertices.reserve(nbSlices*(nbStacks-1) + 2);  
  triangles.reserve((nbSlices+1)*(nbStacks-1)+2);  

  cosT= new float[nbSlices];
  sinT= new float[nbSlices];
  dalpha= 2*M_PI/((float) nbSlices);

  cosTz= new float[nbStacks];
  sinTz= new float[nbStacks];
  dbeta= M_PI/((float) nbStacks);

  alpha= 0;
  for(i=0; i<nbSlices; ++i)
  {
    cosT[i]= cos(alpha);
    sinT[i]= sin(alpha);
    alpha+= dalpha;
  }

  beta= 0;
  for(i=0; i<nbStacks; ++i)
  {
    cosTz[i]= cos(beta);
    sinTz[i]= sin(beta);
    beta+= dbeta;
  }


  // fill the vertex array:
  addVertex(0, 0, radius);
  addNormal(0, 0, 1.0);
  for(i=0; i<nbSlices; ++i)
  {  addTextureVertex(1.0/((float) 2*nbSlices) + ((float) i)/((float) nbSlices), 1.0);  }

  shift1= tvertices.size();
  for(i=1; i<nbStacks; ++i)
  {
    for(j=0; j<nbSlices; j++)
    {
      addVertex(radius*sinTz[i]*cosT[j], radius*sinTz[i]*sinT[j], radius*cosTz[i]);
      addNormal(sinTz[i]*cosT[j], sinTz[i]*sinT[j], cosTz[i]);
      u= ((float) j)/((float) nbSlices);
      v= 1.0 - ((float) i)/((float) nbStacks);
      addTextureVertex(u, v); 
    }
    u= 1.0;
    v= 1.0 - ((float) i)/((float) nbStacks);
    addTextureVertex(u, v);  
  }
  addVertex(0, 0, -radius);
  addNormal(0, 0, -1.0);
  shift2= tvertices.size()-1;
  for(i=0; i<nbSlices; ++i)
  {  addTextureVertex(1.0/((float) 2*nbSlices) + ((float) i)/((float) nbSlices), 0.0);  }


  //fill the triangle array:
  // triangle fan on the top:
  shift1= nbSlices;
  for(i=0; i<nbSlices-1; ++i)
  {
    i1= 0;    i2= 1+i;    i3= 2+i;
    triangle.setVertexIndices(i1, i2, i3);
    triangle.setNormalIndices(i1, i2, i3);

    it1= i;    it2= nbSlices+i;    it3= it2+1;
    triangle.setTextureVertexIndices(it1, it2, it3);
    addTriangle(triangle);
  }
  i1= 0;    i2= nbSlices;    i3= 1;
  triangle.setVertexIndices(i1, i2, i3);
  triangle.setNormalIndices(i1, i2, i3);

  it1= nbSlices-1;  it2= 2*nbSlices-1;  it3= 2*nbSlices;
  triangle.setTextureVertexIndices(it1, it2, it3);
  addTriangle(triangle);

  // triangle strip around the sphere:
  for(i=0; i<nbStacks-2; ++i)
  {
    for(j=0; j<nbSlices-1; j++)
    {
      i1= 1+j+i*nbSlices;    i2= 1+j+(i+1)*nbSlices;    i3= i1+1;
      triangle.setVertexIndices(i1, i2, i3);
      triangle.setNormalIndices(i1, i2, i3);

      it1= nbSlices+j+i*(nbSlices+1);    it2= nbSlices+j+(i+1)*(nbSlices+1);    it3= it1+1;
      triangle.setTextureVertexIndices(it1, it2, it3);
      addTriangle(triangle);

      i1= i2;    i2= i2+1;
      triangle.setVertexIndices(i1, i2, i3);
      triangle.setNormalIndices(i1, i2, i3);

      it1= nbSlices+j+(i+1)*(nbSlices+1);    it2= nbSlices+j+(i+1)*(nbSlices+1)+1;
      triangle.setTextureVertexIndices(it1, it2, it3);
      addTriangle(triangle);
    }
    i1= nbSlices+i*nbSlices;    i2= nbSlices+(i+1)*nbSlices;    i3= 1+i*nbSlices;
    triangle.setVertexIndices(i1, i2, i3);
    triangle.setNormalIndices(i1, i2, i3);

    it1= 2*nbSlices-1+i*(nbSlices+1);    it2= 2*nbSlices-1+(i+1)*(nbSlices+1);    it3= it1+1;
    triangle.setTextureVertexIndices(it1, it2, it3);
    addTriangle(triangle);

    i1= nbSlices+(i+1)*nbSlices;    i2= (i+1)*nbSlices+1;    i3= 1+i*nbSlices;
    triangle.setVertexIndices(i1, i2, i3);
    triangle.setNormalIndices(i1, i2, i3);

    it1= 2*nbSlices-1+(i+1)*(nbSlices+1);    it2= 2*nbSlices-1+(i+1)*(nbSlices+1)+1;
    triangle.setTextureVertexIndices(it1, it2, it3);
    addTriangle(triangle);
  }

  // triangle fan on the bottom:
  for(i=0; i<nbSlices-1; ++i)
  {
    i1= vertices.size()-1;    i2= vertices.size()-1-nbSlices+i;    i3= i2+1;
    triangle.setVertexIndices(i1, i3, i2);
    triangle.setNormalIndices(i1, i3, i2);

    it1= tvertices.size()-1-(nbSlices-1)+i;
    it2= nbSlices+ (nbSlices-2)*(nbSlices+1)+i;
    it3= it2+1;
    triangle.setTextureVertexIndices(it1, it3, it2);
    addTriangle(triangle);
  }
  i1= vertices.size()-1;    i2= vertices.size()-2;    i3= vertices.size()-1-nbSlices;
  triangle.setVertexIndices(i1, i3, i2);
  triangle.setNormalIndices(i1, i3, i2);

  it1= tvertices.size()-1;   it2= it1-nbSlices-1;    it3= it1-nbSlices;
  triangle.setTextureVertexIndices(it1, it3, it2);
  addTriangle(triangle);

  delete [] cosT;
  delete [] sinT; 
  delete [] cosTz;
  delete [] sinTz; 

  return 1;
}


//! Loads the mesh from a file in wavefront format (.obj file) and scales it with the given factor.
//! \param filename name of the .obj file to load
//! \param scale scale factor between the model described in the file and the created TriangleMesh
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::readOBJfile(std::string filename, float scale)
{
  unsigned int i, length, n1, n2, n3;
  float x, y, z;
  int line_count, countV, countVN, countVT, countF;
  char buffer[512];
  bool result;
  size_t pos;
  Triangle triangle;
  std::ifstream in;  
  std::string line, element, strings[3], separator;
  std::istringstream iss;

  reset();

  length= filename.length(); 
  pos= filename.find(".obj"); 
  if(pos==filename.npos || pos!=length-4)
  {
    printf("%s: %d: TriangleMesh::readOBJfile(): file %s does not have the good extension (it should be a .obj file).\n", __FILE__, __LINE__, filename.c_str());
    return 0;
  }    


  in.open(filename.c_str());  
  if( in.fail() ) 
  {
    printf("%s: %d: TriangleMesh::readOBJfile(): file %s was not found.\n", __FILE__, __LINE__, filename.c_str());
    in.close();
    return 0;
  }


  //We perform a first reading to count all the elements in order to allocate the right memory size
  //and check the consistency of vertex indices in the face descriptions:
  line_count= countV= countVN= countVT= countF= 0;
  while(in.eof()==false)
  {
     in.getline(buffer, 300);
     line_count++;
     line= buffer;
 
     //Remove all that's after a '#' character
     pos= line.find_first_of('#');
     if(pos!=line.npos)
     { 
       if(pos==0) continue;
       line= line.substr(0, pos);
     }

     //Remove all the spaces at the end of the line:
     pos= line.find_last_not_of(' ');
     line= line.substr(0, pos+1);

     if( line.empty() )
       continue;

     //Create a stream from the line string:
     iss.clear();
     iss.str(line);

     result= ( iss >> element );

     if(element=="v") //vertex
     { 
        countV++;
        continue;
     } 
     if(element=="vn") //vertex normal
     { 
        countVN++;
        continue;
     } 
     if(element=="vt") //vertex texture coordinates
     { 
        countVT++;
        continue;
     } 
     if(element=="f") //face
     { 
        countF++;
        continue;
     } 
  }


  in.clear(); //Set eof() to false
  in.seekg(0,std::ios::beg);

  vertices.reserve(countV);  
  normals.reserve(countVN);  
  tvertices.reserve(countVT);  
  triangles.reserve(countF);  


  line_count= countV= countVN= countVT= countF= 0;
  while(in.good())
  {
     in.getline(buffer, 300);
     line_count++;
     line= buffer;
 
     //Remove all that's after a '#' character
     pos= line.find_first_of('#');
     if(pos!=line.npos)
     {  line= line.substr(0, pos);  }

     //Remove all the spaces at the end of the line:
     pos= line.find_last_not_of(' ');
     line= line.substr(0, pos+1);

     //Remove the newline character for UNIX file (\n) at the end of the line:
     pos= line.find_first_of('\n');
     if(pos!=line.npos)
     {  line= line.substr(0, pos);  }

     //Remove the newline character for WINDOWS/DOS file (\r\n) at the end of the line:
     pos= line.find_first_of('\r');
     if(pos!=line.npos)
     {  line= line.substr(0, pos);  }


     if( line.empty() )
     {  continue;  }

     //Create a stream from the line string:
     iss.clear();
     iss.str(line);

     //Read the type of the element described at the given line:
     iss >> element;

     if(element=="v") //vertex
     {
       result= ( iss >> x >> y >> z );
       if( !result || !iss.eof() )
       {
         printf("TriangleMesh::readOBJfile(): syntax error in file %s at line %d:\n %s\n", filename.c_str(), line_count, line.c_str());
         printf("expected format is: v x y z (char float float float).\n");
       } 
       addVertex(scale*x, scale*y, scale*z);
       countV++;
     } 


     if(element=="vn") //vertex normal
     {
       result= ( iss >> x >> y >> z );
       if( !result || !iss.eof() )
       {
         printf("TriangleMesh::readOBJfile(): syntax error in file %s at line %d:\n %s\n", filename.c_str(), line_count, line.c_str());
         printf("expected format is: vn x y z (char* float float float).\n");
       }
       addNormal(x, y, z);
       countVN++;
     } 

     if(element=="vt") //texture coordinates
     {
       result= ( iss >> x >> y );
       if( !result || !iss.eof() )
       {
         printf("TriangleMesh::readOBJfile(): syntax error in file %s at line %d:\n %s\n", filename.c_str(), line_count, line.c_str());
         printf("expected format is: vt u v (char* float float).\n");
       } 
       addTextureVertex(x, y);
       countVT++;
     } 


     if(element=="f") //face
     {
       strings[0].clear();
       strings[1].clear();
       strings[2].clear();
       result= ( iss >> strings[0] >> strings[1] >> strings[2] );

       if( !result || !iss.eof() )
       {
         printf("TriangleMesh::readOBJfile(): syntax error in file %s at line %d:\n %s\n", filename.c_str(), line_count, line.c_str());
         printf("expected format is: \n\t f vertex1 vertex2 vertex3 (char int int int)\n");
         printf("   or: \n\t f vertex1//normal1 vertex2//normal2 vertex3//normal3 (char int//int int//int int//int)\n");
         printf("   or: \n\t f vertex1/normal1/texture_vertex1 vertex2/normal2/texture_vertex2 vertex3/normal3/texture_vertex3 (char int/int/int int/int/int int/int/int)\n");
         printf("Non triangular faces are not supported.\n");
         continue;
       } 

       for(i=0; i<3; ++i)
       {
          //remove all the possible ending '/':
          while(strings[i].at(strings[i].size()-1)=='/')
          {
            strings[i].resize(strings[i].size()-1);
          }

          separator.clear();
          triangle.v[i]= triangle.vn[i]= triangle.tv[i]= -1;

          pos= strings[i].find("//");
          if(pos!=std::string::npos)
          {  separator= "//";  }
          else
          {
             pos= strings[i].find("/");
             if(pos!=std::string::npos)
             {  separator= "/";  }
             else
             {  separator.clear(); }
          }

          pos= strings[i].find(separator);

          if(!separator.empty())
          {         
            while(pos!=std::string::npos)
            {
              strings[i].replace(pos, separator.length(), " ");
              pos= strings[i].find(separator);
            }
          }

          iss.clear();
          iss.str(strings[i]);

          if(separator=="/")
          {
             result= (iss >> n1 >> n2 >> n3); 
             if(!result || !iss.eof())
             {
               printf("TriangleMesh::readOBJfile(): syntax error in file %s at line %d:\n %s.\n", filename.c_str(), line_count, line.c_str());
               printf("expected format was int/int/int for vertex %d.\n", i+1);
               continue; 
             } 
             //if the face uses negative reference number:
             if(n1<0) n1= countV  + n1; 
             if(n2<0) n2= countVT + n2;
             if(n3<0) n3= countVN + n3;

             //we use indices starting from zero:
             n1--;
             n2--;
             n3--;

             triangle.v[i] = n1;
             triangle.tv[i]= n2;
             triangle.vn[i]= n3;    
           }

           if(separator=="//")
           {
             result= (iss >> n1 >> n2); 
             if(!result || !iss.eof())
             {
               printf("TriangleMesh::readOBJfile(): syntax error in file %s at line %d:\n %s.\n", filename.c_str(), line_count, line.c_str());
              printf("expected format was int//int for vertex %d.\n", i+1);
              continue;
             } 
             //if the face uses negative reference number:
             if(n1<0) n1= countV  + n1; 
             if(n2<0) n2= countVN + n2;

             //we use indices starting from zero:
             n1--;
             n2--;

             triangle.v[i] = n1;
             triangle.tv[i]= -1;  
             triangle.vn[i]= n2;
           }

           if(separator.empty())
           {
             result= (iss >> n1); 
             if(!result || !iss.eof())
             {
               printf("TriangleMesh::readOBJfile(): syntax error in file %s at line %d:\n %s.\n", filename.c_str(), line_count, line.c_str());
              printf("expected format was int for vertex %d.\n", i+1);
              continue;
             } 
             //if the face uses negative reference number:
             if(n1<0) n1= countV + n1; 

             //we use indices starting from zero:
             n1--;

             triangle.v[i] = n1;
             triangle.tv[i]= -1;
             triangle.vn[i]= -1;    
           }

           if( triangle.v[i]<0 || triangle.v[i]>((unsigned int) vertices.size()-1) )
           { 
             printf("TriangleMesh::readOBJfile(): wrong %d-th vertex index %d in file %s at line %d:\n %s.\n", i,triangle.v[i], filename.c_str(), line_count, line.c_str());
             triangle.v[i]= 0;
           }

           if( triangle.vn[i]<-1 || triangle.vn[i]>((int) normals.size()-1) )
           {  
             printf("TriangleMesh::readOBJfile(): wrong %d-th vertex index %d in file %s at line %d:\n %s.\n", i,triangle.v[i], filename.c_str(), line_count, line.c_str());
             triangle.vn[i]= -1;
           }

           if( triangle.tv[i]<-1 || triangle.tv[i]>((int) tvertices.size()-1) )
           {  
             printf("TriangleMesh::readOBJfile(): wrong %d-th vertex index %d in file %s at line %d:\n %s.\n", i,triangle.v[i], filename.c_str(), line_count, line.c_str());
             triangle.tv[i]= -1;
           }


	     } //end of 'for' loop
       addTriangle(triangle);

       } //end of if( (line[0]=='f')&&(line[1]==' ') )

     } //end of file reading 'while' loop

  in.close();

  return 1;
}


//! Saves the mesh in a file according to the wavefront format (.obj file) after scaling it with the given factor.
//! \param filename name of the file to save
//! \param scale scale factor between the TriangleMesh and the model that will be written in the file
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::saveAsOBJfile(std::string filename, float scale)
{
  unsigned int i;
  std::ofstream out;

  out.open(filename.c_str());

  if( out.fail() ) 
  {
    printf("TriangleMesh::saveAsOBJfile: can not write file \"%s\".\n", filename.c_str());
    out.close();
    return 0;
  }

  out << "# exported from COLDMAN\n";

  out << "# " << vertices.size() << " vertices, "  << triangles.size() << " triangles\n";
  out << "# " << normals.size() << " vertex normals, " << tvertices.size() << " texture vertices\n\n"; 

  for(i=0; i<vertices.size(); ++i)
  {
    out << "v " << scale*vertices[i].x << " " << scale*vertices[i].y << " " << scale*vertices[i].z << std::endl;
  }

  for(i=0; i<normals.size(); ++i)
  {
    out << "vn " << normals[i].x << " " << normals[i].y << " " << normals[i].z << std::endl;
  }

  for(i=0; i<tvertices.size(); ++i)
  {
    out << "vt " << tvertices[i].x << " " << tvertices[i].y << std::endl;
  }

  for(i=0; i<triangles.size(); ++i)
  {
    if( triangles[i].vn[0]!=-1 && triangles[i].vn[1]!=-1 && triangles[i].vn[2]!=-1 )
    {
      if( triangles[i].tv[0]!=-1 && triangles[i].tv[1]!=-1 && triangles[i].tv[2]!=-1 )
      {
        out << "f " << triangles[i].v[0]+1 << "/" << triangles[i].tv[0]+1 << "/" << triangles[i].vn[0]+1;
        out << " "  << triangles[i].v[1]+1 << "/" << triangles[i].tv[1]+1 << "/" << triangles[i].vn[1]+1;
        out << " "  << triangles[i].v[2]+1 << "/" << triangles[i].tv[2]+1 << "/" << triangles[i].vn[2]+1 << std::endl;
      }
      else
      {
        out << "f " << triangles[i].v[0]+1 << "//" << triangles[i].vn[0]+1;
        out << " "  << triangles[i].v[1]+1 << "//" << triangles[i].vn[1]+1;
        out << " "  << triangles[i].v[2]+1 << "//" << triangles[i].vn[2]+1 << std::endl;
      }
    }
    else
    {
      out << "f " << triangles[i].v[0]+1;
      out << " "  << triangles[i].v[1]+1;
      out << " "  << triangles[i].v[2]+1 << std::endl;
    }
  }

  out.close();

  return 1;
}


//! Loads the mesh from a file in Move3d format (.macro file)) and scales it with the given factor.
//! \param filename name of the .obj file to load
//! \param scale scale factor between the model described in the file and the created TriangleMesh
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::readP3Dfile(std::string filename, float scale)
{
  unsigned int i, length;
  int n1, n2, n3;
  float x, y, z;  
  int line_count, countV, countF, shift= 0;
  char buffer[512];
  bool result, triangle_is_good;
  size_t pos;
  Triangle triangle;
  std::ifstream in;  
  std::string line, command;
  std::istringstream iss;

  vertices.clear();
  normals.clear();
  tvertices.clear();
  triangles.clear();

  length= filename.length(); 
  pos= filename.find(".macro"); 
  if(pos==filename.npos || pos!=length-6)
  {
    printf("File %s does not have the good extension (it should be a .macro file).\n", filename.c_str());
    return 0;
  }    

  in.open(filename.c_str());  
  if(in.fail()) 
  {
    printf("File %s was not found.\n", filename.c_str());
    in.close();
    return 0;
  }


  //We perform a first reading to count all the elements in order to allocate the right memory size
  //and check the consistency of vertex indices in the face descriptions:
  line_count= countV= countF= 0;
  while(in.eof()==false)
  {
     in.getline(buffer, 300);
     line_count++;
     line= buffer;
 
     //Remove all that's after a '#' character
     pos= line.find_first_of('#');
     if(pos!=line.npos)
     { 
       if(pos==0) continue;
       line= line.substr(0, pos);
     }

     //Remove all the spaces at the end of the line:
     pos= line.find_last_not_of(' ');
     line= line.substr(0, pos+1);

     if( line.empty() )
       continue;

     //Create a stream from the line string:
     iss.clear();
     iss.str(line);

     iss >> command;

     if(command=="p3d_add_desc_poly")
     {
        countV++;
        continue;
     }

     if(command=="p3d_add_desc_face")
     {
        countF++;
        continue;
     }
  }



  in.clear(); //Set eof() to false
  in.seekg(0,std::ios::beg);

  vertices.reserve(countV);   
  triangles.reserve(countF);  



  line_count= countV= countF= 0;
  while(in.eof()==false)
  {
     in.getline(buffer, 300);
     line_count++;
     line= buffer;
 
     //Remove all that's after a '#' character
     pos= line.find_first_of('#');
     if(pos!=line.npos)
     { 
       if(pos==0) continue;
       line= line.substr(0, pos);
     }

     //Remove all the spaces at the end of the line:
     pos= line.find_last_not_of(' ');
     line= line.substr(0, pos+1);

     if( line.empty() )
       continue;

     //Create a stream from the line string:
     iss.clear();
     iss.str(line);

     iss >> command;

     if(command=="p3d_add_desc_poly")
     {
        shift= countV;
        continue;
     }

     if(command=="p3d_add_desc_vert")
     {
        result= ( iss >> x >> y >> z );
        if( !result || !iss.eof() )
        {
          printf("TriangleMesh::readP3Dfile(): syntax error in file \"%s\" at line %d:\n %s\n", filename.c_str(), line_count, line.c_str());
         printf("expected format is: p3d_add_desc_vert x y z (char* float float float).\n");
        } 
//         p3*= scale;
//         vertices.push_back(p3);
        addVertex(scale*x, scale*y, scale*z);
//         countV++;
        continue;
     }

     if(command=="p3d_add_desc_face")
     {
        result= ( iss >> n1 >> n2 >> n3 );
        if( !result || !iss.eof() )
        {
          printf("TriangleMesh::readP3Dfile(): syntax error in file \"%s\" at line %d:\n %s\n", filename.c_str(), line_count, line.c_str());
         printf("expected format is: p3d_add_desc_face index1 index2 index3 (char* int int int).\n");
         printf("Non triangular faces are not supported.\n");
        } 
     
        //we use indices starting from zero:
        n1--;
        n2--;
        n3--;

        triangle.v[0]= n1 + shift;
        triangle.v[1]= n2 + shift;
        triangle.v[2]= n3 + shift;

        triangle_is_good= true;
        for(i=0; i<3; ++i)
        {
           if( triangle.v[i]<0 || triangle.v[i]>(vertices.size()-1) )
           { 
             printf("TriangleMesh::reaP3Dfile(): wrong %d-th vertex index %d in file \"%s\" at line %d:\n %s.\n", i, triangle.v[i], filename.c_str(), line_count, line.c_str());
             triangle_is_good= false;
           }
        }

        if(triangle_is_good)
        {
//           triangles.push_back(triangle);
          addTriangle(triangle);
        }

        continue;

     } //end of 'if(command=="p3d_add_desc_face")'

  }

  //Compute the triangle normals:
//   computeTriangleNormals();


  return 1;
}

//! Resets the content of the mesh.
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::reset()
{
  vertices.clear();
  normals.clear();
  tvertices.clear();
  triangles.clear();

  return 1;
}

//! Draws a TriangleMesh.
//! \param smooth use smooth (true) or flat (false) shading
void TriangleMesh::draw(bool smooth)
{
#if defined(ENABLE_DISPLAY)
 // GLboolean lighting_enable, texture_enable, polygon_offset_enable;
  unsigned int i, j;
  //  float s= 0.05;

//  glGetBooleanv(GL_TEXTURE_2D, &lighting_enable);
//  glGetBooleanv(GL_TEXTURE_2D, &texture_enable);
//  glGetBooleanv(GL_POLYGON_OFFSET_FILL, &polygon_offset_enable);  

  if(tvertices.empty())
  {  glDisable(GL_TEXTURE_2D);  }
//   else
//   {  loadTexture(texture); }

  glPushAttrib(GL_ENABLE_BIT);

  glEnable(GL_LIGHTING);
  glEnable(GL_POLYGON_OFFSET_FILL);
  glPolygonOffset(1.2, 1.2);
  glBegin(GL_TRIANGLES);
   for(i=0; i<triangles.size(); ++i)
   {
     for(j=0; j<3; j++)
     {
        if(triangles[i].vn[j]>-1 && smooth==true)
        { glNormal3f(normals[triangles[i].vn[j]].x, normals[triangles[i].vn[j]].y, normals[triangles[i].vn[j]].z); }
        else
        { glNormal3f(triangles[i].normal.x, triangles[i].normal.y, triangles[i].normal.z); } 
        if(triangles[i].tv[j]>-1)
        { glTexCoord2f(tvertices[triangles[i].tv[j]].x, tvertices[triangles[i].tv[j]].y); }

        glVertex3f(vertices[triangles[i].v[j]].x, vertices[triangles[i].v[j]].y, vertices[triangles[i].v[j]].z);
     }
   }
  glEnd();

/*
  glColor3f(0.0, 0.0, 0.0);
  glDisable(GL_LIGHTING);
  glBegin(GL_LINES);
   for(i=0; i<triangles.size(); ++i)
   {
     glVertex3f(vertices[triangles[i].v[0]].x, vertices[triangles[i].v[0]].y, vertices[triangles[i].v[0]].z);
     glVertex3f(vertices[triangles[i].v[1]].x, vertices[triangles[i].v[1]].y, vertices[triangles[i].v[1]].z);

     glVertex3f(vertices[triangles[i].v[1]].x, vertices[triangles[i].v[1]].y, vertices[triangles[i].v[1]].z);
     glVertex3f(vertices[triangles[i].v[2]].x, vertices[triangles[i].v[2]].y, vertices[triangles[i].v[2]].z);

     glVertex3f(vertices[triangles[i].v[0]].x, vertices[triangles[i].v[0]].y, vertices[triangles[i].v[0]].z);
     glVertex3f(vertices[triangles[i].v[2]].x, vertices[triangles[i].v[2]].y, vertices[triangles[i].v[2]].z);
   }
  glEnd();
*/

  // display the mesh vertices as points:
//   glPointSize(3);
//   glDisable(GL_LIGHTING);
//   glColor3f(1, 0, 0);
//   glBegin(GL_POINTS);
//    for(i=0; i<vertices.size(); ++i)
//    {
//      glVertex3f(vertices[i].x, vertices[i].y, vertices[i].z);
//    }
//   glEnd();

  //display the vertex normals, if they exist:
//   glBegin(GL_LINES);
//    for(i=0; i<triangles.size(); ++i)
//    {
//      for(j=0; j<3; j++)
//      {
//         if(triangles[i].vn[j]>-1)
//         {
//           glVertex3f(vertices[triangles[i].v[j]].x, vertices[triangles[i].v[j]].y, vertices[triangles[i].v[j]].z);
//           glVertex3f(vertices[triangles[i].v[j]].x + s*normals[triangles[i].vn[j]].x, vertices[triangles[i].v[j]].y + s*normals[triangles[i].vn[j]].y, vertices[triangles[i].v[j]].z + s*normals[triangles[i].vn[j]].z);
//         }
//      }
//    }
//   glEnd();

/*
  if(lighting_enable)
  {  glEnable(GL_LIGHTING);  }
  else
  {  glDisable(GL_LIGHTING);  }

  if(texture_enable)
  {  glEnable(GL_TEXTURE_2D);  }
  else
  {  glDisable(GL_TEXTURE_2D);  }

  if(polygon_offset_enable)
  { glEnable(GL_POLYGON_OFFSET_LINE);  }
  else
  { glDisable(GL_POLYGON_OFFSET_LINE); }  */

  glPopAttrib();

  glDisable(GL_TEXTURE_GEN_R);
  glDisable(GL_TEXTURE_GEN_S);
  glDisable(GL_TEXTURE_GEN_T);
#endif
}


//! Computes the normal vector of each triangle.
//! NB: the triangle normals are already computed when the triangles
//! are added withe the addTriangle() function.
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::computeTriangleNormals()
{
  unsigned int i;
  Vector3D e1, e2;

  for(i=0; i<triangles.size(); ++i)
  {
     e1= vertices[triangles[i].v[1]] - vertices[triangles[i].v[0]];
     e1.normalize();
     e2= vertices[triangles[i].v[2]] - vertices[triangles[i].v[0]];
     e2.normalize();
     triangles[i].normal=  crossProduct(e1, e2);
     triangles[i].normal.normalize();
  }

  return 1;
}

//! Computes the normal vector of each vertex.
//! The computed normal is a weighted sum of the normals of all the triangles the vertex belongs to.
//! For each normal, the weight is the angle between the two edges sharing the vertex.
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::computeVertexNormals()
{
  unsigned int i, j, vertex_index;
  bool triangle_found= false;
  float vertex_angle;
  Vector3D e1, e2;

  if(!normals.empty())
  {
    return 1;
  }
  
  normals.resize(vertices.size());

  for(i=0; i<normals.size(); ++i)
  { triangle_found= false;
     normals[i].set(0.0, 0.0, 0.0);
     for(j=0; j<triangles.size(); j++)
     {
       if(triangles[j].v[0]==i)
       {
         e1= vertices[triangles[j].v[1]] - vertices[triangles[j].v[0]];
         e2= vertices[triangles[j].v[2]] - vertices[triangles[j].v[0]];
         vertex_index= 0; triangle_found= true;
       }
       else
       {
          if(triangles[j].v[1]==i)
          {
            e1= vertices[triangles[j].v[0]] - vertices[triangles[j].v[1]];
            e2= vertices[triangles[j].v[2]] - vertices[triangles[j].v[1]];
            vertex_index= 1; triangle_found= true;
          }
          else     
          {        
            if(triangles[j].v[2]==i)
            {
              e1= vertices[triangles[j].v[0]] - vertices[triangles[j].v[2]];
              e2= vertices[triangles[j].v[1]] - vertices[triangles[j].v[2]];
              vertex_index= 2; triangle_found= true;
            }
            else
            {
              continue;
            }
          }
       }

       e1.normalize();
       e2.normalize();

       vertex_angle= fabs( acos( dotProduct(e1, e2) ) );
       if( isnan(vertex_angle) )
       {  vertex_angle= 0.0;  }

       normals[i]+= vertex_angle*(triangles[j].normal);
       triangles[j].vn[vertex_index]= i;
     }

     if(normals[i].normalize()==0)
     {
        normals[i].set(1.0, 0.0, 0.0);
     }
  }

  return 1;
}

//! Inverts all the triangle normals.
//! The order in the vertex index array of each triangle is modified as well as its field "normal".
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::invertTriangleNormals()
{
   unsigned int i, tmp;

   for(i=0; i<triangles.size(); ++i)
   {
     tmp= triangles[i].v[2];
     triangles[i].v[2]= triangles[i].v[1];
     triangles[i].v[1]= tmp;
    
     triangles[i].normal= -triangles[i].normal;
   }
  
   return 1;
}


//! WIP
Vector3D TriangleMesh::closestPoint(Vector3D p)
{
  unsigned int i;
  float d, dmin= 1e17;
  Vector3D tri[3], closestPoint, result;

  for(i=0; i<triangles.size(); ++i)
  {
    getTriangleVertices(i, tri[0], tri[1], tri[2]);
    d= pointToTriangleDistance(p, tri[0], tri[1], tri[2], closestPoint);

    if(i==0)
    {
      dmin= d;
      result= closestPoint;
    }
    else
    {
      if(d < dmin)
      {
        dmin= d;
        result= closestPoint;
      }
    }
  }

  return result;

}

//! Tests wether or not a point is included in the mesh.
//! The algorithm first finds the closest triangle to the point and
//! then tests on which side of the triangle (half-plane) the point is.
//! If it is not in the half-plane where points the triangle's normal, the point
//! is inside the mesh volume.
//! This algorithm should give good results even if the mesh has holes but the triangles
//! must be correctly oriented.
//! \param point the point to test
//! \return true is the point is in the mesh's interior, false otherwise
bool TriangleMesh::pointInclusionTest(Vector3D point)
{
  unsigned int i;
  bool inside;
  float d, dmin= 1e17;
  Vector3D p[3], closestPoint, diff;

  inside= false;
  for(i=0; i<triangles.size(); ++i)
  {
    getTriangleVertices(i, p[0], p[1], p[2]);
    d= pointToTriangleDistance(point, p[0], p[1], p[2], closestPoint);
    diff= closestPoint - point;
    diff.normalize();

    if(i==0)
    { dmin= d; }

    if(d < dmin || i==0)
    {
      dmin= d;
      if(dotProduct(triangles[i].normal, diff) > 0)
      {  inside= true; }
      else
      {  inside= false; }
    }
  }

  return inside;
}

//! Scales the mesh so that all its vertices fit in a sphere with given radius.
int TriangleMesh::fitInSphere(float radius)
{
  unsigned int i;
  float d, dmax, scale;


  dmax= vertices[0].norm();

  for(i=1; i<vertices.size(); ++i)
  {
    d= vertices[i].norm();

    if(d > dmax)
    {  dmax= d; }
  }

  scale= radius/dmax;

  for(i=0; i<vertices.size(); ++i)
  {
    vertices[i]*= scale;
  }

  return 1;
}

//! AABB root constructor
AABB::AABB(Vector3D center, float dimX, float dimY, float dimZ, float minSize, TriangleMesh *mesh, std::list<unsigned int> inner_triangles)
{
  if(mesh==NULL)
  {  
    printf("%s: %d: AABB::AABB(Vector3D, float, float, float, TriangleMesh*, std::list<unsigned int>): mesh is NULL.\n",__FILE__,__LINE__);   
    return;
  }

  _mesh= mesh;
  _root= true;
  _leaf= true;
  _terminate= false;
  _level= 1;

  _center= center;
  _dimX  = dimX;
  _dimY  = dimY;
  _dimZ  = dimZ;
  _minimum_size= minSize;

  _children[0]= NULL;
  _children[1]= NULL;
  _brother= NULL;

  _inner_triangles= inner_triangles;
}


//! AABB root constructor
AABB::AABB(float xmin, float xmax, float ymin, float ymax, float zmin, float zmax, float minSize, TriangleMesh *mesh, std::list<unsigned int> inner_triangles)
{
  if(mesh==NULL)
  {  
    printf("%s: %d: AABB::AABB(float, float, float, float, float, float, TriangleMesh*, std::list<unsigned int>): mesh is NULL.\n",__FILE__,__LINE__);   
    return;
  }

  _state= NEITHER;
  _mesh= mesh;
  _root= true;
  _leaf= true;
  _terminate= false;
  _level= 1;

  _xmin= xmin;
  _xmax= xmax;
  _ymin= ymin;
  _ymax= ymax;
  _zmin= zmin;
  _zmax= zmax;
  _minimum_size= minSize;

  _children[0]= NULL;
  _children[1]= NULL;
  _brother= NULL;

  _inner_triangles= inner_triangles;
}


//! Constructor for non root AABB.
AABB::AABB(AABB *previous, Vector3D center, float dimX, float dimY, float dimZ)
{
  if(previous==NULL)
  {  
    printf("%s: %d: AABB::AABB(AABB*, Vector3D, float, float, float): previous is NULL.\n",__FILE__,__LINE__);   
    return;
  }
  
  if(previous->_mesh==NULL)
  {  
    printf("%s: %d: AABB::AABB(AABB*, Vector3D, float, float, float): the \"_mesh\" field of previous is NULL.\n",__FILE__,__LINE__);  
    return;
  }  
  
  previous->_leaf= false;
  _root= false;
  _leaf= true;
  _terminate= false;
  _level= previous->_level + 1;
  
  _mesh= previous->_mesh;

  _center= center;
  _dimX  = dimX;
  _dimY  = dimY;
  _dimZ  = dimZ;
  _minimum_size= previous->_minimum_size;

  _children[0]= NULL;
  _children[1]= NULL;

  _inner_triangles= previous->_inner_triangles;

}


//! Constructor for non root AABB.
AABB::AABB(AABB *previous, float xmin, float xmax, float ymin, float ymax, float zmin, float zmax)
{
  if(previous==NULL)
  {  
    printf("%s: %d: AABB::AABB(AABB*, float, float, float, float, float, float): previous is NULL.\n",__FILE__,__LINE__);   
    return;
  }
  
  if(previous->_mesh==NULL)
  {  
    printf("%s: %d: AABB::AABB(AABB*, float, float, float, float, float, float): the \"_mesh\" field of previous is NULL.\n",__FILE__,__LINE__);  
    return;
  }  
  
  _state= previous->_state;
  previous->_leaf= false;
  _root= false;
  _leaf= true;
  _terminate= false;
  _level= previous->_level + 1;
  
  _mesh= previous->_mesh;

  _xmin= xmin;
  _xmax= xmax;
  _ymin= ymin;
  _ymax= ymax;
  _zmin= zmin;
  _zmax= zmax;
  _minimum_size= previous->_minimum_size;

  _children[0]= NULL;
  _children[1]= NULL;

  _inner_triangles= previous->_inner_triangles;
}

int AABB::testTriangles()
{
  if(_mesh==NULL)
  {
    printf("%sdi: %d: AABB::testTriangles(): the \"_mesh\" field of the calling AABB is NULL.\n",__FILE__,__LINE__);   
    return 0;
  }
  unsigned int nbInnerTriangles;
  float x, y, z, xmin, xmax, ymin, ymax, zmin, zmax;
  Vector3D p1, p2, p3, new_center;
  std::list<unsigned int>::iterator triangle_index;

  xmin= _center.x - 0.5*_dimX;
  xmax= _center.x + 0.5*_dimX;
  ymin= _center.y - 0.5*_dimY;
  ymax= _center.y + 0.5*_dimY;
  zmin= _center.z - 0.5*_dimZ;
  zmax= _center.z + 0.5*_dimZ;


  nbInnerTriangles= _inner_triangles.size();
  triangle_index= _inner_triangles.begin();
  while( triangle_index != _inner_triangles.end() && !_inner_triangles.empty() )
  {
    _mesh->getTriangleVertices(*triangle_index, p1, p2, p3);
    if( (p1.x < _xmin) && (p2.x < _xmin) && (p3.x < _xmin) )
    {
      triangle_index= _inner_triangles.erase(triangle_index);
      continue;
    }
    if( (p1.x > _xmax) && (p2.x > _xmax) && (p3.x > _xmax) )
    {
      triangle_index= _inner_triangles.erase(triangle_index);
      continue;
    }

    if( (p1.y < _ymin) && (p2.y < _ymin) && (p3.y < _ymin) )
    {
      triangle_index= _inner_triangles.erase(triangle_index);
      continue;
    }
    if( (p1.y > _ymax) && (p2.y > _ymax) && (p3.y > _ymax) )
    {
      triangle_index= _inner_triangles.erase(triangle_index);
      continue;
    }

    if( (p1.z < _zmin) && (p2.z < _zmin) && (p3.z < _zmin) )
    {
      triangle_index= _inner_triangles.erase(triangle_index);
      continue;
    }
    if( (p1.z  > _zmax) && (p2.z > _zmax) && (p3.z > _zmax) )
    {
      triangle_index= _inner_triangles.erase(triangle_index);
      continue;
    }
    triangle_index++;
  }
  
  if(nbInnerTriangles==_inner_triangles.size())
  {
    _terminate= true;
    return 1;
  }

  if(!_inner_triangles.empty())
  {//printf("inner tri size= %d\n", _inner_triangles.size());
      _mesh->getTriangleVertices(_inner_triangles.front(), p1, p2, p3);
      xmin= min( p1.x, min(p2.x, p3.x)); 
      xmax= max( p1.x, max(p2.x, p3.x)); 
      ymin= min( p1.y, min(p2.y, p3.y)); 
      ymax= max( p1.y, max(p2.y, p3.y));
      zmin= min( p1.z, min(p2.z, p3.z)); 
      zmax= max( p1.z, max(p2.z, p3.z));
  
      triangle_index= _inner_triangles.begin();
      triangle_index++;
      while( triangle_index != _inner_triangles.end() )
      {
        _mesh->getTriangleVertices(*triangle_index, p1, p2, p3);

        x= min( p1.x, min(p2.x, p3.x)); 
        if(x < xmin)
        {  xmin= x;  }
        x= max( p1.x, max(p2.x, p3.x)); 
        if(x > xmax)
        {  xmax= x;  }  
        
        y= min( p1.y, min(p2.y, p3.y)); 
        if(y < ymin)
        {  ymin= y;  }
        y= max( p1.y, max(p2.y, p3.y)); 
        if(y > ymax)
        {  ymax= y;  }  
        
        z= min( p1.z, min(p2.z, p3.z)); 
        if(z < zmin)
        {  zmin= z;  }
        z= max( p1.z, max(p2.z, p3.z)); 
        if(z > zmax)
        {  zmax= z;  }      
        
        triangle_index++;
      }    
      
    // _center.print("center");
    // printf("dim= %f %f %f\n", _dimX, _dimY, _dimZ);

/*
      new_center= _center;
    
      if(xmax - xmin < _dimX)
      {
        new_center.x= 0.5*(xmin+xmax);
        _dimX= xmax - xmin;
      }
  
      if(ymax - ymin < _dimY)
      {
        new_center.y= 0.5*(ymin+ymax);
        _dimY= ymax - ymin;
      }
  
      if(zmax - zmin < _dimZ)
      {
        new_center.z= 0.5*(zmin+zmax);
        _dimZ= zmax - zmin;
      }
  
      _center= new_center;
*/
      if(_xmin==xmin && _xmax==xmax && _ymin==ymin && _ymax==ymax && _zmin==zmin && _zmax==zmax)
      {
        _terminate= true;
        return 1;
      }


      _xmin= xmin;
      _xmax= xmax;
      _ymin= ymin;
      _ymax= ymax;
      _zmin= zmin;
      _zmax= zmax;    
  /*
      _center.set( 0.5*(xmin+xmax), 0.5*(ymin+ymax), 0.5*(zmin+zmax) );
  
      if( (_dimX <= xmax - xmin) && (_dimY <= ymax - ymin) && (_dimZ <= zmax - zmin) )
      {
        _terminate= true;
        return 1;
      }*/
  /*
      _dimX= xmax - xmin;
      _dimY= ymax - ymin;    
      _dimZ= zmax - zmin;   */
  
      //_center.print("---center");
      //printf("---dim= %f %f %f\n", _dimX, _dimY, _dimZ); 
                      
  }
  else
  {
    _terminate= false;
  }
  
  
  
//printf("nb tris= %d\n", _inner_triangles.size());
  return 1;
}


int AABB::testTrianglesSmall(int axis)
{
  if(_mesh==NULL)
  {
    printf("%s: %d: AABB::testTriangles(): the \"_mesh\" field of the calling AABB is NULL.\n",__FILE__,__LINE__);   
    return 0;
  }
  
  float x, y, z, xmin, xmax, ymin, ymax, zmin, zmax;
  Vector3D p1, p2, p3, new_center;
  std::list<unsigned int>::iterator triangle_index;

  triangle_index= _inner_triangles.begin();
  while( triangle_index != _inner_triangles.end() && !_inner_triangles.empty() )
  {
    _mesh->getTriangleVertices(*triangle_index, p1, p2, p3);
    xmin= _center.x - 0.5*_dimX;
    xmax= _center.x + 0.5*_dimX;
    ymin= _center.y - 0.5*_dimY;
    ymax= _center.y + 0.5*_dimY;
    zmin= _center.z - 0.5*_dimZ;
    zmax= _center.z + 0.5*_dimZ;


    if(axis==1)
    {
      if( (p1.x < xmin) || (p2.x < xmin) || (p3.x < xmin) )
      {
        triangle_index= _inner_triangles.erase(triangle_index);
        continue;
      }
      if( (p1.x > xmax) || (p2.x > xmax) || (p3.x > xmax) )
      {
        triangle_index= _inner_triangles.erase(triangle_index);
        continue;
      }
    }
    else
    {
      if( (p1.x < xmin) && (p2.x < xmin) && (p3.x < xmin) )
      {
        triangle_index= _inner_triangles.erase(triangle_index);
        continue;
      }
      if( (p1.x > xmax) && (p2.x > xmax) && (p3.x > xmax) )
      {
        triangle_index= _inner_triangles.erase(triangle_index);
        continue;
      }
    }

    if(axis==2)
    {
      if( (p1.y < ymin) || (p2.y < ymin) || (p3.y < ymin) )
      {
        triangle_index= _inner_triangles.erase(triangle_index);
        continue;
      }
      if( (p1.y > ymax) || (p2.y > ymax) || (p3.y > ymax) )
      {
        triangle_index= _inner_triangles.erase(triangle_index);
        continue;
      }
    }
    else
    {
      if( (p1.y < ymin) && (p2.y < ymin) && (p3.y < ymin) )
      {
        triangle_index= _inner_triangles.erase(triangle_index);
        continue;
      }
      if( (p1.y > ymax) && (p2.y > ymax) && (p3.y > ymax) )
      {
        triangle_index= _inner_triangles.erase(triangle_index);
        continue;
      }
    }


    if(axis==3)
    {
      if( (p1.z < zmin) || (p2.z < zmin) || (p3.z < zmin) )
      {
        triangle_index= _inner_triangles.erase(triangle_index);
        continue;
      }
      if( (p1.z  > zmax) || (p2.z > zmax) || (p3.z > zmax) )
      {
        triangle_index= _inner_triangles.erase(triangle_index);
        continue;
      }
    }
    else
    {
      if( (p1.z < zmin) && (p2.z < zmin) && (p3.z < zmin) )
      {
        triangle_index= _inner_triangles.erase(triangle_index);
        continue;
      }
      if( (p1.z  > zmax) && (p2.z > zmax) && (p3.z > zmax) )
      {
        triangle_index= _inner_triangles.erase(triangle_index);
        continue;
      }
    }

    triangle_index++;
  }
  

  if(!_inner_triangles.empty())
  {//printf("inner tri size= %d\n", _inner_triangles.size());
      _mesh->getTriangleVertices(_inner_triangles.front(), p1, p2, p3);
      xmin= min( p1.x, min(p2.x, p3.x)); 
      xmax= max( p1.x, max(p2.x, p3.x)); 
      ymin= min( p1.y, min(p2.y, p3.y)); 
      ymax= max( p1.y, max(p2.y, p3.y));
      zmin= min( p1.z, min(p2.z, p3.z)); 
      zmax= max( p1.z, max(p2.z, p3.z));
  
      triangle_index= _inner_triangles.begin();
      triangle_index++;
      while( triangle_index != _inner_triangles.end() )
      {
        _mesh->getTriangleVertices(*triangle_index, p1, p2, p3);

        x= min( p1.x, min(p2.x, p3.x)); 
        if(x < xmin)
        {  xmin= x;  }
        x= max( p1.x, max(p2.x, p3.x)); 
        if(x > xmax)
        {  xmax= x;  }  
        
        y= min( p1.y, min(p2.y, p3.y)); 
        if(y < ymin)
        {  ymin= y;  }
        y= max( p1.y, max(p2.y, p3.y)); 
        if(y > ymax)
        {  ymax= y;  }  
        
        z= min( p1.z, min(p2.z, p3.z)); 
        if(z < zmin)
        {  zmin= z;  }
        z= max( p1.z, max(p2.z, p3.z)); 
        if(z > zmax)
        {  zmax= z;  }      
        
        triangle_index++;
      }    
      
    // _center.print("center");
    // printf("dim= %f %f %f\n", _dimX, _dimY, _dimZ);
      new_center= _center;
    
      if(xmax - xmin < _dimX)
      {
        new_center.x= 0.5*(xmin+xmax);
        _dimX= xmax - xmin;
      }
  
      if(ymax - ymin < _dimY)
      {
        new_center.y= 0.5*(ymin+ymax);
        _dimY= ymax - ymin;
      }
  
      if(zmax - zmin < _dimZ)
      {
        new_center.z= 0.5*(zmin+zmax);
        _dimZ= zmax - zmin;
      }
      _center= new_center;             
  }
  else
  {
    _terminate= false;
  }
  
  return 1;
}


bool AABB::isPointInside(Vector3D p)
{
  if( (_xmin < p.x) && (p.x < _xmax) &&
      (_ymin < p.y) && (p.y < _ymax) &&
      (_zmin < p.z) && (p.z < _zmax)  )
  {  return true;  }
  else
  {  return false;  }

}

int AABB::divide()
{//printf("aabb\n");
  Vector3D center;

  center= _center;

  if( (_dimX < _minimum_size) || (_dimY < _minimum_size) || (_dimZ < _minimum_size) )
  {
    _leaf= true;
    return 1;
  }

  if( (_dimX > _dimY) && (_dimX > _dimZ) )
  {
     center.x= _center.x + 0.25*_dimX;
     _children[0]= new AABB(this, center, 0.5*_dimX, _dimY, _dimZ);

     center.x= _center.x - 0.25*_dimX;
     _children[1]= new AABB(this, center, 0.5*_dimX, _dimY, _dimZ);
  }
  else  
  {
     if(_dimY > _dimZ) 
     {
        center.y= _center.y + 0.25*_dimY;
        _children[0]= new AABB(this, center, _dimX, 0.5*_dimY, _dimZ);

        center.y= _center.y - 0.25*_dimY;
        _children[1]= new AABB(this, center, _dimX, 0.5*_dimY, _dimZ);
     }
     else
     {    
        center.z= _center.z + 0.25*_dimZ;
        _children[0]= new AABB(this, center, _dimX, _dimY, 0.5*_dimZ);

        center.z= _center.z - 0.25*_dimZ;
        _children[1]= new AABB(this, center, _dimX, _dimY, 0.5*_dimZ);
     }
  }

  _children[0]->_brother= _children[1];
  _children[1]->_brother= _children[0];


  _children[0]->testTriangles();
  if(!_children[0]->_inner_triangles.empty() && !_children[0]->_terminate)
  {  _children[0]->divide();  }

  _children[1]->testTriangles();
  if(!_children[1]->_inner_triangles.empty() && !_children[1]->_terminate)
  {  _children[1]->divide();  }


  return 1;
}


int AABB::divide2()
{
  static unsigned int count= 0;

  int axis;
  float dimX, dimY, dimZ;
  //  float min, max;
  Vector3D center;

  center= _center;

  dimX= _xmax - _xmin;
  dimY= _ymax - _ymin;
  dimZ= _zmax - _zmin;


  count++;
  if(count>50000)
  {
    printf("too many iterations\n");
    _leaf= true;
    return 1;
  }  


  if( (dimX < _minimum_size) || (dimY < _minimum_size) || (dimZ < _minimum_size) )
  {
    _leaf= true;
    return 1;
  }

  if( (dimX > dimY) && (dimX > dimZ) )
  {  axis= 1;  }
  else if(dimY > dimZ) 
  {  axis= 2;  }
  else
  {  axis= 3;  }

  switch(axis)
  {
    case 1:
      _children[0]= new AABB(this, _xmin, _xmin+0.5*dimX, _ymin, _ymax, _zmin, _zmax);
      _children[1]= new AABB(this, _xmin+0.5*dimX, _xmax, _ymin, _ymax, _zmin, _zmax);    
/*
      _children[0]->adjust(2);
      _children[0]->adjust(3);
      _children[1]->adjust(2);
      _children[1]->adjust(3);*/
    break;
    case 2:
      _children[0]= new AABB(this, _xmin, _xmax, _ymin, _ymin+0.5*dimY, _zmin, _zmax);
      _children[1]= new AABB(this, _xmin, _xmax, _ymin+0.5*dimY, _ymax, _zmin, _zmax);   
/*
      _children[0]->adjust(1);
      _children[0]->adjust(3);
      _children[1]->adjust(1);
      _children[1]->adjust(3);*/
    break;
    case 3:
      _children[0]= new AABB(this, _xmin, _xmax, _ymin, _ymax, _zmin, _zmin+0.5*dimZ);
      _children[1]= new AABB(this, _xmin, _xmax, _ymin, _ymax, _zmin+0.5*dimZ, _zmax);    
/*
      _children[0]->adjust(1);
      _children[0]->adjust(2);
      _children[1]->adjust(1);
      _children[1]->adjust(2);*/
    break;
  }

  _children[0]->_brother= _children[1];
  _children[1]->_brother= _children[0];

  _children[0]->testTriangles();
  _children[1]->testTriangles();


  if(!_children[0]->_inner_triangles.empty() && !_children[0]->_terminate)
  {  _children[0]->divide2();  }


  if(!_children[1]->_inner_triangles.empty() && !_children[1]->_terminate)
  {  _children[1]->divide2();  }


  return 1;
}

int AABB::adjust(int axis)
{

  if(_mesh==NULL)
  {
    printf("%s: %d: AABB::adjust(): the \"_mesh\" field of the calling AABB is NULL.\n",__FILE__,__LINE__);   
    return 0;
  }
  
  float x, y, z, xmin, xmax, ymin, ymax, zmin, zmax;
  Vector3D p1, p2, p3, new_center;
  std::list<unsigned int>::iterator triangle_index;




  triangle_index= _inner_triangles.begin();
  while( triangle_index != _inner_triangles.end() && !_inner_triangles.empty() )
  {
    _mesh->getTriangleVertices(*triangle_index, p1, p2, p3);

    if(axis==1)
    {
      //case 1:
        if( (p1.x < _xmin) && (p2.x < _xmin) && (p3.x < _xmin) )
        {
          triangle_index= _inner_triangles.erase(triangle_index);
          continue;
        }
        if( (p1.x > _xmax) && (p2.x > _xmax) && (p3.x > _xmax) )
        {
          triangle_index= _inner_triangles.erase(triangle_index);
          continue;
        }
     }
     // break;
    //  case 2:
    if(axis==2)
    {
        if( (p1.y < _ymin) && (p2.y < _ymin) && (p3.y < _ymin) )
        {
          triangle_index= _inner_triangles.erase(triangle_index);
          continue;
        }
        if( (p1.y > _ymax) && (p2.y > _ymax) && (p3.y > _ymax) )
        {
          triangle_index= _inner_triangles.erase(triangle_index);
          continue;
        }
    }
/*
        break;
        case 3:*/
    if(axis==3)
    {
        if( (p1.z < _zmin) && (p2.z < _zmin) && (p3.z < _zmin) )
        {
          triangle_index= _inner_triangles.erase(triangle_index);
          continue;
        }
        if( (p1.z  > _zmax) && (p2.z > _zmax) && (p3.z > _zmax) )
        {
          triangle_index= _inner_triangles.erase(triangle_index);
          continue;
        }
      //break; 
    }
    triangle_index++;
  }
  

  if(!_inner_triangles.empty())
  {//printf("inner tri size= %d\n", _inner_triangles.size());
      _mesh->getTriangleVertices(_inner_triangles.front(), p1, p2, p3);
      xmin= min( p1.x, min(p2.x, p3.x)); 
      xmax= max( p1.x, max(p2.x, p3.x)); 
      ymin= min( p1.y, min(p2.y, p3.y)); 
      ymax= max( p1.y, max(p2.y, p3.y));
      zmin= min( p1.z, min(p2.z, p3.z)); 
      zmax= max( p1.z, max(p2.z, p3.z));
  
      triangle_index= _inner_triangles.begin();
      triangle_index++;
      while( triangle_index != _inner_triangles.end() )
      {
        _mesh->getTriangleVertices(*triangle_index, p1, p2, p3);

        x= min( p1.x, min(p2.x, p3.x)); 
        if(x < xmin)
        {  xmin= x;  }
        x= max( p1.x, max(p2.x, p3.x)); 
        if(x > xmax)
        {  xmax= x;  }  
        
        y= min( p1.y, min(p2.y, p3.y)); 
        if(y < ymin)
        {  ymin= y;  }
        y= max( p1.y, max(p2.y, p3.y)); 
        if(y > ymax)
        {  ymax= y;  }  
        
        z= min( p1.z, min(p2.z, p3.z)); 
        if(z < zmin)
        {  zmin= z;  }
        z= max( p1.z, max(p2.z, p3.z)); 
        if(z > zmax)
        {  zmax= z;  }      
        
        triangle_index++;
      }    

  
      _xmin= xmin;
      _xmax= xmax;
      _ymin= ymin;
      _ymax= ymax;
      _xmin= zmin;
      _zmax= zmax;
  }
  else
  {
    _terminate= false;
  }
  
  
  return 1;
}


int AABB::adjust()
{
  if(_mesh==NULL)
  {
    printf("%s: %d: AABB::adjust(): the \"_mesh\" field of the calling AABB is NULL.\n",__FILE__,__LINE__);   
    return 0;
  }
  
  float x, y, z, xmin, ymin, zmin, xmax, ymax, zmax;
  bool in1, in2, in3;
  Vector3D p1, p2, p3;
  std::list<unsigned int>::iterator triangle_index;


  triangle_index= _inner_triangles.begin();
  while(triangle_index != _inner_triangles.end() && !_inner_triangles.empty())
  {
    _mesh->getTriangleVertices(_inner_triangles.front(), p1, p2, p3);

    in1= isPointInside(p1);
    in2= isPointInside(p2);
    in3= isPointInside(p3);

    //if the triangle is completely outside the box:
    if( (p1.x<_xmin && p2.x<_xmin && p3.x<_xmin) || (p1.x>_xmax && p2.x>_xmax && p3.x>_xmax) ||
        (p1.y<_ymin && p2.y<_ymin && p3.y<_ymin) || (p1.y>_ymax && p2.y>_ymax && p3.y>_ymax) ||
        (p1.z<_zmin && p2.z<_zmin && p3.z<_zmin) || (p1.z>_zmax && p2.z>_zmax && p3.z>_zmax) )
    {
      triangle_index= _inner_triangles.erase(triangle_index);
      continue;
    }

    //if the triangle is completely inside the box:
    if( in1 && in2 && in3 )
    {
       x= min( p1.x, min(p2.x, p3.x)); 
       if(x < xmin)
       {  xmin= x;  }
       x= max( p1.x, max(p2.x, p3.x)); 
       if(x > xmax)
       {  xmax= x;  }  
  
       y= min( p1.y, min(p2.y, p3.y)); 
       if(y < ymin)
       {  ymin= y;  }
       y= max( p1.y, max(p2.y, p3.y)); 
       if(y > ymax)
       {  ymax= y;  }   
       y= min( p1.y, min(p2.y, p3.y));
  
       if(z < zmin)
       {  zmin= z;  }
       z= max( p1.z, max(p2.z, p3.z)); 
       if(z > zmax)
       {  zmax= z;  }   
    }

 

    if( 0 )
    {
       triangle_index= _inner_triangles.erase(triangle_index);
      continue;
    }
    triangle_index++;
  }

  return 1;
}


int AABB::draw()
{
#if defined(ENABLE_DISPLAY)
  //_center.print("center");
 //printf("level= %d %d %d \n", _level, _leaf, _root); 
  if(_inner_triangles.empty())
  {
    glColor3f(0.0, 1.0, 0.0);                            
  }
  else
  {
    glColor3f(1.0, 0.0, 0.0);           
  }
 
 
  glBegin(GL_LINES);
    glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
    glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
    
    glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);
    glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);
    
    glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);
    glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);    
    
    glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);
    glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);
        
  
    glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
    glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ); 
    
    glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);
    glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);  
     
    glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
    glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ); 
    
    glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);
    glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);       
       
    
    glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
    glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ); 
    
    glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);
    glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);  
     
    glVertex3f(_center.x + 0.5*_dimX ,_center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
    glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ); 
    
    glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);
    glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);      
  glEnd(); 
  
  if(_leaf)
  { return 1; }
  
  _children[0]->draw();
  _children[1]->draw();  
#endif
  return 1;  
}



int AABB::draw(unsigned int level, bool fill)
{
#if defined(ENABLE_DISPLAY)
  GLboolean lighting_enable;


  glGetBooleanv(GL_TEXTURE_2D, &lighting_enable);

  if(_root)
  { CURRENT_COLOR= 0; }

  //_center.print("center");
 // printf("dim= %f %f %f\n", _dimX,_dimY,_dimZ); 
 //printf("draw level= %d %d %d %d\n", _level, level, _leaf, _root); 
  if(_inner_triangles.empty())
  {
    glColor3f(0.0, 1.0, 0.0);
  }
  else
  {
    glColor3f(1.0, 0.0, 0.0);
  }
 
  

  if( !_inner_triangles.empty() )
  if( _leaf || _level==level ) 
  {
    CURRENT_COLOR++;
    if(CURRENT_COLOR>PALETTE_SIZE)
    {  CURRENT_COLOR= 0;  }
    glColor3f(PALETTE[CURRENT_COLOR][0], PALETTE[CURRENT_COLOR][1], PALETTE[CURRENT_COLOR][2]);

    if(!fill)
    {   
      glDisable(GL_LIGHTING);
      glBegin(GL_LINES);/*
        glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
        
        glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);
        
        glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);    
        
        glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);
            
      
        glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ); 
        
        glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);
        glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);  
         
        glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ); 
        
        glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);       
           
        
        glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ); 
        
        glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);  
         
        glVertex3f(_center.x + 0.5*_dimX ,_center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ); 
        
        glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);  */

        glVertex3f(_xmin, _ymin, _zmin);
        glVertex3f(_xmax, _ymin, _zmin);

        glVertex3f(_xmin, _ymax, _zmin);
        glVertex3f(_xmax, _ymax, _zmin);

        glVertex3f(_xmin, _ymax, _zmax);
        glVertex3f(_xmax, _ymax, _zmax);

        glVertex3f(_xmin, _ymin, _zmax);
        glVertex3f(_xmax, _ymin, _zmax);

        glVertex3f(_xmin, _ymin, _zmin);
        glVertex3f(_xmin, _ymax, _zmin);

        glVertex3f(_xmax, _ymin, _zmin);
        glVertex3f(_xmax, _ymax, _zmin);

        glVertex3f(_xmax, _ymin, _zmax);
        glVertex3f(_xmax, _ymax, _zmax);

        glVertex3f(_xmin, _ymin, _zmax);
        glVertex3f(_xmin, _ymax, _zmax);

        glVertex3f(_xmin, _ymin, _zmin);
        glVertex3f(_xmin, _ymin, _zmax);

        glVertex3f(_xmax, _ymin, _zmin);
        glVertex3f(_xmax, _ymin, _zmax);

        glVertex3f(_xmax, _ymax, _zmin);
        glVertex3f(_xmax, _ymax, _zmax);

        glVertex3f(_xmin, _ymax, _zmin);
        glVertex3f(_xmin, _ymax, _zmax);
      glEnd(); 
    }
    else
    {   
      glEnable(GL_LIGHTING);
      glBegin(GL_QUADS);/*
        glNormal3f(1.0, 0.0, 0.0);
        glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);

        glNormal3f(-1.0, 0.0, 0.0);
        glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);
        glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);

        glNormal3f(0.0, 1.0, 0.0);
        glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);
      
        glNormal3f(0.0, -1.0, 0.0);
        glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);
        glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);

        glNormal3f(0.0, 0.0, 1.0);
        glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z + 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);
        glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z + 0.5*_dimZ);

        glNormal3f(0.0, 0.0, -1.0);
        glVertex3f(_center.x + 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x - 0.5*_dimX, _center.y - 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x - 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);
        glVertex3f(_center.x + 0.5*_dimX, _center.y + 0.5*_dimY, _center.z - 0.5*_dimZ);*/
        glNormal3f(1.0, 0.0, 0.0);
        glVertex3f(_xmax, _ymin, _zmin);
        glVertex3f(_xmax, _ymax, _zmin);
        glVertex3f(_xmax, _ymax, _zmax);
        glVertex3f(_xmax, _ymin, _zmax);

        glNormal3f(-1.0, 0.0, 0.0);
        glVertex3f(_xmin, _ymax, _zmin);
        glVertex3f(_xmin, _ymin, _zmin);
        glVertex3f(_xmin, _ymin, _zmax);
        glVertex3f(_xmin, _ymax, _zmax);

        glNormal3f(0.0, 1.0, 0.0);
        glVertex3f(_xmax, _ymax, _zmin);
        glVertex3f(_xmin, _ymax, _zmin);
        glVertex3f(_xmin, _ymax, _zmax);
        glVertex3f(_xmax, _ymax, _zmax);
      
        glNormal3f(0.0, -1.0, 0.0);
        glVertex3f(_xmin, _ymin, _zmin);
        glVertex3f(_xmax, _ymin, _zmin);
        glVertex3f(_xmax, _ymin, _zmax);
        glVertex3f(_xmin, _ymin, _zmax);

        glNormal3f(0.0, 0.0, 1.0);
        glVertex3f(_xmin, _ymin, _zmax);
        glVertex3f(_xmax, _ymin, _zmax);
        glVertex3f(_xmax, _ymax, _zmax);
        glVertex3f(_xmin, _ymax, _zmax);

        glNormal3f(0.0, 0.0, -1.0);
        glVertex3f(_xmax, _ymin, _zmin);
        glVertex3f(_xmin, _ymin, _zmin);
        glVertex3f(_xmin, _ymax, _zmin);
        glVertex3f(_xmax, _ymax, _zmin);
      glEnd(); 
    }
  }
  
  if(lighting_enable)
  {  glEnable(GL_LIGHTING);  }
  else
  {  glDisable(GL_LIGHTING);  }

  if(_leaf || _level>=level)
  {  return 1;  }
  
  _children[0]->draw(level, fill);
  _children[1]->draw(level, fill);  
#endif    
  return 1;  

}

//! WIP
AABB* TriangleMesh::kdTreeDecomposition(float resolution)
{
  unsigned int i;
  float xmin, xmax, ymin, ymax, zmin, zmax;
  Vector3D center;
  AABB *root= NULL;
  std::list<unsigned int> inner_triangles;

  if( (vertices.size() < 4) || (triangles.size()<4) )
  {
     printf("%s: %d: octreeDecomposition(): the mesh must have at least four vertices and four triangles to describe a volume.\n",__FILE__,__LINE__);
    return NULL;
  }

  xmin= vertices[0].x;
  xmax= vertices[0].x;
  ymin= vertices[0].y;
  ymax= vertices[0].y;
  zmin= vertices[0].z;
  zmax= vertices[0].z;

  for(i=1; i<vertices.size(); ++i)
  {
    if( vertices[i].x < xmin ) 
    { xmin=  vertices[i].x; }

    if( vertices[i].x > xmax ) 
    { xmax=  vertices[i].x; }

    if( vertices[i].y < ymin ) 
    { ymin=  vertices[i].y; }

    if( vertices[i].y > ymax ) 
    { ymax=  vertices[i].y; }

    if( vertices[i].z < zmin ) 
    { zmin=  vertices[i].z; }

    if( vertices[i].z > zmax ) 
    { zmax=  vertices[i].z; }
  }

  center.set( 0.5*(xmin+xmax), 0.5*(ymin+ymax), 0.5*(zmin+zmax) );

  for(i=0; i<triangles.size(); ++i)
  {
    inner_triangles.push_back(i);
  }

  //root= new AABB(center, xmax-xmin, ymax-ymin, zmax-zmin, resolution, this, inner_triangles);
  root= new AABB(xmin, xmax, ymin, ymax, zmin, zmax, resolution, this, inner_triangles);

  root->divide2();

  return root;
}

//! Transforms the vertex positions and normals of the mesh.
//! \param T the homogeneous transform matrix to apply
//! \return 1 in case of success, 0 otherwise
int TriangleMesh::vertexTransform(const HTMatrix &T)
{
  unsigned int i;
 
  for(i=0; i<vertices.size(); ++i)
  {
    vertices[i]= T*vertices[i];  
  }

  for(i=0; i<normals.size(); ++i)
  {
    normals[i]= rotationMult(T, normals[i]);
  }

  return 1;
}


Body::Body()
{ 
  _name= "none";
  _joint= NULL;
  _pose.setIdentity();
  _type= DISPLAY_AND_COLLISION;
  _color.set(0.6, 0.6, 0.6);
  _texture= "none";
  material= PLASTIC;


  _pqpModel= NULL; 


  colliding       = false;
  enable_collision= true;
  enable_display  = true;
  enable_shadow   = true;
  enable_texture  = false;
}

Body::Body(BodyType type)
{ 
  _name= "none";
  _joint= NULL;
  _pose.setIdentity();
  _type= type;
  _color.set(0.6, 0.6, 0.6);
  _texture= "none";
  material= PLASTIC;

  _pqpModel= NULL; 

  colliding       = false;
 
  switch(type)
  {
    case DISPLAY_ONLY:
      enable_collision= false;
      enable_display  = true;
    break;
    case COLLISION_ONLY:
      enable_collision= true;
      enable_display  = false;
    break;
    case DISPLAY_AND_COLLISION:
      enable_collision= true;
      enable_display  = true;
    break;
    case NO_DISPLAY_NO_COLLISION:
      enable_collision= false;
      enable_display  = false;
    break;
  }

  enable_shadow   = true;
  enable_texture  = false;
}

//! Gets the body's name.
std::string Body::name()
{ return _name; }

//! Gets the body's pose.
HTMatrix Body::pose()
{ return _pose; }

//! Gets a reference to the body's mesh.
TriangleMesh& Body::mesh()
{ return _mesh; }

//! Gets the joint the body is linked to body.
Joint * Body::joint()
{ return _joint; }

//! Gets the body's Tjoint matrix.
HTMatrix Body::Tjoint()
{ return _Tjoint; }

//! Gets a pointer to the body's PQP_Model.
PQP_Model* Body::pqpModel()
{ return _pqpModel; }

//! Gets the body's type.
BodyType Body::type()
{ return _type; }

//! Creates a Body from a mesh described in a file (.obj or .macro). 
//! \param name _name of the body
//! \param type type of the body (see Body.h)
//! \param filename name of the file to load (.obj or .macro)
//! \param scale scale factor between the model described in the file and the Body model that will be created
Body::Body(std::string name, BodyType type, std::string filename, float scale)
{
  _name= name;
  _pose.setIdentity();
  _joint= NULL;
  _type= type;
  _color.set(0.2, 0.9, 0.9);
  _texture= "none";


  _pqpModel= NULL; 

  colliding= false;
  switch(type)
  {
    case DISPLAY_ONLY:
      enable_collision= false;
      enable_display  = true;
    break;
    case COLLISION_ONLY:
      enable_collision= true;
      enable_display  = false;
    break;
    case DISPLAY_AND_COLLISION:
      enable_collision= true;
      enable_display  = true;
    break;
    case NO_DISPLAY_NO_COLLISION:
      enable_collision= false;
      enable_display  = false;
    break;
  }

  enable_shadow   = true;
  enable_texture  = false;

  loadMeshFile(filename, scale);
}

Body::Body(const std::string &name, const BodyType type, const TriangleMesh &mesh)
{
  if( (mesh.nbVertices() < 3) || (mesh.nbTriangles() < 1) )
  {
    printf("%s: %d: Body::Body(): the input TriangleMesh is not valid (%u vertices, %u triangles).\n",__FILE__,__LINE__,mesh.nbVertices(),mesh.nbTriangles());
    return;
  }

  _name= name;
  _pose.setIdentity();
  _joint= NULL;
  _type= type;
  _color.set(0.2, 0.9, 0.9);
  _texture= "none";


  _pqpModel= NULL; 

  colliding= false;
  switch(type)
  {
    case DISPLAY_ONLY:
      enable_collision= false;
      enable_display  = true;
    break;
    case COLLISION_ONLY:
      enable_collision= true;
      enable_display  = false;
    break;
    case DISPLAY_AND_COLLISION:
      enable_collision= true;
      enable_display  = true;
    break;
    case NO_DISPLAY_NO_COLLISION:
      enable_collision= false;
      enable_display  = false;
    break;
  }

  enable_shadow   = true;
  enable_texture  = false;


  _mesh= mesh;
}

Body::Body(const Body &body)
{
  _name  = body._name;
  _mesh  = body._mesh;
  _pose  = body._pose;
  _joint = body._joint;
  _Tjoint= body._Tjoint;
  _type  = body._type;

  _color  = body._color;
  material= body.material;

  colliding       = body.colliding;
  enable_collision= body.enable_collision;  
  enable_display  = body.enable_display;  
  enable_shadow   = body.enable_shadow;  

  if(body._pqpModel!=NULL)
  {  createPQPmodel();  }
}


Body::~Body()
{ 
  if(_pqpModel!=NULL)
  {  delete _pqpModel;  }
}    


Body & Body::operator = (const Body &body)
{
  if(this==&body) 
  {  return *this;  }

  _name  = body._name;
  _mesh  = body._mesh;
  _pose  = body._pose;
  _joint = body._joint;
  _Tjoint= body._Tjoint;
  _type  = body._type;

  _color  = body._color;
  material= body.material;

  colliding       = body.colliding;
  enable_collision= body.enable_collision;  
  enable_display  = body.enable_display;  
  enable_shadow   = body.enable_shadow;  

  if(body._pqpModel!=NULL)
  {  createPQPmodel();  }

  return *this;
}

Box::Box(std::string name, BodyType type, float dimX, float dimY, float dimZ): Body(type)
{
  _name= name;
  _type= type;
  switch(type)
  {
    case DISPLAY_ONLY:
      enable_collision= false;
      enable_display  = true;
    break;
    case COLLISION_ONLY:
      enable_collision= true;
      enable_display  = false;
    break;
    case DISPLAY_AND_COLLISION:
      enable_collision= true;
      enable_display  = true;
    break;
    case NO_DISPLAY_NO_COLLISION:
      enable_collision= false;
      enable_display  = false;
    break;
  }
  _dimX= dimX;
  _dimY= dimY;
  _dimZ= dimZ;
  _mesh.makeBox(_dimX, _dimY, _dimZ);
}

Cylinder::Cylinder(std::string name, BodyType type, float radius, float height, unsigned int nbSlices):Body(type)
{
  _name= name;
  _type= type;
  switch(type)
  {
    case DISPLAY_ONLY:
      enable_collision= false;
      enable_display  = true;
    break;
    case COLLISION_ONLY:
      enable_collision= true;
      enable_display  = false;
    break;
    case DISPLAY_AND_COLLISION:
      enable_collision= true;
      enable_display  = true;
    break;
    case NO_DISPLAY_NO_COLLISION:
      enable_collision= false;
      enable_display  = false;
    break;
  }
  _radius= radius;
  _height= height;
  _mesh.makeCylinder(_radius, _height, nbSlices);
}

Sphere::Sphere(std::string name, BodyType type, float radius, unsigned int nbSlices, unsigned int nbStacks):Body(type)
{
  _name= name;
  _type= type;
  switch(type)
  {
    case DISPLAY_ONLY:
      enable_collision= false;
      enable_display  = true;
    break;
    case COLLISION_ONLY:
      enable_collision= true;
      enable_display  = false;
    break;
    case DISPLAY_AND_COLLISION:
      enable_collision= true;
      enable_display  = true;
    break;
    case NO_DISPLAY_NO_COLLISION:
      enable_collision= false;
      enable_display  = false;
    break;
  }
  _radius= radius;
 // _mesh.makeSphere(radius, nbSlices, nbStacks);
  _mesh.makeSphere(_radius, 55, 55);
}

//! Creates the TriangleMesh of the body from a 3D model file (.obj or .macro).
//! \param filename name of the file to load (.obj or .macro)
//! \param scale scale factor between the model described in the file and the Body model that will be created
//! \return 1 in case of success, 0 otherwise
int Body::loadMeshFile(std::string filename, float scale)
{
  unsigned int length;
  size_t pos;


  length= filename.length(); 
  pos= filename.find(".obj"); 
  if(pos==length-4)
  {
    return _mesh.readOBJfile(filename, scale);
  }    


  pos= filename.find(".macro"); 
  if(pos==length-6)
  {
    return _mesh.readP3Dfile(filename, scale);
  }  


  printf("File %s does not have the good extension (it should be a .obj or a .macro file).\n", filename.c_str());
  return 0;

}


//! Creates the collision model of the body from its TriangleMesh.
//! \return 1 in case of success, 0 otherwise
int Body::createPQPmodel()
{
   unsigned int i;
   PQP_REAL a[3], b[3], c[3];

   if(_pqpModel!=NULL)
   {  delete _pqpModel;  }
   _pqpModel= NULL;

   if(_type==DISPLAY_ONLY || _type==NO_DISPLAY_NO_COLLISION)
   {  return 0;  }

//    if(enable_collision==false)
//    {
//      printf("%s: %d: Body::createPQPmodel(): collision is deactivated for body \"%s\".\n", __FILE__, __LINE__, _name.c_str());
//      return 0;
//    }

   if(_mesh.triangles.empty())
   {
     printf("%s: %d: Body::createPQPmodel(): the body mesh of \"%s\" does not have any triangles.\n", __FILE__, __LINE__, _name.c_str());
     enable_collision= false;
     return 0;
   }

   _pqpModel= new PQP_Model; 

   _pqpModel->BeginModel(_mesh.triangles.size());

   for(i=0; i<_mesh.triangles.size(); ++i)
   {
     a[0]= _mesh.vertices[ _mesh.triangles[i].v[0] ].x;
     a[1]= _mesh.vertices[ _mesh.triangles[i].v[0] ].y;
     a[2]= _mesh.vertices[ _mesh.triangles[i].v[0] ].z;

     b[0]= _mesh.vertices[ _mesh.triangles[i].v[1] ].x;
     b[1]= _mesh.vertices[ _mesh.triangles[i].v[1] ].y;
     b[2]= _mesh.vertices[ _mesh.triangles[i].v[1] ].z;

     c[0]= _mesh.vertices[ _mesh.triangles[i].v[2] ].x;
     c[1]= _mesh.vertices[ _mesh.triangles[i].v[2] ].y;
     c[2]= _mesh.vertices[ _mesh.triangles[i].v[2] ].z;
     _pqpModel->AddTri(a, b, c, i);

     if( i!= (unsigned int) _pqpModel->tris[i].id )
       printf("tri: index= %d id= %d ---\n", i, _pqpModel->tris[i].id);
  }
  _pqpModel->EndModel();

  for(i=0; i<_mesh.triangles.size(); ++i)
  {
    _pqpModel->tris[i].id= i;
  }

  return 1;
}


void Body::setPosition(float t[3])
{
  _pose.setTranslation(t[0], t[1], t[2]);
}

void Body::setPosition(float x, float y, float z)
{
  _pose.setTranslation(x, y, z);
}

void Body::setOrientation(float yaw, float pitch, float roll)
{
  _pose.setRotationFromYPRangles(yaw, pitch, roll);
}

void Body::setPose(float x, float y, float z, float yaw, float pitch, float roll)
{
  _pose.setTranslation(x, y, z);
  _pose.setRotationFromYPRangles(yaw, pitch, roll);
}

void Body::setPose(const HTMatrix &M)
{
  _pose= M;
}

//! Sets the joint the body will be linked to.
//! \param joint pointer to the joint
//! \return 1 in case of success, 0 otherwise
int Body::setJoint(Joint *joint)
{
  if(joint==NULL)
  {
    printf("Body::setJoint(): input joint is NULL.\n");
    return 0;
  }
 
  _joint= joint;

  return 0;
}

void Body::setTjoint(const HTMatrix &M)
{
  _Tjoint= M;
}

void Body::rename(const std::string name)
{
  _name= name;
}

void Body::setTexture(const std::string texture)
{
  _texture= texture;
}


void Body::printConfiguration()
{
  printf("R= [ %f %f %f ]\n", _pose.m11, _pose.m12, _pose.m13);
  printf("   [ %f %f %f ]\n", _pose.m21, _pose.m22, _pose.m33);
  printf("   [ %f %f %f ]\n", _pose.m31, _pose.m32, _pose.m33);
  printf("t= [ %f ]\n", _pose.m14);
  printf("   [ %f ]\n", _pose.m24);
  printf("   [ %f ]\n", _pose.m34);
}

void Body::setColor(float r, float g, float b, float alpha)
{
  _color.set(r, g, b, alpha);
}

void Body::setMaterial(Material _material)
{
  material= _material;
}

//! Draws the body.
//! \param smooth use smooth (true) or flat (false) shading
//! \return 1 in case of success, 0 otherwise
int Body::draw(bool smooth)
{
#if defined(ENABLE_DISPLAY)
  if(!enable_display && !colliding)
  {  return 1;  }


  GLboolean texture_enable;
  GLfloat mat[16];


  glGetBooleanv(GL_TEXTURE_2D, &texture_enable);

  _pose.convertToOpenGLFormat(mat);

  if(colliding)
  {  glColor3f(1.0, 0.0, 0.0);  }
  else
  {  glColor3f(_color.r, _color.g, _color.b);  }


  if(enable_texture)
  {  glEnable(GL_TEXTURE_2D);  }
  else
  {  glDisable(GL_TEXTURE_2D);  }

  glPushMatrix();
   glMultMatrixf(mat);
   _mesh.draw(smooth);
  glPopMatrix();

  if(texture_enable)
  {  glEnable(GL_TEXTURE_2D);  }
  else
  {  glDisable(GL_TEXTURE_2D);  }
#endif
  return 1;
}


//! Draws the body in black and unlit.
//! This function is meant to be used to display planar shadows.
//! \return 1 in case of success, 0 otherwise
int Body::drawInBlack()
{
#if defined(ENABLE_DISPLAY)
  if( !enable_display || !enable_shadow )
  {  return 1;  }

  GLboolean lighting_enable;
  GLfloat mat[16];

  _pose.convertToOpenGLFormat(mat);

  glGetBooleanv(GL_LIGHTING, &lighting_enable);

 // glDisable(GL_LIGHTING);
 // glColor3f(0.0, 0.0, 0.0);

  glPushMatrix();
   glMultMatrixf(mat);
   _mesh.draw(false);
  glPopMatrix();

  if(lighting_enable)
  {
    glEnable(GL_LIGHTING);
  }
#endif
  return 1;
}

//! Draws the OBBs (Oriented Bounding Box) of the PQP model of the Body.
//! \param level only the OBBs from the root level to the given level in the hierarchy tree are drawn
//! \return 1 in case of success, 0 otherwise
int Body::drawOBBs(unsigned int level)
{
#if defined(ENABLE_DISPLAY)
  if(_pqpModel==NULL)
  {
    printf("Body::drawOBBs(): the body has no PQP model.\n");
    return 0;
  }
  GLfloat mat[16];
  float T[4][4];

  _pose.convertToOpenGLFormat(mat);

  T[0][0]=  1.0;   T[0][1]=  0.0;   T[0][2]=  0.0;   T[0][3]=  0.0; 
  T[1][0]=  0.0;   T[1][1]=  1.0;   T[1][2]=  0.0;   T[1][3]=  0.0;  
  T[2][0]=  0.0;   T[2][1]=  0.0;   T[2][2]=  1.0;   T[2][3]=  0.0; 
  T[3][0]=  0.0;   T[3][1]=  0.0;   T[3][2]=  0.0;   T[3][3]=  1.0; 

  glPushMatrix();
    glMultMatrixf(mat);
    drawOBBs_recursive(_pqpModel, T, 0, 0, level);
  glPopMatrix();
#endif
  return 1;
}


//! Recursive function needed by drawOBBs(). It must not be used directly.
//! \sa drawOBBs.
int drawOBBs_recursive(PQP_Model *pqpModel, float T[4][4], int bn, unsigned int currentLevel, unsigned int level)
{
#if defined(ENABLE_DISPLAY)
  if(pqpModel==NULL)
  {
    printf("Body::drawOBBs_recursive(): the body has no PQP model.\n");
    return 0;
  }

  if(currentLevel + 1 > level)
  {   return 1;   }

  int i, j;
  float T2[4][4], T3[4][4];
  float a[3], b[3], c[3], d[3], e[3], f[3], g[3], h[3];    

  BV* bv= pqpModel->child(bn);
  if(bv->Leaf())
  {  
    i= -bv->first_child-1;   
    glColor3f(0.0, 0.0, 0.0);  
      glBegin(GL_TRIANGLES);
        glVertex3f(pqpModel->tris[i].p1[0], pqpModel->tris[i].p1[1], pqpModel->tris[i].p1[2]); 
        glVertex3f(pqpModel->tris[i].p2[0], pqpModel->tris[i].p2[1], pqpModel->tris[i].p2[2]); 
        glVertex3f(pqpModel->tris[i].p3[0], pqpModel->tris[i].p3[1], pqpModel->tris[i].p3[2]); 
     glEnd();
    return 1; 
  }
  else
  {    
     T2[0][0]= bv->R[0][0];   T2[0][1]= bv->R[0][1];   T2[0][2]= bv->R[0][2];   T2[0][3]= bv->To[0]; 
     T2[1][0]= bv->R[1][0];   T2[1][1]= bv->R[1][1];   T2[1][2]= bv->R[1][2];   T2[1][3]= bv->To[1];  
     T2[2][0]= bv->R[2][0];   T2[2][1]= bv->R[2][1];   T2[2][2]= bv->R[2][2];   T2[2][3]= bv->To[2]; 
     T2[3][0]=           0;   T2[3][1]=           0;   T2[3][2]=           0;   T2[3][3]=         1; 

     for(i=0; i<4; ++i)
     {
       for(j=0; j<4; j++)
       {
         T3[i][j]= T[i][0]*T2[0][j] + T[i][1]*T2[1][j] + T[i][2]*T2[2][j] + T[i][3]*T2[3][j] ;
       }
     }
 

     if(currentLevel + 1==level)
     {
       for(i=0; i<3; ++i)
       {  
         a[i]=   T3[i][0]*bv->d[0] +  T3[i][1]*bv->d[1] +  T3[i][2]*bv->d[2] +  T3[i][3];
         b[i]=   T3[i][0]*bv->d[0] +  T3[i][1]*bv->d[1] -  T3[i][2]*bv->d[2] +  T3[i][3];
         c[i]=   T3[i][0]*bv->d[0] -  T3[i][1]*bv->d[1] +  T3[i][2]*bv->d[2] +  T3[i][3];
         d[i]=   T3[i][0]*bv->d[0] -  T3[i][1]*bv->d[1] -  T3[i][2]*bv->d[2] +  T3[i][3];
         e[i]= - T3[i][0]*bv->d[0] +  T3[i][1]*bv->d[1] +  T3[i][2]*bv->d[2] +  T3[i][3];
         f[i]= - T3[i][0]*bv->d[0] +  T3[i][1]*bv->d[1] -  T3[i][2]*bv->d[2] +  T3[i][3];
         g[i]= - T3[i][0]*bv->d[0] -  T3[i][1]*bv->d[1] +  T3[i][2]*bv->d[2] +  T3[i][3];
         h[i]= - T3[i][0]*bv->d[0] -  T3[i][1]*bv->d[1] -  T3[i][2]*bv->d[2] +  T3[i][3];
       }

       glColor3f(0.0, 0.0, 0.0); 
       glBegin(GL_LINES);
         glVertex3fv(a);  glVertex3fv(b);
         glVertex3fv(a);  glVertex3fv(c);
         glVertex3fv(a);  glVertex3fv(e);
         glVertex3fv(b);  glVertex3fv(d);
         glVertex3fv(b);  glVertex3fv(f);

         glVertex3fv(c);  glVertex3fv(d);
         glVertex3fv(c);  glVertex3fv(g);
         glVertex3fv(d);  glVertex3fv(h);
         glVertex3fv(e);  glVertex3fv(f);
         glVertex3fv(e);  glVertex3fv(g);

         glVertex3fv(d);  glVertex3fv(h);
         glVertex3fv(e);  glVertex3fv(f);
         glVertex3fv(e);  glVertex3fv(g);
         glVertex3fv(f);  glVertex3fv(h);
         glVertex3fv(g);  glVertex3fv(h);
       glEnd();
     }
  }

  drawOBBs_recursive(pqpModel, T3, bv->first_child, currentLevel+1, level);
  drawOBBs_recursive(pqpModel, T3, bv->first_child+1, currentLevel+1, level);
 #endif
  return 1;
} 




//! Draws the PQP model of the Body with the given color.
//! \param red red component of the desired color
//! \param green green component of the desired color
//! \param blue blue component of the desired color
//! \return 1 in case of success, 0 otherwise
int Body::drawPQPmodel(float red, float green, float blue)
{
#if defined(ENABLE_DISPLAY)

  if(_pqpModel==NULL)
  {
    printf("Body::drawPQPmodel(): the body has no PQP model.\n");
    return 0;
  }

  int i;

  glColor3f(red, green, blue); 
  float u[3], v[3], n[3];
  float norm;
  for(i=0; i<_pqpModel->num_tris; ++i)
  { 
    v[0]= _pqpModel->tris[i].p2[0] - _pqpModel->tris[i].p3[0];
    v[1]= _pqpModel->tris[i].p2[1] - _pqpModel->tris[i].p3[1];   
    v[2]= _pqpModel->tris[i].p2[2] - _pqpModel->tris[i].p3[2];   
    u[0]= _pqpModel->tris[i].p1[0] - _pqpModel->tris[i].p3[0];
    u[1]= _pqpModel->tris[i].p1[1] - _pqpModel->tris[i].p3[1];   
    u[2]= _pqpModel->tris[i].p1[2] - _pqpModel->tris[i].p3[2];  
    n[0]= u[1]*v[2] - u[2]*v[1];
    n[1]= u[2]*v[0] - u[0]*v[2];
    n[2]= u[0]*v[1] - u[1]*v[0];
    norm= sqrt( pow(n[0],2) + pow(n[1],2) + pow(n[2],2) );

    glNormal3f(n[0]/norm, n[1]/norm, n[2]/norm);   
    glBegin(GL_TRIANGLES);
      glVertex3f(_pqpModel->tris[i].p1[0], _pqpModel->tris[i].p1[1], _pqpModel->tris[i].p1[2]);
      glVertex3f(_pqpModel->tris[i].p2[0], _pqpModel->tris[i].p2[1], _pqpModel->tris[i].p2[2]);
      glVertex3f(_pqpModel->tris[i].p3[0], _pqpModel->tris[i].p3[1], _pqpModel->tris[i].p3[2]);      
    glEnd(); 
    glNormal3f(-n[0]/norm,-n[1]/norm,-n[2]/norm);   
    glBegin(GL_TRIANGLES);
      glVertex3f(_pqpModel->tris[i].p1[0], _pqpModel->tris[i].p1[1], _pqpModel->tris[i].p1[2]);
      glVertex3f(_pqpModel->tris[i].p3[0], _pqpModel->tris[i].p3[1], _pqpModel->tris[i].p3[2]);      
      glVertex3f(_pqpModel->tris[i].p2[0], _pqpModel->tris[i].p2[1], _pqpModel->tris[i].p2[2]);
    glEnd();  
  
    glColor3f(0,0,0);
    glBegin(GL_LINES);
      glVertex3f(_pqpModel->tris[i].p1[0],_pqpModel->tris[i].p1[1],_pqpModel->tris[i].p1[2]);
      glVertex3f(_pqpModel->tris[i].p2[0],_pqpModel->tris[i].p2[1],_pqpModel->tris[i].p2[2]);
    glEnd();  
    glBegin(GL_LINES);
      glVertex3f(_pqpModel->tris[i].p1[0],_pqpModel->tris[i].p1[1],_pqpModel->tris[i].p1[2]);
      glVertex3f(_pqpModel->tris[i].p3[0],_pqpModel->tris[i].p3[1],_pqpModel->tris[i].p3[2]);        
    glEnd(); 
    glBegin(GL_LINES);
      glVertex3f(_pqpModel->tris[i].p3[0],_pqpModel->tris[i].p3[1],_pqpModel->tris[i].p3[2]);      
      glVertex3f(_pqpModel->tris[i].p2[0],_pqpModel->tris[i].p2[1],_pqpModel->tris[i].p2[2]);
    glEnd(); 
    glColor3f(red, green, blue); 
  
  }   
#endif
  return 1;
}




//! Draws a triangle of the PQP model of the Body. 
//! \param index index of the selected triangle
//! \return 1 in case of success, 0 otherwise
int Body::drawPQPtriangle(unsigned int index)
{
#if defined(ENABLE_DISPLAY)
  if(_pqpModel==NULL)
  {
    printf("Body::drawPQPtriangle(): the body has no PQP model.\n");
    return 0;
  }

  if( (index > (unsigned int) _pqpModel->num_tris-1) )
  {
    printf("Body::drawPQPtriangle(): index exceeds triangle array dimension.\n");
    return 0;
  }

  float norm, u[3], v[3], n[3];
  GLfloat mat[16];

  _pose.convertToOpenGLFormat(mat);

  if((unsigned int) _pqpModel->tris[index].id!=index) printf("id!=index (%d %d)\n", _pqpModel->tris[index].id, index);

  v[0]= _pqpModel->tris[index].p2[0] - _pqpModel->tris[index].p3[0];
  v[1]= _pqpModel->tris[index].p2[1] - _pqpModel->tris[index].p3[1];   
  v[2]= _pqpModel->tris[index].p2[2] - _pqpModel->tris[index].p3[2];   
  u[0]= _pqpModel->tris[index].p1[0] - _pqpModel->tris[index].p3[0];
  u[1]= _pqpModel->tris[index].p1[1] - _pqpModel->tris[index].p3[1];   
  u[2]= _pqpModel->tris[index].p1[2] - _pqpModel->tris[index].p3[2];  
  n[0]= u[1]*v[2] - u[2]*v[1];
  n[1]= u[2]*v[0] - u[0]*v[2];
  n[2]= u[0]*v[1] - u[1]*v[0];
  norm= sqrt( n[0]*n[0] + n[1]*n[1] + n[2]*n[2] );
  n[0]/= norm;
  n[1]/= norm;
  n[2]/= norm;


  glPushMatrix();
  glMultMatrixf(mat);
    glNormal3f(n[0], n[1], n[2]); 
    glBegin(GL_TRIANGLES);
     glVertex3f(_pqpModel->tris[index].p1[0], _pqpModel->tris[index].p1[1], _pqpModel->tris[index].p1[2] );
     glVertex3f(_pqpModel->tris[index].p2[0], _pqpModel->tris[index].p2[1], _pqpModel->tris[index].p2[2] );
     glVertex3f(_pqpModel->tris[index].p3[0], _pqpModel->tris[index].p3[1], _pqpModel->tris[index].p3[2] );
    glEnd(); 
    glBegin(GL_TRIANGLES);
     glVertex3f(_pqpModel->tris[index].p1[0], _pqpModel->tris[index].p1[1], _pqpModel->tris[index].p1[2] );
     glVertex3f(_pqpModel->tris[index].p3[0], _pqpModel->tris[index].p3[1], _pqpModel->tris[index].p3[2] );
     glVertex3f(_pqpModel->tris[index].p2[0], _pqpModel->tris[index].p2[1], _pqpModel->tris[index].p2[2] );
    glEnd(); 
  glPopMatrix();
 #endif
  return 1;
}



//! Performs a collision test between two bodies.
//! \param body1 pointer to the first body
//! \param body2 pointer to the second body 
//! \return 0 if there is no collision, 1 otherwise
int collisionTest(Body *body1, Body *body2)
{
  if(body1==NULL || body2==NULL)
  { 
    printf("collisionTest(): NULL input(s).\n");
    return 0;
  }
  if(body1->type()==DISPLAY_ONLY || body1->type()==NO_DISPLAY_NO_COLLISION)
  { 
    return 0;
  }
  if(body2->type()==DISPLAY_ONLY || body2->type()==NO_DISPLAY_NO_COLLISION)
  { 
    return 0;
  }
  if(body1->enable_collision==false || body2->enable_collision==false)
  { 
    return 0;
  }



  if(body1->pqpModel()==NULL || body2->pqpModel()==NULL)
  { 
    printf("%s: %d: collisionTest(): the PQP models of the bodies have not been created.\n",__FILE__,__LINE__);
    return 0;
  }

  HTMatrix M1, M2;
  PQP_REAL t1[3], t2[3];  
  PQP_REAL R1[3][3], R2[3][3];
  PQP_CollideResult cres;
  
  //body1->printConfiguration();
  //body2->printConfiguration();
  M1= body1->pose();
  M2= body2->pose();

  t1[0]= M1.m14;
  t1[1]= M1.m24;
  t1[2]= M1.m34;
  R1[0][0]= M1.m11;      R1[0][1]= M1.m12;      R1[0][2]= M1.m13;
  R1[1][0]= M1.m21;      R1[1][1]= M1.m22;      R1[1][2]= M1.m23;
  R1[2][0]= M1.m31;      R1[2][1]= M1.m32;      R1[2][2]= M1.m33;

  t2[0]= M2.m14;
  t2[1]= M2.m24;
  t2[2]= M2.m34;
  R2[0][0]= M2.m11;      R2[0][1]= M2.m12;      R2[0][2]= M2.m13;
  R2[1][0]= M2.m21;      R2[1][1]= M2.m22;      R2[1][2]= M2.m23;
  R2[2][0]= M2.m31;      R2[2][1]= M2.m32;      R2[2][2]= M2.m33;


  PQP_Collide(&cres, R1, t1, body1->pqpModel(), R2, t2, body2->pqpModel(), PQP_FIRST_CONTACT);

  if( cres.NumPairs()!=0 )
  {
    body1->colliding= true;
    body2->colliding= true;
  }
  else
  {
    body1->colliding= false;
    body2->colliding= false;
  }  

  //printf("NumPairs= %d\n", cres.NumPairs());
  return cres.NumPairs();


}


//! Computes the distance between two given bodies.
//! \param body1 pointer to the first body
//! \param body2 pointer to the second body
//! \param p1 the closest point on the first body (in absolute coordinates)
//! \param p2 the closest point on the second body (in absolute coordinates)
//! \return 0 in case of error or if the bodies are colliding, the distance otherwise
float distanceComputation(Body *body1, Body *body2, float p1[3], float p2[3])
{
  if(body1==NULL || body2==NULL)
  { 
    printf("distanceComputation(): NULL input(s).\n");
    return 0;
  }
  if(body1->type()==DISPLAY_ONLY || body1->type()==NO_DISPLAY_NO_COLLISION)
  { 
    return 0;
  }
  if(body2->type()==DISPLAY_ONLY || body2->type()==NO_DISPLAY_NO_COLLISION)
  { 
    return 0;
  }


  if(body1->pqpModel()==NULL || body2->pqpModel()==NULL)
  { 
    printf("distanceComputation(): the PQP models have not been created.\n");
    return 0;
  }

  float p1_rel[3], p2_rel[3];

  HTMatrix M1, M2;
  PQP_REAL t1[3], t2[3];  
  PQP_REAL R1[3][3], R2[3][3];
  PQP_REAL rel_err= 0.0;
  PQP_REAL abs_err= 0.0;
  PQP_DistanceResult dres;

  M1= body1->pose();
  M2= body2->pose();

  t1[0]= M1.m14;
  t1[1]= M1.m24;
  t1[2]= M1.m34;
  R1[0][0]= M1.m11;      R1[0][1]= M1.m12;      R1[0][2]= M1.m13;
  R1[1][0]= M1.m21;      R1[1][1]= M1.m22;      R1[1][2]= M1.m23;
  R1[2][0]= M1.m31;      R1[2][1]= M1.m32;      R1[2][2]= M1.m33;

  t2[0]= M2.m14;
  t2[1]= M2.m24;
  t2[2]= M2.m34;
  R2[0][0]= M2.m11;      R2[0][1]= M2.m12;      R2[0][2]= M2.m13;
  R2[1][0]= M2.m21;      R2[1][1]= M2.m22;      R2[1][2]= M2.m23;
  R2[2][0]= M2.m31;      R2[2][1]= M2.m32;      R2[2][2]= M2.m33;

  PQP_Distance(&dres, R1, t1, body1->pqpModel(), R2, t2, body2->pqpModel(), rel_err, abs_err);
 
  p1_rel[0]= dres.p1[0];
  p1_rel[1]= dres.p1[1];
  p1_rel[2]= dres.p1[2];
  p2_rel[0]= dres.p2[0];
  p2_rel[1]= dres.p2[1];
  p2_rel[2]= dres.p2[2];

  p1[0]= R1[0][0]*p1_rel[0] + R1[0][1]*p1_rel[1] + R1[0][2]*p1_rel[2] + t1[0];
  p1[1]= R1[1][0]*p1_rel[0] + R1[1][1]*p1_rel[1] + R1[1][2]*p1_rel[2] + t1[1];
  p1[2]= R1[2][0]*p1_rel[0] + R1[2][1]*p1_rel[1] + R1[2][2]*p1_rel[2] + t1[2];

  p2[0]= R2[0][0]*p2_rel[0] + R2[0][1]*p2_rel[1] + R2[0][2]*p2_rel[2] + t2[0];
  p2[1]= R2[1][0]*p2_rel[0] + R2[1][1]*p2_rel[1] + R2[1][2]*p2_rel[2] + t2[1];
  p2[2]= R2[2][0]*p2_rel[0] + R2[2][1]*p2_rel[1] + R2[2][2]*p2_rel[2] + t2[2];


  return dres.distance;


}



//! Exports a body to a .pov file (for POVRAY ray tracer).
//! \param foldername folder where the exported file will be saved
//! \return 1 in case of success, 0 otherwise
int Body::exportMeshToPOVRay(std::string foldername)
{
  unsigned int i;
  size_t pos;
  std::string inc_name;
  FILE *file= NULL;

  inc_name= _name;
  pos= inc_name.find(".");
  while(pos!=std::string::npos)
  {
    inc_name.replace(pos, 1, "_");
    pos= inc_name.find(".");
  }

  inc_name= foldername + inc_name + ".inc";

  file= fopen(inc_name.c_str(), "w");
  if(file==NULL)
  {
    printf("%s: %d: Body::exportMeshToPOVRay(): can not open file \"%s\".\n",__FILE__,__LINE__, inc_name.c_str()); 
    return 0;
  }

//   fprintf(file, "#declare Pigment=\n");
//   fprintf(file, "\tpigment  { color rgb <%f, %f, %f> }\n", color.r, color.g, color.b);

//   fprintf(file, "#declare Finish=\n");
//   fprintf(file, "\tfinish { ambient 0.0\n");
//   fprintf(file, "\t         diffuse 0.6\n");
//   fprintf(file, "\t         specular 1.0\n");
//   fprintf(file, "\t         brilliance 1.0\n");
//   fprintf(file, "\t         roughness 0.01\n");
//   //fprintf(file, "\t         reflection 0.1\n");
// //  fprintf(file, "\t         metallic 0.9\n");
//   fprintf(file, "\t         reflection { 0.8 metallic }\n");
//   fprintf(file, "\t         conserve_energy\n");
//   fprintf(file, "\t         phong 1\n");
//   fprintf(file, "\t         phong_size 5\n");
//   fprintf(file, "\t       }\n");
// 
//    fprintf(file, "#declare default_material=\n");
//    fprintf(file, "\ttexture{ Pigment Finish }\n");


  fprintf(file, "#declare %s_mesh = mesh2 {\n", _name.c_str());

  fprintf(file, "\t vertex_vectors {\n");
  fprintf(file, "\t\t %d,\n", _mesh.nbVertices());
  for(i=0; i<_mesh.nbVertices(); ++i)
  {
    fprintf(file, "\t\t <%f, %f, %f>,\n", _mesh.vertices[i].x, _mesh.vertices[i].y, _mesh.vertices[i].z); 
  }
  fprintf(file, "\t}\n\n");
    
  if(_mesh.nbNormals()!=0)
  {
    fprintf(file, "\t normal_vectors {\n");
    fprintf(file, "\t\t %d,\n", _mesh.nbNormals());
    for(i=0; i<_mesh.nbNormals(); ++i)
    {
      fprintf(file, "\t\t <%f, %f, %f>,\n", _mesh.normals[i].x, _mesh.normals[i].y, _mesh.normals[i].z); 
    }
    fprintf(file, "\t}\n\n");
  }


//  fprintf(file, "\t texture_list { 1, texture{ default_material } }\n");

    
  fprintf(file, "\t face_indices {\n");
  fprintf(file, "\t\t %d,\n", _mesh.nbTriangles());
  for(i=0; i<_mesh.nbTriangles(); ++i)
  {
    fprintf(file, "\t\t <%d, %d, %d>,\n", _mesh.triangles[i].v[0], _mesh.triangles[i].v[1], _mesh.triangles[i].v[2]); 
  }
  fprintf(file, "\t}\n\n");
    

  if(_mesh.nbNormals()!=0)
  {
    fprintf(file, "\t normal_indices {\n");
    fprintf(file, "\t\t %d\n", _mesh.nbTriangles());
    for(i=0; i<_mesh.nbTriangles(); ++i)
    {
      fprintf(file, "\t\t <%d, %d, %d>,\n", _mesh.triangles[i].vn[0], _mesh.triangles[i].vn[1], _mesh.triangles[i].vn[2]); 
    }
    fprintf(file, "\t}\n\n");
  }


  fprintf(file, "}\n\n");
  
  fclose(file);

  return 1;
}

int Body::addToPOVfile(FILE *file)
{
  size_t pos;
  std::string inc_name;

  // Replace all the '.' characters by '_' the file names:
  inc_name= _name;
  pos= inc_name.find(".");
  while(pos!=std::string::npos)
  {
    inc_name.replace(pos, 1, "_");
    pos= inc_name.find(".");
  }
  fprintf(file, "#include \"%s.inc\"\n", inc_name.c_str());

  inc_name+= "_mesh";

  fprintf(file, "object { %s\n", inc_name.c_str());

  fprintf(file, "\t matrix < %f, %f, %f,\n",  _pose.m11, _pose.m21, _pose.m31);
  fprintf(file, "\t          %f, %f, %f,\n",  _pose.m12, _pose.m22, _pose.m32);
  fprintf(file, "\t          %f, %f, %f,\n",  _pose.m13, _pose.m23, _pose.m33);
  fprintf(file, "\t          %f, %f, %f >\n", _pose.m14, _pose.m24, _pose.m34);
  fprintf(file, "\t pigment { color rgb <%f, %f, %f> }\n", _color.r, _color.g, _color.b);
  fprintf(file, "\t material {");

  switch(material)
  {
    case PLASTIC:
       fprintf(file, " M_PLASTIC }\n");
    break;
    case SHINY_PLASTIC:
       fprintf(file, " M_SHINY_PLASTIC }\n");
    break;
    case METAL:
       fprintf(file, " M_METAL }\n");
    break;
    case GLASS:
       fprintf(file, " M_GLASS }\n");
    break;
    case CONCRETE:
       fprintf(file, " M_CONCRETE }\n");
    break;
  }
  fprintf(file, "}\n\n");    

  return 1;
}

