#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <iostream>
#include <sstream>
#include <string>
#include <vector>
#include <file_reader.h>
#include <robot.h>

//! Names of all the elements that are recognized by the file parser:
static char DEFINED_ELEMENTS[]= "body joint constraint robot name file models_directory robot_directory robot_file subrobot connection scale ht_matrix color texture material type box cylinder sphere bounding_box translation YPR_angles rotation_angle_and_axis dh_parameters q qmin qmax robot_configuration collision display shadow master slave master_dof slave_dof parameters";



//! This function is used like a printf but it writes the text in a file named "logfile".
//! The first time it is called, the previous content of the file is erased.
//! \param format format string + variables (like for a printf)
//! \return 1 in case of success, 0 otherwise 
int logfile(const char *format,...)
{
  static bool firstTime= true;
  FILE *file= NULL;

  if(firstTime)
  {
    firstTime= 0;
    file= fopen("logfile","w");
  }
  else
  {
    file= fopen("logfile","a+");
  }

  if(file==NULL)
  { 
    printf("%s: %d: logfile(): can not open file \"logfile\".\n", __FILE__, __LINE__);
    return 0;
  }

  va_list args;
  va_start(args,format);
  vfprintf(file,format,args);
  va_end(args);

  fclose(file); 

  return 1;
}

 
bool isElementDefined(std::string element)
{
  size_t pos1, pos2, start;
  std::string current;
  std::string element_list= DEFINED_ELEMENTS;
  
  start= 0;
  pos1= pos2= 0;
  while(pos1!=std::string::npos && pos2!=std::string::npos)
  {
    pos1= element_list.find_first_not_of(" ", start);
    pos2= element_list.find_first_of(" ", pos1+1);
    current= element_list.substr(pos1, pos2-pos1); 
    start= pos2+1;

    if(element==current)
    { return true; }
  }

  return false;
}


//! Converts the char string of the given node to a std::string after removing all newline characters
//! as well as space characters at the end of the string.
std::string getNodeString(xmlDocPtr doc, xmlNodePtr node)
{
  size_t pos;
  xmlChar *key;
  std::string text, message, word;
  key= xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
  text= (char *) key;
  xmlFree(key);

  //Remove all the newline characters for UNIX file (\n) at the end of the line:
  pos= text.find("\n");
  while(pos!=std::string::npos)
  {
    text.replace(pos, 1, " ");
    pos= text.find("\n");
  }

  //Remove all the newlines character for WINDOWS/DOS file (\r\n) at the end of the line:
  pos= text.find("\r");
  while(pos!=std::string::npos)
  {
    text.replace(pos, 1, " ");
    pos= text.find("\r");
  }

  //Replace all the tab characters with spaces:
  pos= text.find("\t");
  while(pos!=std::string::npos)
  {
    text.replace(pos, 1, " ");
    pos= text.find("\t");
  }

  //Remove all the spaces at the end of the line:
  pos= text.find_last_not_of(' ');
  text= text.substr(0, pos+1);



  return text;
}

void warningMessage(int line_number, const xmlChar *URL, const xmlChar *element_name, std::string &message)
{
   printf("At line %d in %s: warning: in an element of <%s>: \n", line_number, URL, element_name);
   printf(" %s\n\n", message.c_str());
   return;
}


void formatErrorMessage(int line_number, const xmlChar *URL, const xmlChar *element_name, std::string &message)
{
   printf("At line %d in %s: format error in the data of an element <%s>.\n", line_number, URL, element_name);
   printf(" %s\n\n", message.c_str());

   return;
}

void elementMissingMessage(int line_number, const xmlChar *URL, const xmlChar *element_name, std::string &message)
{
   printf("At line %d in %s: A sub-element of element <%s> is missing.\n", line_number, URL, element_name);
   printf(" %s\n\n", message.c_str());

   return;
}


//! Reads the content of the given node and fills the appropriate field in the ElementParserData structure.
//! This function is used for simple element (with no chidren nodes).
//! \param doc pointer to the structure used by libxml2 to parse the XML document
//! \param entry_node the node whose children will be browsed
//! \param element the name of the element that will be looked for among the entry node children
//! \param data the data structure that will be used to store the read information
//! \return true if the searched element was found, false otherwise
bool parseElement(xmlDocPtr doc, xmlNodePtr entry_node, std::string element, ElementParserData &data)
{
  bool result;
  float x;
  xmlNodePtr cur= entry_node->xmlChildrenNode;
  std::string text, message, word, word2;
  std::istringstream iss;


  for(cur= entry_node->xmlChildrenNode; cur!=NULL; cur= cur->next)
  {
    if( !xmlStrcmp(cur->name, (xmlChar*) element.c_str() ) )
    {
       iss.str(getNodeString(doc, cur));

       if(element=="name")
       { 
         result= (iss >> data.name );
         if( !result || !iss.eof() )
         {
           message= "Usage: <name> name </name>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="file")
       {
         result= (iss >> data.file );
         if( !result || !iss.eof() )
         {
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="models_directory")
       {
         result= (iss >> data.models_directory );
         if( !result || !iss.eof() )
         {
           message= "Usage: <models_directory> name of the directory that contains the 3D models of the robot bodies </models_directory>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="robot_directory")
       {
         result= (iss >> data.robot_directory );
         if( !result || !iss.eof() )
         {
           message= "Usage: <robot_directory> name of the directory that contains the XML file describing the robot </robot_directory>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="robot_file")
       {
         result= (iss >> data.robot_file );
         if( !result || !iss.eof() )
         {
           message= "Usage: <robot_file> name of the XML file describing the robot </robot_file>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="subrobot")
       {
         result= (iss >> data.subrobot );
         if( !result || !iss.eof() )
         {
           message= "Usage: <subrobot> name of the XML file describing a sub-robot of the current robot </subrobot>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="connection")
       {
         result= (iss >> data.connection );
         if( !result || !iss.eof() )
         {
           message= "Usage: <connection> name of the robot's joint the sub-robot will be connected to </connection>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="scale")
       {
         result= (iss >> data.scale );
         if( !result || !iss.eof() )
         {
           message= "Usage: <scale> scale </scale>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="ht_matrix")
       { 
         result= (iss >> data.ht_matrix.m11 >> data.ht_matrix.m12 >> data.ht_matrix.m13 >> data.ht_matrix.m14 >> data.ht_matrix.m21 >> data.ht_matrix.m22 >> data.ht_matrix.m23 >> data.ht_matrix.m24 >> data.ht_matrix.m31 >> data.ht_matrix.m32 >> data.ht_matrix.m33 >> data.ht_matrix.m34 );

         if( !result || !iss.eof() )
         {
           message= "Usage: <ht_matrix> R11 R12 R13 tx R21 R22 R23 ty R31 R32 R33 tz </ht_matrix>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="color")
       {
         result= (iss >> data.color[0] >> data.color[1] >> data.color[2] >> data.color[3]);
         if( !result || !iss.eof() )
         {
           message= "Usage: <color> red green blue alpha </color>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="texture")
       {
         result= (iss >> data.texture);
         if( !result || !iss.eof() )
         {
           message= "Usage: <texture> name of the texture to apply to the body  </texture>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="material")
       {
         result= (iss >> data.material_type);
         if( !result || !iss.eof() )
         {
           message= "Usage: <material> name of the body's material  </material>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="type")
       {
         result= (iss >> data.type );
         if( !result || !iss.eof() )
         {
           message= "Usage: <type> joint or constraint type </type>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="box")
       {
         result= (iss >> data.dimX >>  data.dimY >>  data.dimZ );
         if( !result || !iss.eof() )
         {
           message= "Usage: <box> dimX dimY dimZ </box>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="cylinder")
       {
         result= (iss >> data.radius >>  data.height );
         if( !result || !iss.eof() )
         {
           message= "Usage: <cylinder> radius height </cylinder>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="sphere")
       {
         result= (iss >> data.radius );
         if( !result || !iss.eof() )
         {
           message= "Usage: <sphere> radius </sphere>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="bounding_box")
       {
         result= (iss >> data.bounding_box[0] >>  data.bounding_box[1] >>  data.bounding_box[2] >>  data.bounding_box[3] >>  data.bounding_box[4] >>  data.bounding_box[5]);
         if( !result || !iss.eof() )
         {
           message= "Usage: <bounding_box> xmin xmax ymin ymax zmin zmax </bounding_box>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }


       if(element=="translation")
       {
         result= (iss >> data.translation.x >> data.translation.y >> data.translation.z );
         if( !result || !iss.eof() )
         {
           message= "Usage: <translation> x y z </translation>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }

       if(element=="YPR_angles")
       {
         result= (iss >> data.YPR_angles[0] >> data.YPR_angles[1] >> data.YPR_angles[2] );
         if( !result || !iss.eof() )
         {
           message= "Usage: <YPR_angles> ax ay az </YPR_angles>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         data.YPR_angles[0]*= DEGTORAD;
         data.YPR_angles[1]*= DEGTORAD;
         data.YPR_angles[2]*= DEGTORAD;
         return true;
       }

       if(element=="rotation_angle_and_axis")
       {
         result= (iss >> data.angle_axis[0] >> data.angle_axis[1] >> data.angle_axis[2] >> data.angle_axis[3] );
         if( !result || !iss.eof() )
         {
           message= "Usage: <rotation_angle_and_axis> angle ux uy uz </rotation_angle_and_axis>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         data.angle_axis[0]*= DEGTORAD;
         return true;
       }

       if(element=="dh_parameters")
       {
         result= (iss >> data.dh_parameters[0] >> data.dh_parameters[1] >> data.dh_parameters[2] >> data.dh_parameters[3] );
         if( !result || !iss.eof() )
         {
           message= "Usage: <dh_parameters> theta d a alpha </dh_parameters>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         data.dh_parameters[1]*= DEGTORAD;
         data.dh_parameters[3]*= DEGTORAD;
         return true;
       }

       if(element=="q")
       {
          data.q.clear();
          while(!iss.eof())
          {
            result= (iss >> x );
            if(!result)
            {
              message= "Usage: <q> joint parameters (angles are given in degrees, lengths in meters) </q>.";
              formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
              return false;
            } 
            else
            {
               data.q.push_back(x);  
            }
          }
       }

       if(element=="qmin")
       {
          data.qmin.clear();
          while(!iss.eof())
          {
            result= (iss >> x );
            if(!result)
            {
              message= "Usage: <qmin> joint minimal bounds (angles are given in degrees, lengths in meters) </qmin>.";
              formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
              return false;
            } 
            else
            {
               data.qmin.push_back(x);  
            }
          }
       }

       if(element=="qmax")
       {
          data.qmax.clear();
          while(!iss.eof())
          {
            result= (iss >> x );
            if(!result)
            {
              message= "Usage: <qmax> joint minimal bounds (angles are given in degrees, lengths in meters) </qmax>.";
              formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
              return false;
            } 
            else
            {
               data.qmax.push_back(x);  
            }
          }
       }

       if(element=="robot_configuration")
       {
          data.robot_configuration.clear();
          while(!iss.eof())
          {
            result= (iss >> x );
            if(!result)
            {
              message= "Usage: <robot_configuration> robot joint parameters (angles are given in degrees, lengths in meters) </robot_configuration>.";
              formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
              return false;
            } 
            else
            {
               data.robot_configuration.push_back(x);  
            }
          }
       }

       if(element=="collision")
       {
         result= (iss >> word );
         if( !result || !iss.eof() )
         {
           message= "Usage: <collision> \"yes\" or \"no\" </collision>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         if(word=="yes")
         { 
           data.enable_collision= true;
           return true;
         }
         if(word=="no")
         { 
           data.enable_collision= false;
           return true;
         }
         message= "Usage: <collision> \"yes\" or \"no\" </collision>.";
         formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
         return false;
       }     

       if(element=="display")
       {
         result= (iss >> word );
         if( !result || !iss.eof() )
         {
           message= "Usage: <display> \"yes\" or \"no\" </display>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         if(word=="yes")
         { 
           data.enable_display= true;
           return true;
         }
         if(word=="no")
         { 
           data.enable_display= false;
           return true;
         }
         message= "Usage: <display> \"yes\" or \"no\" </display>.";
         formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
         return false;
       }     

       if(element=="shadow")
       {
         result= (iss >> word );
         if( !result || !iss.eof() )
         {
           message= "Usage: <shadow> \"yes\" or \"no\" </shadow>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         if(word=="yes")
         { 
           data.enable_shadow= true;
           return true;
         }
         if(word=="no")
         { 
           data.enable_shadow= false;
           return true;
         }
         message= "Usage: <shadow> \"yes\" or \"no\" </shadow>.";
         formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
         return false;
       }  

       if(element=="master")
       {
         result= (iss >> data.master );
         if( !result || !iss.eof() )
         {
           message= "Usage: <master> name of the master joint in a joint constraint </master>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }  

       if(element=="slave")
       {
         result= (iss >> data.slave );
         if( !result || !iss.eof() )
         {
           message= "Usage: <slave> name of the slave joint in a joint constraint </slave>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }  

       if(element=="master_dof")
       {
         result= (iss >> data.master_dof );
         if( !result || !iss.eof() )
         {
           message= "Usage: <master_dof> index of the DOF of the master joint that will constrain one of the slave joint DOFs </master_dof>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       }  
       if(element=="slave_dof")
       {
         result= (iss >> data.slave_dof );
         if( !result || !iss.eof() )
         {
           message= "Usage: <slave_dof> index of the DOF of the slave joint that will be constrained by one of the master joint DOFs </slave_dof>.";
           formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
           return false;
         } 
         return true;
       } 

       if(element=="parameters")
       {
          data.parameters.clear();
          while(!iss.eof())
          {
            result= (iss >> x );
            if(!result)
            {
              message= "Usage: <parameters> constraint parameters </parameters>.";
              formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
              return false;
            } 
            else
            {
               data.parameters.push_back(x);  
            }
          }
       }

       return true;
    }

  }


  return false;
}


//! Reads the content of a <transformation> node and fills the TransformationParserData structure.
//! \param doc pointer to the structure used by libxml2 to parse the XML document
//! \param entry_node the node whose children will be browsed
//! \param data the data structure that will be used to store all the read information about the transformation
//! \return true if all the mandatory elements were found, false otherwise
bool parseTransformation(xmlDocPtr doc, xmlNodePtr entry_node, TransformationParserData &data)
{
  bool has_translation, has_rotation;
  ElementParserData elementData;
  std::string message;

  has_translation= false;
  has_rotation= false;

  data.matrix.setIdentity();

  if(parseElement(doc, entry_node, "ht_matrix", elementData))
  {
    data.matrix= elementData.ht_matrix;
    return true;
  }  

  if(parseElement(doc, entry_node, "dh_parameters", elementData))
  {
    data.matrix.setFromDHparameters(elementData.dh_parameters[0], elementData.dh_parameters[1], elementData.dh_parameters[2], elementData.dh_parameters[3]);
    return true;
  }  

  if(parseElement(doc, entry_node, "translation", elementData))
  {
    has_translation= true; 
    data.matrix.setTranslation(elementData.translation.x, elementData.translation.y, elementData.translation.z);
  }

  if(parseElement(doc, entry_node, "YPR_angles", elementData))
  {
    data.matrix.setRotationFromYPRangles(elementData.YPR_angles[0], elementData.YPR_angles[1], elementData.YPR_angles[2]);
    has_rotation= true;
  }

  if(parseElement(doc, entry_node, "rotation_angle_and_axis", elementData))
  {
    data.matrix.setRotationFromAngleAndAxis(elementData.angle_axis[0], elementData.angle_axis[1], elementData.angle_axis[2], elementData.angle_axis[3]);
    has_rotation= true;
  }

  if(!has_translation && !has_rotation)
  {  
    message= "Element <transformation> should have at least a valid sub-element (<ht_matrix>, <dh_parameters>, <translation>, <rotation_angle_and_axis> or <YPR_angles>).";
    elementMissingMessage((int) xmlGetLineNo(entry_node), doc->URL, entry_node->name, message);
    return false;
  }

  return true;
}


//! Reads the content of a <deactivate_collision_pair> node and fills the TransformationParserData structure.
//! \param doc pointer to the structure used by libxml2 to parse the XML document
//! \param entry_node the node whose children will be browsed
//! \param bodyName1 where to store the name of first body
//! \param bodyName2 where to store the name of second body
//! \return true if all the mandatory elements were found, false otherwise
bool parseCollisionPair(xmlDocPtr doc, xmlNodePtr entry_node, std::string &bodyName1, std::string &bodyName2)
{
  bool result;
  std::string message, word1, word2;
  std::istringstream iss;


  if( !xmlStrcmp(entry_node->name, (xmlChar*) "deactivate_collision_pair" ) )
  {
     iss.str(getNodeString(doc, entry_node));

      result= (iss >> word1 >> word2 );
      if(!result)
      {
        message= "Usage: <deactivate_collision_pair> name_of_first_body name_of_second_body </deactivate_collision_pair>.";
        formatErrorMessage((int) xmlGetLineNo(entry_node), doc->URL, entry_node->name, message);
        return false;
      } 
      else
      {
        bodyName1= word1;
        bodyName2= word2; 
        return true;
      }
   }

   return false;
}


//! Reads the content of a <subrobot> node and fills the SubrobotParserData structure.
//! \param doc pointer to the structure used by libxml2 to parse the XML document
//! \param entry_node the node whose children will be browsed
//! \param data the data structure that will be used to store all the read information about the subrobot
//! \return true if all the mandatory elements were found, false otherwise
bool parseSubrobot(xmlDocPtr doc, xmlNodePtr entry_node, SubrobotParserData &data)
{
  xmlNodePtr cur= entry_node;
  ElementParserData elementData;
  TransformationParserData transformationData;
  std::string message;

  if(!parseElement(doc, entry_node, "robot_file", elementData)) 
  {  
    message= "Element <subrobot> should have an element <robot_file>.";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false;
  }  
  else
  {
    data.robot_file= elementData.robot_file;
  }

  //Read the name of the joint the subrobot will be connected to:
  if(!parseElement(doc, entry_node, "connection", elementData)) 
  {  
    message= "Element <subrobot> should have an element <connection>.";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false;
  }  
  else
  {
    data.connection= elementData.connection;
  }


  //Read the subrobot's transformation (pose):
  data.T.setIdentity(); //default value is identity
  for(cur= entry_node->xmlChildrenNode; cur!=NULL; cur= cur->next)
  {
    if((!xmlStrcmp(cur->name, (const xmlChar *)"transformation")))
    { 
      if(parseTransformation(doc, cur, transformationData))
      {
        data.T= transformationData.matrix;
        break;
      }
    }
  }

  if(!parseElement(doc, entry_node, "collision", elementData)) 
  {  
    data.enable_collision= true;
  }  
  else
  {
    data.enable_collision= elementData.enable_collision;
  }

  if(!parseElement(doc, entry_node, "display", elementData)) 
  {  
    data.enable_display= true;
  }  
  else
  {
    data.enable_display= elementData.enable_display;
  }


  return true;
}


//! Reads the content of a <body> node and fills the BodyParserData structure.
//! \param doc pointer to the structure used by libxml2 to parse the XML document
//! \param entry_node the node whose children will be browsed
//! \param data the data structure that will be used to store all the read information about the body
//! \return true if all the mandatory elements were found, false otherwise
bool parseBody(xmlDocPtr doc, xmlNodePtr entry_node, BodyParserData &data)
{
  xmlNodePtr cur= entry_node;
  ElementParserData elementData;
  TransformationParserData transformationData;
  std::string message;

  if(!parseElement(doc, entry_node, "name", elementData)) 
  {  
    message= "Element <body> should have an element <name>.";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false;
  }  
  else
  {
    data.name= elementData.name;
  }

  if(parseElement(doc, entry_node, "file", elementData)) 
  {  
    data.is_primitive= false;
    data.file= elementData.file;
  }  
  else if(parseElement(doc, entry_node, "box", elementData)) 
  {   
    data.is_primitive= true;
    data.primitive_type= BOX;    
    data.dimX= elementData.dimX;
    data.dimY= elementData.dimY;
    data.dimZ= elementData.dimZ;
  }
  else if(parseElement(doc, entry_node, "cylinder", elementData))
  {
    data.is_primitive= true;
    data.primitive_type= CYLINDER;    
    data.radius= elementData.radius;
    data.height= elementData.height;
  }
  else if(parseElement(doc, entry_node, "sphere", elementData))
  {
    data.is_primitive= true;
    data.primitive_type= SPHERE;    
    data.radius= elementData.radius;
  }
  else
  {
    message= "Element <body> should have an element to define its shape (<file>, <box>, or <cylinder>).";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false;
  }

  if(!parseElement(doc, entry_node, "scale", elementData)) 
  {  
    data.scale= 1.0;
  }  
  else
  {  
    data.scale= elementData.scale;
  }  

  //Read the body's transformation (pose):
  data.T.setIdentity(); //default value is identity
  for(cur= entry_node->xmlChildrenNode; cur!=NULL; cur= cur->next)
  {
    if((!xmlStrcmp(cur->name, (const xmlChar *)"transformation")))
    { 
      if(parseTransformation(doc, cur, transformationData))
      {
        data.T= transformationData.matrix;
        break;
      }
    }
  }
 
  if(!parseElement(doc, entry_node, "color", elementData)) 
  {  
    data.color[0]= 1.0;
    data.color[1]= 1.0;
    data.color[2]= 0.0;
    data.color[3]= 1.0;
  }  
  else
  {  
    data.color[0]= elementData.color[0];
    data.color[1]= elementData.color[1];
    data.color[2]= elementData.color[2];
    data.color[3]= elementData.color[3];
  }  

  if(parseElement(doc, entry_node, "texture", elementData)) 
  {  
    data.texture= elementData.texture;
  }  
  else
  {
    data.texture= "none";
  }

  if(parseElement(doc, entry_node, "material", elementData)) 
  {  
    if(elementData.material_type=="plastic")
    {  data.material= PLASTIC;  }
    else if(elementData.material_type=="shiny_plastic")
    {  data.material= SHINY_PLASTIC;  }
    else if(elementData.material_type=="metal")
    {  data.material= METAL;    }
    else if(elementData.material_type=="glass")
    {  data.material= GLASS;    }
    else if(elementData.material_type=="concrete")
    {  data.material= CONCRETE; }
    else
    {
      message= "Usage: <material> name of the body's material (plastic, shiny_plastic, metal, glass or concrete) </material>.";
      formatErrorMessage((int) xmlGetLineNo(entry_node), doc->URL, entry_node->name, message);
    }
  }
  else
  {
    data.material= PLASTIC;
  }


  if(!parseElement(doc, entry_node, "collision", elementData))
  {  
    data.enable_collision= true;
  }  
  else
  {  
    data.enable_collision= elementData.enable_collision;
  } 

  if(!parseElement(doc, entry_node, "display", elementData))
  {  
    data.enable_display= true;
  }  
  else
  {  
    data.enable_display= elementData.enable_display;
  } 


  if(!parseElement(doc, entry_node, "shadow", elementData))
  {  
    data.enable_shadow= true;
  }  
  else
  {  
    data.enable_shadow= elementData.enable_shadow;
  } 


  return true;
}

//! Reads the content of a <constraint> node and fills the ConstraintParserData structure.
//! \param doc pointer to the structure used by libxml2 to parse the XML document
//! \param entry_node the node whose children will be browsed
//! \param data the data structure that will be used to store all the read information about the constraint
//! \return true if all the mandatory elements were found, false otherwise
bool parseConstraint(xmlDocPtr doc, xmlNodePtr entry_node, ConstraintParserData &data)
{
  xmlNodePtr cur= entry_node;
  ElementParserData elementData;
  std::string message;

  if(!parseElement(doc, entry_node, "type", elementData)) 
  {  
    message= "Element <constraint> should have an element <type>.";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false;
  }  
  else
  {
    if(elementData.type=="linear")
    { 
      data.type= LINEAR;
    }   
    else
    {
      message= "Unknown constraint type: " + elementData.type + ".";
      formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
      return false;
    }
  }

  if(!parseElement(doc, entry_node, "master", elementData)) 
  {  
    message= "Element <constraint> should have an element <master>.";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false; 
  }  
  else
  {
    data.master= elementData.master;
  }

  if(!parseElement(doc, entry_node, "slave", elementData)) 
  {  
    message= "Element <constraint> should have an element <slave>.";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false; 
  }  
  else
  {
    data.slave= elementData.slave;
  }

  if(!parseElement(doc, entry_node, "master_dof", elementData)) 
  {  
    message= "Element <constraint> should have an element <master_dof>.";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false; 
  }  
  else
  {
    data.master_dof= elementData.master_dof;
  }

  if(!parseElement(doc, entry_node, "slave_dof", elementData)) 
  {  
    message= "Element <constraint> should have an element <slave_dof>.";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false; 
  }  
  else
  {
    data.slave_dof= elementData.slave_dof;
  }

  if(!parseElement(doc, entry_node, "parameters", elementData)) 
  {  
    message= "Element <constraint> should have an element <parameters>.";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false;
  }       
  else
  {
    switch(data.type)
    {
      case LINEAR:
        if(elementData.parameters.size()!=2)
        {
          message= "Usage: <parameters> (linear relationship coefficient)  offset </parameters>.";
          formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
          return false;
        }
      break;
    }
    data.parameters= elementData.parameters;
  }


  return true;
}

//! Reads the content of a <joint> node and adds the resulting joint to the given robot.
//! NB: As a joint node may have chidren nodes (joint or body), this function is recursive.
//! \param doc pointer to the structure used by libxml2 to parse the XML document
//! \param entry_node the node whose children will be browsed
//! \param robot the robot the joint will be added to
//! \return true if all the mandatory elements were found and the joint added to the robot, false otherwise
bool parseJoint(xmlDocPtr doc, xmlNodePtr entry_node, Robot *robot)
{
  unsigned int i;
  xmlNodePtr cur= entry_node;
  std::string message, element;
  ElementParserData elementData;
  TransformationParserData transformationData;
  BodyParserData bodyData;
  JointType type;
  std::string joint_name, previous_joint;
  std::vector<float> qmin, qmax, q;
  HTMatrix T, previous_frame;
  BodyType bodyType;
  Body *body= NULL;

  if(!parseElement(doc, entry_node, "name", elementData)) 
  {  
    message= "Element <joint> should have an element <name>.";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false;
  }  
  else
  {
    joint_name= elementData.name;
  }

  if(!parseElement(doc, entry_node, "type", elementData)) 
  {  
    message= "Element <joint> should have an element <type>.";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false;
  }  
  else
  {
    if(elementData.type=="base")
    { 
      type= BASE;
    }   
    else if(elementData.type=="freeflyer")
    { 
      type= FREEFLYER;
    }   
    else  if(elementData.type=="revolute")
    { 
      type= REVOLUTE;
    }   
    else  if(elementData.type=="s_revolute")
    { 
      type= S_REVOLUTE;
    }   
    else if(elementData.type=="prismatic")
    { 
      type= PRISMATIC;
    }   
    else
    {
      message= "Unknown joint type: " + elementData.type + ".";
      formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
      return false;
    }
  }

  //Read the joint's transformation relative to previous joint:
  T.setIdentity(); // default value
  for(cur= entry_node->xmlChildrenNode; cur!=NULL; cur= cur->next)
  {
    if((!xmlStrcmp(cur->name, (const xmlChar *)"transformation")))
    {
      if(parseTransformation(doc, cur, transformationData))
      {
        T= transformationData.matrix;
        break;
      }
    }    
  }

  switch(type)
  {
    case BASE:
      q.resize(3);
      qmin.resize(2);
      qmax.resize(2);
    break;
    case FREEFLYER:
      q.resize(6);
      qmin.resize(3);
      qmax.resize(3);
    break;
    case REVOLUTE: case S_REVOLUTE: case PRISMATIC:
      q.resize(1);
      qmin.resize(1);
      qmax.resize(1);
    break;
  }


  if(!parseElement(doc, entry_node, "qmin", elementData)) 
  {  
    message= "Element <joint> should have an element <qmin>.";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false; 
  }       
  else
  {
    switch(type)
    {
      case BASE:
        if(elementData.qmin.size()!=2)
        {
          message= "Usage (for base joints): <qmin> Xmin Ymin </qmin>.";
          element= "<qmin>";
          formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
          return false;
        }
      break;
      case FREEFLYER:
        if(elementData.qmin.size()!=3)
        {
          message= "Usage (for freeflyer joints): <qmin> Xmin Ymin Zmin </qmin>.";
          element= "<qmin>";
          formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
          return false;
        }
      break;
      case REVOLUTE: case S_REVOLUTE: case PRISMATIC:
        if(elementData.qmin.size()!=1)
        {
          message= "Usage (for revolute, s_revolute or prismatic joints): <qmin> minimal bound on joint value </qmin>.";
          element= "<qmin>";
          formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
          return false;
        }
      break;
    }
    qmin= elementData.qmin;
  }

 
  if(!parseElement(doc, entry_node, "qmax", elementData)) 
  {  
    message= "Element <joint> should have an element <qmax>.";
    elementMissingMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
    return false;
  }       
  else
  {
    switch(type)
    {
      case BASE:
        if(elementData.qmax.size()!=2)
        {
          message= "Usage (for base joints): <qmax> Xmax Ymax </qmax>.";
          element= "<qmax>";
          formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
          return false;
        }
      break;
      case FREEFLYER:
        if(elementData.qmax.size()!=3)
        {
          message= "Usage (for freeflyer joints): <qmax> Xmax Ymax Zmax </qmax>.";
          element= "<qmax>";
          formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
          return false;
        }
      break;
      case REVOLUTE: case S_REVOLUTE: case PRISMATIC:
        if(elementData.qmax.size()!=1)
        {
          message= "Usage (for revolute, s_revolute or prismatic joints): <qmax> maximal bound on joint value </qmax>.";
          element= "<qmax>";
          formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
          return false;
        }
      break;
    }
    qmax= elementData.qmax;
  }

  if(!parseElement(doc, entry_node, "q", elementData)) 
  {  
    for(i=0; i<q.size(); ++i)
    {
      q[i]= 0;
    }
  }       
  else
  {
    switch(type)
    {
      case BASE:
        if(elementData.q.size()!=3)
        {
          message= "Usage (for base joints): <q> x y theta </q>.";
          element= "<q>";
          formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
          return false;
        }
      break;
      case FREEFLYER:
        if(elementData.q.size()!=6)
        {
          message= "Usage (for freeflyer joints): <q> x y z alpha beta gamma </q>.";
          element= "<q>";
          formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
          return false;
        }
      break;
      case REVOLUTE: case S_REVOLUTE: case PRISMATIC:
        if(elementData.q.size()!=1)
        {
          message= "Usage (for revolute, s_revolute or prismatic joints): <q> current value of joint parameter </q>.";
          element= "<q>";
          formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, entry_node->name, message);
          return false;
        }
      break;
    }
    q= elementData.q;
  }

  if(entry_node->parent!=NULL)
  {
    if(xmlStrcmp(entry_node->parent->name, (const xmlChar *)"joint")==0)
    {
      if(!parseElement(doc, entry_node->parent, "name", elementData))
      {
         previous_joint= "none";
      }
      else
      {
         previous_joint= elementData.name;
      }
    } 
  }
  
  robot->addJoint(joint_name, type, qmin, qmax, T, previous_joint);

  for(i=0; i<robot->joints.back()->q.size(); ++i)
  {
    robot->joints.back()->q[i].setValue(q[i]);
  }

  for(cur= entry_node->xmlChildrenNode; cur!=NULL; cur= cur->next)
  {
    if((!xmlStrcmp(cur->name, (const xmlChar *)"body")))
    {
      if(parseBody(doc, cur, bodyData))
      {    
        if(bodyData.enable_display==true)
        {
          if(bodyData.enable_collision==true)
          {  bodyType= DISPLAY_AND_COLLISION;  }
          else
          {  bodyType= DISPLAY_ONLY;  }
        }
        else
        {
          if(bodyData.enable_collision==true)
          {  bodyType= COLLISION_ONLY;  }
          else
          {  bodyType= NO_DISPLAY_NO_COLLISION;  }
        }

        if(bodyData.is_primitive==false)
        {
          robot->addBody(bodyData.name, bodyType, bodyData.file, bodyData.scale, joint_name, bodyData.T);
        }
        else
        {
          switch(bodyData.primitive_type)
          {
            case BOX:
              body= new Box(bodyData.name, bodyType, bodyData.dimX, bodyData.dimY, bodyData.dimZ);
              robot->addBody(body, joint_name, bodyData.T);
            break;
            case CYLINDER:
              body= new Cylinder(bodyData.name, bodyType, bodyData.radius, bodyData.height);
              robot->addBody(body, joint_name, bodyData.T);
            break;
            case SPHERE:
              body= new Sphere(bodyData.name, bodyType, bodyData.radius);
              robot->addBody(body, joint_name, bodyData.T);
            break;
          }
        }
        robot->bodies.back()->setColor(bodyData.color[0], bodyData.color[1], bodyData.color[2], bodyData.color[3]);
        robot->bodies.back()->setMaterial(bodyData.material);
        robot->bodies.back()->enable_collision= bodyData.enable_collision;
        robot->bodies.back()->enable_display  = bodyData.enable_display;
        robot->bodies.back()->enable_shadow   = bodyData.enable_shadow;
        robot->bodies.back()->setTexture(bodyData.texture);
        robot->bodies.back()->enable_texture  = false;
      }
    }
  }

  for(cur= entry_node->xmlChildrenNode; cur!=NULL; cur= cur->next)
  {
    if((!xmlStrcmp(cur->name, (const xmlChar *)"joint")))
    {
      parseJoint(doc, cur, robot);
    }
  }

  return true;
}


//! Parses an XML file describing a robot.
//! \param filename name of the XML file
//! \param robot pointer to the robot (that has already been created but is still empty)
//! \return 1 in case of success, 0 otherwise
int parseRobotFile(std::string path, std::string filename, Robot *robot)
{
  xmlDocPtr doc;
  xmlNodePtr root, cur;
  std::string text, message;  
  ElementParserData elementData;
  ConstraintParserData constraintData;
  SubrobotParserData subrobotData;
  std::string bodyName1, bodyName2;
  Body *body1= NULL, *body2= NULL;
  Robot *subrobot= NULL;

  xmlLineNumbersDefault(1);
  doc= xmlParseFile(filename.c_str());
	
  if(doc==NULL)
  {
    fprintf(stderr, "Document \"%s\" not parsed successfully.\n", filename.c_str());
    return 0;
  }
	
  root = xmlDocGetRootElement(doc);
	
  if(root==NULL)
  {
    fprintf(stderr, "Document \"%s\" is empty.\n", filename.c_str());
    xmlFreeDoc(doc);
    return 0;
  }
	
  if(xmlStrcmp(root->name, (const xmlChar *) "robot"))
  {
    fprintf(stderr, "Document \"%s\" is of the wrong type, root node != <robot>.\n", filename.c_str());
    xmlFreeDoc(doc);
    return 0;
  }
	
//   //Look for undefined elements:
//   for(cur= root->xmlChildrenNode; cur!=NULL; cur= cur->next)
//   {
//     text.assign((char *) cur->name);
//     if(text!="text" && !isElementDefined(text))
//     { 
//       message= "warning: element \"";
//       message+= text;
//       message+= "\" is not defined.";
//       formatErrorMessage((int) xmlGetLineNo(cur), doc->URL, root->name, message);
//     }
//   }  

  //Read the robot name:
  if(!parseElement(doc, root, "name", elementData)) 
  {  
    message= "Element <robot> should have an element <name>.";
    std::cout << message << std::endl;
  }
  else
  {
    robot->rename(elementData.name);
  }

  //Read the directory where to find the models if they are not given in absolute path:
  if(parseElement(doc, root, "models_directory", elementData)) 
  {
    robot->models_directory= path + "/" + elementData.models_directory;
  }


  //Read all the joints:
  for(cur= root->xmlChildrenNode; cur!=NULL; cur= cur->next)
  {
    if((!xmlStrcmp(cur->name, (const xmlChar *)"joint")))
    { 
      parseJoint(doc, cur, robot);
    }
  }

  //Read the possible sub-robots:
  for(cur= root->xmlChildrenNode; cur!=NULL; cur= cur->next)
  {
    if((!xmlStrcmp(cur->name, (const xmlChar *)"subrobot")))
    { 
      if(parseSubrobot(doc, cur, subrobotData))
      { 
        if(subrobotData.robot_file==robot->file())
        {
          message= "Subrobot files must not be the currently parsed robot file. The file parser would loop with no end.";
          warningMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
        }
        else
        {
          subrobot= new Robot();
          subrobot->loadFile(path,subrobotData.robot_file);
          subrobot->enableDisplay(subrobotData.enable_display);
          subrobot->enableCollision(subrobotData.enable_collision);
          robot->connectRobot(subrobotData.connection, subrobotData.T, subrobot);
          delete subrobot;
          subrobot= NULL;
        }
      }
    }
  }


  //Read the constraints:
  for(cur= root->xmlChildrenNode; cur!=NULL; cur= cur->next)
  {
    if((!xmlStrcmp(cur->name, (const xmlChar *)"constraint")))
    { 
      if(parseConstraint(doc, cur, constraintData))
      {
        robot->addConstraint(constraintData.master, constraintData.slave, constraintData.type, constraintData.master_dof, constraintData.slave_dof, constraintData.parameters);
      }
    }
  }

  //Read the collision pairs to deactivate:
  for(cur= root->xmlChildrenNode; cur!=NULL; cur= cur->next)
  {
    if((!xmlStrcmp(cur->name, (const xmlChar *)"deactivate_collision_pair")))
    { 
      if(parseCollisionPair(doc, cur, bodyName1, bodyName2)==false)
      {  continue; }

      body1= robot->getBodyByName(bodyName1);
      body2= robot->getBodyByName(bodyName2);

      if(body1==NULL)
      {
        message= "The robot has no body named \"" + bodyName1 + "\".";
        warningMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
      }
      else if(body2==NULL)
      {
        message= "The robot has no body named \"" + bodyName2 + "\".";
        warningMessage((int) xmlGetLineNo(cur), doc->URL, cur->name, message);
      }
      else
      {
        robot->deactivateCollisionPair(body1, body2);
      }
    }
  }

  xmlFreeDoc(doc);
  return 1;
}


//! Parses an XML file describing the world.
//! \param filename name of the XML file
//! \param robot pointer to the world (that has already been created but is still empty)
//! \return 1 in case of success, 0 otherwise
int parseWorldFile(std::string path, std::string filename, World *world)
{
  unsigned int i;
  xmlDocPtr doc;
  xmlNodePtr root, cur;
  std::string text, message, robot_file;  
  ElementParserData elementData;
  BodyParserData bodyData;
  DOFunit unit;
  BodyType bodyType;
  Body *body= NULL;

  xmlLineNumbersDefault(1);
  doc= xmlParseFile(filename.c_str());

  if(doc==NULL)
  {
    fprintf(stderr, "Document \"%s\" was not parsed successfully.\n", filename.c_str());
    return 0;
  }
	
  root = xmlDocGetRootElement(doc);
	
  if(root==NULL)
  {
    fprintf(stderr, "Document \"%s\" is empty.\n", filename.c_str());
    xmlFreeDoc(doc);
    return 0;
  }
	
  if(xmlStrcmp(root->name, (const xmlChar *) "world"))
  {
    fprintf(stderr, "Document \"%s\" is of the wrong type, root node != <world>.\n", filename.c_str());
    xmlFreeDoc(doc);
    return 0;
  }
	

  //Read the directory where to find the robot description file if it is not given in absolute path:
  if(parseElement(doc, root, "robot_directory", elementData)) 
  {
    world->robot_directory= path + "/" + elementData.robot_directory;
  }

  //Read the robot:
  if(!parseElement(doc, root, "robot_file", elementData)) 
  {
    message= "Element <world> should have an element <robot_file>.";
    elementMissingMessage((int) xmlGetLineNo(root), doc->URL, root->name, message);
    xmlFreeDoc(doc);
    return 0;
  }
  else
  { 
    robot_file=  elementData.robot_file;
    world->robot.loadFile(path, robot_file.c_str());
/*
    if(world->robot_directory.empty())
    {
      world->robot_directory= "./";
    }

    if(elementData.robot_file[0]!='/') //if the file is not given in absolute path
    {
      if(world->robot_directory.at(world->robot_directory.size()-1)!='/')
      {   world->robot_directory+= '/';   }
      elementData.robot_file= world->robot_directory + elementData.robot_file;
     }
     world->robot.loadFile(elementData.robot_file.c_str());*/
  }



  //Read the directory where to find the obstacle models if they are not given in absolute path:
  if(parseElement(doc, root, "models_directory", elementData)) 
  {
    world->models_directory= path + "/" + elementData.models_directory;
  }

  //Read all the obstacle bodies:
  for(cur= root->xmlChildrenNode; cur!=NULL; cur= cur->next)
  {
     if((!xmlStrcmp(cur->name, (const xmlChar *)"obstacle")))
     { 
        if(parseBody(doc, cur, bodyData))
        { 
          if(bodyData.enable_display==true)
          {
            if(bodyData.enable_collision==true)
            {  bodyType= DISPLAY_AND_COLLISION;  }
            else
            {  bodyType= DISPLAY_ONLY;  }
          }
          else
          {
            if(bodyData.enable_collision==true)
            {  bodyType= COLLISION_ONLY;  }
            else
            {  bodyType= NO_DISPLAY_NO_COLLISION;  }
          }

          if(bodyData.is_primitive==false)
          {
             world->addObstacle(bodyData.name, bodyType, bodyData.file, bodyData.scale, bodyData.T);
          }
          else
          {
            switch(bodyData.primitive_type)
            {
              case BOX:
                body= new Box(bodyData.name, bodyType, bodyData.dimX, bodyData.dimY, bodyData.dimZ);
                world->addObstacle(body, bodyData.T);
              break;
              case CYLINDER:
                body= new Cylinder(bodyData.name, bodyType, bodyData.radius, bodyData.height);
                world->addObstacle(body, bodyData.T);
              break;
              case SPHERE:
                body= new Sphere(bodyData.name, bodyType, bodyData.radius);
                world->addObstacle(body, bodyData.T);
              break;
            }
          }
          world->obstacles.back()->setColor(bodyData.color[0], bodyData.color[1], bodyData.color[2], bodyData.color[3]);
          world->obstacles.back()->setMaterial(bodyData.material);
          world->obstacles.back()->enable_collision= bodyData.enable_collision;
          world->obstacles.back()->enable_display  = bodyData.enable_display;
          world->obstacles.back()->enable_shadow   = bodyData.enable_shadow;
          world->obstacles.back()->setTexture(bodyData.texture);
          world->obstacles.back()->enable_texture  = false;
        }
      }
  }

  //Read the world bounding box:
  if(parseElement(doc, root, "bounding_box", elementData)) 
  {
    world->xmin= elementData.bounding_box[0];
    world->xmax= elementData.bounding_box[1];
    world->ymin= elementData.bounding_box[2];
    world->ymax= elementData.bounding_box[3];
    world->zmin= elementData.bounding_box[4];
    world->zmax= elementData.bounding_box[5];
  }

  //Read the robot configuration:
  if(parseElement(doc, root, "robot_configuration", elementData)) 
  { 
    if( elementData.robot_configuration.size() != world->robot.nbDOFs() )
    {
      message= "<robot_configuration> does not have the appropriate number of elements (it should have " +  convertToString(world->robot.nbDOFs()) + " elements).";
      formatErrorMessage((int) xmlGetLineNo(root), doc->URL, root->name, message);
    }
    else
    {
      for(i=0; i<elementData.robot_configuration.size(); ++i)
      {
        world->robot.getDOFunit(i, unit);
        if(unit==RADIAN)
        {  elementData.robot_configuration[i]*= DEGTORAD;   }
      }
      world->robot.setConfiguration(elementData.robot_configuration);
    }
  }

  xmlFreeDoc(doc);
  return 1;
}



