#if defined(ENABLE_DISPLAY)
#include <stdio.h>
#include <math.h>
#include <GL/freeglut.h>
#include <coldman.h>
#include <vector3D.h>
#include <graphics.h>

//! Draws a tesselated rectangle with its normal vector parallel to z axis.
//! \param bottomLeftCornerX x coordinate of the bottom left corner of the rectangle when seen from the top
//! \param bottomLeftCornerY y coordinate of the bottom left corner of the rectangle when seen from the top
//! \param z z coordinate of the rectangle
//! \param dimX side length of the rectangle along x axis
//! \param dimY side length of the rectangle along y axis
//! \param delta side length of the small squares used to tesselate the rectangle  
void drawTesselatedRectangle(float bottomLeftCornerX, float bottomLeftCornerY, float z, float dimX, float dimY, float delta)
{   
  //  GLboolean enable_cullface;
  unsigned int i, j, nx, ny;
  //  GLint smooth;
  float x1, x2, y, xmax, ymax;

  //glGetBooleanv(GL_CULL_FACE, &enable_cullface);
  //glGetIntegerv(GL_SHADE_MODEL, &smooth);
  glPushAttrib(GL_CULL_FACE | GL_SHADE_MODEL);
//  glPushAttrib(GL_ENABLE_BIT | GL_LIGHTING_BIT);

  xmax= bottomLeftCornerX + dimX;
  ymax= bottomLeftCornerY + dimY;

  nx= (unsigned int) ceil(dimX/delta);
  ny= (unsigned int) ceil(dimY/delta);

  for(i=0; i<=nx; ++i)
  {
    x1= bottomLeftCornerX + i*delta;
    if(x1>xmax)
    {  break;  }
    x2= bottomLeftCornerX + (i+1)*delta;
    if(x2>xmax)
    {  x2= xmax;  }

    glBegin(GL_TRIANGLE_STRIP);
      for(j=0; j<=ny; j++)
      {   
        y= bottomLeftCornerY + j*delta;

        if(y>ymax)
        {  y= ymax; }

        glNormal3f(0.0f, 0.0f, 1.0f);
        glTexCoord2f(x1/dimX, y/dimY);
        glVertex3f(x1, y, z);
        glTexCoord2f(x2/dimX, y/dimY);
        glVertex3f(x2, y, z);
      }
    glEnd();
  } 

//   if(enable_cullface)
//   {  glEnable(GL_CULL_FACE);  }
//   else
//   {  glDisable(GL_CULL_FACE);  }
// 
//   if(smooth)
//   {  glShadeModel(GL_SMOOTH);  }
//   else
//   {  glShadeModel(GL_FLAT);  }

  glPopAttrib();
}


//! Draws an axis-aligned wire box.
//! \param xmin smallest coordinate of the box along X-axis 
//! \param xmax biggest coordinate of the box along X-axis 
//! \param ymin smallest coordinate of the box along Y-axis 
//! \param ymax biggest coordinate of the box along Y-axis 
//! \param zmin smallest coordinate of the box along Z-axis 
//! \param zmax biggest coordinate of the box along Z-axis 
void drawAAbox(double xmin, double xmax, double ymin, double ymax, double zmin, double zmax)
{ 
  //GLint line_width;

  //glGetIntegerv(GL_LINE_WIDTH, &line_width);
  glPushAttrib(GL_LINE_BIT);

  glColor3f(0.0, 0.0, 0.0);
  glLineWidth(3);
  glBegin(GL_LINES);
  glVertex3f(xmin, ymin, zmin);
  glVertex3f(xmin, ymin, zmax);

  glVertex3f(xmax, ymin, zmin);
  glVertex3f(xmax, ymin, zmax);

  glVertex3f(xmax,  ymax, zmin);
  glVertex3f(xmax,  ymax, zmax);

  glVertex3f(xmin, ymax, zmin);
  glVertex3f(xmin, ymax, zmax);
  glEnd();


  glBegin(GL_LINE_LOOP);
  glVertex3f(xmin, ymin, zmin);
  glVertex3f(xmax, ymin, zmin);
  glVertex3f(xmax, ymax, zmin);
  glVertex3f(xmin, ymax, zmin);
  glEnd();

  glBegin(GL_LINE_LOOP);
  glVertex3f(xmin, ymin, zmax);
  glVertex3f(xmax, ymin, zmax);
  glVertex3f(xmax, ymax, zmax);
  glVertex3f(xmin, ymax, zmax);
  glEnd();

//   glLineWidth(line_width);

  glPopAttrib();
}


//! Draws a floor, tiled with rectangles and surrounded by a wire box.
//! \param dx length of a tile along x axis
//! \param dy length of a tile along y axis
//! \param xmin smallest coordinate of the floor along X-axis 
//! \param xmax biggest coordinate of the floor along X-axis 
//! \param ymin smallest coordinate of the floor along Y-axis 
//! \param ymax biggest coordinate of the floor along Y-axis 
//! \param zmin smallest coordinate of the box along Z-axis 
//! \param zmax biggest coordinate of the box along Z-axis 
//! \return 1 in case of success, 0 otherwise
int drawTiledFloor(float dx, float dy, float xmin, float xmax, float ymin, float ymax, float zmin, float zmax)
{
  if( (xmin>=xmax) || (ymin>=ymax) || (zmin>=zmax) )
  {
    printf("%s: %d: drawTiledFloor(): some of the box limit values are inconsistent.\n",__FILE__,__LINE__);
    return 0;
  }

  unsigned int i, j, k, nx, ny;
  float space, delta, shiftX, shiftY, dx2, dy2;
  float center[3], p1[4][3], p2[4][3];
  //float p3[4][3];

  nx= (unsigned int) ceil( (xmax-xmin)/dx );
  ny= (unsigned int) ceil( (ymax-ymin)/dy );
  space= ((dx<dy ? dx : dy)/50.0); //width of the border between the tiles (half of the gap between two adjacent tiles) 
  delta= ((dx<dy ? dx : dy)/20.0); //tesselation size


  //quit if there is too many tiles to display:
  if(nx>100 || ny>100)
  {
    printf("%s: %d: drawTiledFloor(): too many tiles to display: change input values.\n",__FILE__,__LINE__);
    return 0;
  }
glDisable(GL_TEXTURE_2D);


  glPushAttrib(GL_SHADE_MODEL | GL_CULL_FACE);

  glShadeModel(GL_SMOOTH);
  glEnable(GL_CULL_FACE);
  glCullFace(GL_BACK);

  //draw the tiles:
  shiftX= ( (xmax-xmin)/dx - floor((xmax-xmin)/dx) )/2.0;
  shiftX= 0.5*dx - (dx-shiftX);
 
  shiftY= ( (ymax-ymin)/dy - floor((ymax-ymin)/dy) )/2.0;
  shiftY= 0.5*dx - (dy-shiftY);
  

  for(i=0; i<=nx; ++i)
  {
    for(j=0; j<=ny; j++)
      {
        center[0]=  xmin + shiftX + i*dx;
        center[1]=  ymin - shiftY + j*dy;
        center[2]=  zmin;
  
        p1[0][0]= center[0] - 0.5*dx;
        p1[0][1]= center[1] - 0.5*dy; 
        p1[0][2]= center[2]; 
  
        p1[1][0]= center[0] + 0.5*dx;
        p1[1][1]= center[1] - 0.5*dy;
        p1[1][2]= center[2]; 
  
        p1[2][0]= center[0] + 0.5*dx;
        p1[2][1]= center[1] + 0.5*dy; 
        p1[2][2]= center[2]; 
  
        p1[3][0]= center[0] - 0.5*dx;
        p1[3][1]= center[1] + 0.5*dy; 
        p1[3][2]= center[2]; 

        p2[0][0]= p1[0][0] + space;
        p2[0][1]= p1[0][1] + space; 
        p2[0][2]= p1[0][2]; 
  
        p2[1][0]= p1[1][0] - space;
        p2[1][1]= p1[1][1] + space; 
        p2[1][2]= p1[1][2]; 
  
        p2[2][0]= p1[2][0] - space;
        p2[2][1]= p1[2][1] - space; 
        p2[2][2]= p1[2][2]; 
  
        p2[3][0]= p1[3][0] + space;
        p2[3][1]= p1[3][1] - space; 
        p2[3][2]= p1[3][2]; 

//         p3[0][0]= p1[0][0];
//         p3[0][1]= p1[0][1] + space; 
//         p3[0][2]= p1[0][2]; 
//   
//         p3[1][0]= p1[1][0];
//         p3[1][1]= p1[1][1] + space; 
//         p3[1][2]= p1[1][2]; 
//   
//         p3[2][0]= p1[2][0];
//         p3[2][1]= p1[2][1] - space; 
//         p3[2][2]= p1[2][2]; 
//   
//         p3[3][0]= p1[3][0];
//         p3[3][1]= p1[3][1] - space; 
//         p3[3][2]= p1[3][2]; 


        for(k=0; k<4; k++)
        {
          if(p1[k][0]>xmax)   p1[k][0]= xmax;
          if(p1[k][0]<xmin)   p1[k][0]= xmin;
          if(p1[k][1]>ymax)   p1[k][1]= ymax;
          if(p1[k][1]<ymin)   p1[k][1]= ymin;

          if(p2[k][0]>xmax)   p2[k][0]= xmax;
          if(p2[k][0]<xmin)   p2[k][0]= xmin;
          if(p2[k][1]>ymax)   p2[k][1]= ymax;
          if(p2[k][1]<ymin)   p2[k][1]= ymin;

//           if(p3[k][0]>xmax)   p3[k][0]= xmax;
//           if(p3[k][0]<xmin)   p3[k][0]= xmin;
//           if(p3[k][1]>ymax)   p3[k][1]= ymax;
//           if(p3[k][1]<ymin)   p3[k][1]= ymin;
        } 
        
        //draw the inner rectangle of the tile:
        glColor3f(0, 0.8, 1);
        dx2= p2[1][0] - p2[0][0];
        dy2= p2[3][1] - p2[0][1];
        drawTesselatedRectangle(p2[0][0], p2[0][1], p2[0][2], dx2, dy2, delta);


        //draw the tile borders:
        glDisable(GL_LIGHTING);
        glColor4f(0.0, 0.0, 0.0, 1.0);
        glBegin(GL_QUADS);
        glNormal3f(0.0, 0.0, 1.0);
        glVertex3fv(p1[0]);
        glVertex3fv(p1[1]);
        glVertex3fv(p2[1]);
        glVertex3fv(p2[0]);

        glVertex3fv(p1[1]);
        glVertex3fv(p1[2]);
        glVertex3fv(p2[2]);
        glVertex3fv(p2[1]);

        glVertex3fv(p1[2]);
        glVertex3fv(p1[3]);
        glVertex3fv(p2[3]);
        glVertex3fv(p2[2]);

        glVertex3fv(p1[3]);
        glVertex3fv(p1[0]);
        glVertex3fv(p2[0]);
        glVertex3fv(p2[3]);
        glEnd();        
        glEnable(GL_LIGHTING);

//         setShadeLight(LIGHT_POSITION);
//         glColor4f(0.0, 0.0, 0.0, 1.0);
//         dx2= p1[1][0] - p1[0][0];
//         dy2= p3[0][1] - p1[0][1];
//         drawTesselatedRectangle(p1[0][0], p1[0][1], p1[0][2], dx2, dy2, delta);
// 
//         dx2= p3[1][0] - p2[1][0];
//         dy2= p2[3][1] - p2[1][1];
//         drawTesselatedRectangle(p2[1][0], p2[1][1], p2[1][2], dx2, dy2, delta);
// 
//         dx2= p3[2][0] - p3[3][0];
//         dy2= p1[3][1] - p3[3][1];
//         drawTesselatedRectangle(p3[3][0], p3[3][1], p3[3][2], dx2, dy2, delta);
// 
// 
//         dx2= p2[0][0] - p3[0][0];
//         dy2= p3[3][1] - p3[0][1];
//         drawTesselatedRectangle(p3[0][0], p3[0][1], p3[0][2], dx2, dy2, delta);
    }
  }

//   glDisable(GL_CULL_FACE);
  glPopAttrib();


  //draw the wire bounding box:
  glDisable(GL_LIGHTING);
  drawAAbox(xmin, xmax, ymin, ymax, zmin, zmax);
  glEnable(GL_LIGHTING);

  return 1;
}


//! Builds a projection matrix that can be used by OpenGL to project shadows on a plane.
//! \param lightPos position (in homegeneous coordinates) of the light source that will create the shadows
//! \param plane equation of the plane: (a,b,c,d) parameters such that the plane equation is a*x + b*y + c*z + d = 0
//! \param matrix the array the projection matrix is copied in
void buildShadowMatrix(float lightPos[4], float plane[4],  float matrix[16])
{
  float dotp;

  // Calculate the dot-product between the plane and the light's position
  dotp = plane[0] * lightPos[0] +
         plane[1] * lightPos[1] +
         plane[2] * lightPos[2] +
         plane[3] * lightPos[3];

  // First column
  matrix[0]  = dotp - lightPos[0] * plane[0];
  matrix[4]  = 0.0f - lightPos[0] * plane[1];
  matrix[8]  = 0.0f - lightPos[0] * plane[2];
  matrix[12] = 0.0f - lightPos[0] * plane[3];

  // Second column
  matrix[1]  = 0.0f - lightPos[1] * plane[0];
  matrix[5]  = dotp - lightPos[1] * plane[1];
  matrix[9]  = 0.0f - lightPos[1] * plane[2];
  matrix[13] = 0.0f - lightPos[1] * plane[3];

  // Third column
  matrix[2]  = 0.0f - lightPos[2] * plane[0];
  matrix[6]  = 0.0f - lightPos[2] * plane[1];
  matrix[10] = dotp - lightPos[2] * plane[2];
  matrix[14] = 0.0f - lightPos[2] * plane[3];

  // Fourth column
  matrix[3]  = 0.0f - lightPos[3] * plane[0];
  matrix[7]  = 0.0f - lightPos[3] * plane[1];
  matrix[11] = 0.0f - lightPos[3] * plane[2];
  matrix[15] = dotp - lightPos[3] * plane[3];
}


//! Sets the OpenGL projection matrix to an orthographic transformation.
//! \param width width of the display window
//! \param height height of the display window
void setOrthographicProjection(int width, int height)
{
  GLint current_mode;

  glGetIntegerv(GL_MATRIX_MODE, &current_mode);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glOrtho(0, width, 0, height, -1, 1);
  glMatrixMode(current_mode);
}

//! Sets the OpenGL projection matrix to a perspective transformation.
//! \param fov field of view angle (in degrees)
//! \param width width of the display window
//! \param height height of the display window
void setPerspectiveProjection(float fov, int width, int height)
{
  GLint current_mode;

  glGetIntegerv(GL_MATRIX_MODE, &current_mode);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(fov, ((GLdouble) width)/((GLdouble) height), 0.005, 150);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glMatrixMode(current_mode);
}


//! Draws a 3D frame.
//! \param pose a 4x4 homogeneous transform matrix representing the frame pose
//! \param length length of the arrows that will be drawn to represent the frame axes
void drawFrame(HTMatrix &pose, float length)
{
  GLfloat mat[16];

  pose.convertToOpenGLFormat(mat);
 

  glPushAttrib(GL_NORMALIZE | GL_MATRIX_MODE);
//   glPushAttrib(GL_TRANSFORM_BIT);

  glEnable(GL_NORMALIZE);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glMultMatrixf(mat);
    //X axis:
    glColor3f(1.0, 0.0, 0.0);
    glPushMatrix();
      glPushMatrix();
        glTranslatef(0.45*length, 0.0, 0.0);
        glScalef(0.9*length, 0.01*length, 0.01*length);
        glutSolidCube(1.0);
      glPopMatrix();
      glTranslatef(0.9*length, 0.0, 0.0);
      glRotatef(90.0, 0.0, 1.0, 0.0);
      glutSolidCone(0.03*length, 0.1*length, 8, 2);
    glPopMatrix();

    //Y axis:
    glColor3f(0.0, 1.0, 0.0);
    glPushMatrix();
      glPushMatrix();
        glTranslatef(0.0, 0.45*length, 0.0);
        glScalef(0.01*length, 0.9*length, 0.01*length);
        glutSolidCube(1.0);
      glPopMatrix();
      glTranslatef(0.0, 0.9*length, 0.0);
      glRotatef(-90.0, 1.0, 0.0, 0.0);
      glutSolidCone(0.03*length, 0.1*length, 8, 2);
    glPopMatrix();

    //Z axis:
    glColor3f(0.0, 0.0, 1.0);
    glPushMatrix();
      glPushMatrix();
        glTranslatef(0.0, 0.0, 0.45*length);
        glScalef(0.01*length, 0.01*length, 0.9*length);
        glutSolidCube(1.0);
      glPopMatrix();
      glTranslatef(0.0, 0.0, 0.9*length);
      glutSolidCone(0.03*length, 0.1*length, 8, 2);
    glPopMatrix();
 
  glPopMatrix();

  glPopAttrib();
}


//! Draws a cylinder from its two face centers.
//! \param p1 center of the first face (disc) of the cylinder
//! \param p2 center of the second face (disc) of the cylinder
//! \param radius cylinder's radius
int drawCylinder(Vector3D p1, Vector3D p2, float radius)
{
  unsigned int i, n;
  float alpha, dalpha;
  Vector3D d, u, v, c1, c2, normal;

  d= p2 - p1;

  if(d.norm() < EPSILON)
  {
    return 0;
  }  

  d.normalize();
  u= d.orthogonalVector();
  v= crossProduct(d, u);

  n= 20;
  dalpha= 2*M_PI/((float) n);

  alpha= 0;
  glBegin(GL_TRIANGLE_FAN);
  glNormal3f(d.x, d.y, d.z);
  glVertex3f(p2.x, p2.y, p2.z);
  for(i=0; i<n; ++i)
  {
    c1= p2 + radius*cos(alpha)*u + radius*sin(alpha)*v;
    c2= p2 + radius*cos(alpha+dalpha)*u + radius*sin(alpha+dalpha)*v;

    glVertex3f(c1.x, c1.y, c1.z);
    glVertex3f(c2.x, c2.y, c2.z);
    alpha+= dalpha; 
  }
  glEnd();

  alpha= 0;
  glBegin(GL_TRIANGLE_FAN);
  glNormal3f(-d.x, -d.y, -d.z);
  glVertex3f(p1.x, p1.y, p1.z);
  for(i=0; i<n; ++i)
  {
    c1= p1 + radius*cos(alpha)*v + radius*sin(alpha)*u;
    c2= p1 + radius*cos(alpha+dalpha)*v + radius*sin(alpha+dalpha)*u;

    glVertex3f(c1.x, c1.y, c1.z);
    glVertex3f(c2.x, c2.y, c2.z);
    alpha+= dalpha; 
  }
  glEnd();

  alpha= 0;
  glBegin(GL_TRIANGLE_STRIP);
  for(i=0; i<=n; ++i)
  {
    c1= p1 + radius*cos(alpha)*v + radius*sin(alpha)*u;
    c2= p2 + radius*cos(alpha)*v + radius*sin(alpha)*u;

    normal= cos(alpha)*v + sin(alpha)*u;
    glNormal3f(normal.x, normal.y, normal.z);
    glVertex3f(c1.x, c1.y, c1.z);
    glVertex3f(c2.x, c2.y, c2.z);
    alpha+= dalpha; 
  }
  glEnd();


  return 1;
}



//! Saves the current OpenGL pixel buffer as a ppm (PortablePixMap) image file (uncompressed format).
//! In other words: takes a screenshot of the active OpenGL window.
//! \param filename name of the image file where to save the pixel buffer
//! \return 1 in case of success, 0,otherwise
int exportGLdisplay(std::string filename)
{
  size_t pos;
  unsigned int i,j, width, height, length;
  GLint viewport[4];
  unsigned char *pixels= NULL;
  unsigned char *pixels_inv= NULL;
  FILE *file= NULL;

  glGetIntegerv(GL_VIEWPORT, viewport);
  width = viewport[2];
  height= viewport[3];

  pixels    = new unsigned char[3*width*height];
  pixels_inv= new unsigned char[3*width*height];

  length= filename.length(); 

  pos= filename.find(".ppm"); 
  if(length<5 || pos==filename.npos || pos!=length-4)
  {
    printf("%s: %d: exportGLdisplay(): file \"%s\" does not have the good extension (it should be a .ppm file).\n",__FILE__,__LINE__, filename.c_str());
    filename+= ".ppm";
    printf("It is renamed \"%s\".\n",  filename.c_str());
    return 0;
  } 

  file= fopen(filename.c_str(),"w");


  if(file==NULL) 
  {
    printf("%s: %d: exportGLdisplay(): can not open \"%s\".\n",__FILE__,__LINE__,filename.c_str());
    fclose(file);
    return 0;
  }

  glReadBuffer(GL_FRONT);

  // choose 1-byte alignment:
  glPixelStorei(GL_PACK_ALIGNMENT, 1);

  // get the image pixels (from (0,0) position):
  glReadPixels(0, 0, width, height, GL_RGB, GL_UNSIGNED_BYTE, pixels);

  // glReadPixels returns an upside-down image.
  // we have to first flip it
  // NB: in pixels the 3 colors of a pixel follows each other immediately (RGBRGBRGB...RGB).
  for(i=0; i<width; i++)
  { 
    for(j=0; j<height; j++)
    { 
      pixels_inv[3*(i+j*width)]  = pixels[3*(i+(height-j)*width)+0];
      pixels_inv[3*(i+j*width)+1]= pixels[3*(i+(height-j)*width)+1];
      pixels_inv[3*(i+j*width)+2]= pixels[3*(i+(height-j)*width)+2];
    }
  } 

  fprintf(file, "P6\n");
  fprintf(file, "# creator: COLDMAN\n");
  fprintf(file, "%d %d\n", width, height);
  fprintf(file, "255\n");

  fwrite(pixels_inv, sizeof(unsigned char), 3*width*height, file);

  fclose(file);

  delete [] pixels;
  delete [] pixels_inv;

  return 1;
}

//! Draws a cube.
//! \param center cube's center
//! \param edge_length cube's edge length
//! \return OK in case of success, ERROR otherwise
int drawCube(Vector3D center, float edge_length)
{
  glPushMatrix();
   glTranslatef(center.x, center.y, center.z);
   glutSolidCube(edge_length);
  glPopMatrix();

  return OK;
}

#endif
